
#
# Perl Module for EMBL::CC
#
# Cared for by Matthew Pocock <mrp@sanger.ac.uk>
#
# Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE

package EMBL::CC;

use vars qw($AUTOLOAD @ISA @EXPORT_OK %fields);

use Exporter;
use Carp;
use strict;

use EMBL::_LineGroup;
use EMBL::TextFunc;

#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = qw();


BEGIN {
    @ISA = ( 'Exporter', 'EMBL::_LineGroup' );

    %fields = (
	       'comment' => undef,
	       'lineWidth' => undef
    );
}

# class variable to hold width of the comment lines
my $lineWidth = 70;

sub initialize {}

sub new {
    my $ref = shift;
    my $class = ref($ref) || $ref;
    my $self = new EMBL::_LineGroup;

    $self->rebless($class);
    return $self;
}

sub fromEMBLStream {
    my $class = shift;
    my $fh = shift;
    my $lineRef = shift;
    my $lineGroup = new EMBL::CC;
    my $line;
    my $text = undef;

    $line = $$lineRef;
    while($line =~ s/^CC\s{0,3}//) { # remove leading text
	defined($text) && ($text .= " ".$line) || ($text = $line); # add this line to previous text
	$line = <$fh>;
	if(not defined $line) { warn "CC: End of stream"; last; }
    }

    $lineGroup->comment($text);

    $$lineRef = $line;
    return $lineGroup;
}

sub toEMBLStream {
    my $self = shift;
    my $fh = shift;
    my $line;

    foreach $line (EMBL::TextFunc->wrapText($self->comment(), $self->lineWidth() )) {
	print $fh "CC   $line\n";
    }
}

sub comment {
    my $self = shift;
    my $comment = shift; 
    my $switch = shift;

    if (defined $comment) {
	defined $switch && $comment =~ s/\n\s*/ /g;
	return $self->SUPER::comment($comment);
    } else {
	return $self->SUPER::comment();
    }
}

sub rewrap {
    my $self = shift;
    my $comment = $self->comment;
    $comment =~ s/\s*\n/ /g;
    $self->comment($comment);
}

sub lineWidth {
    my $thingy = shift;                        # the thingy being queried for lineWidth
    my $width = shift;                         # optional - the new value for lineWidth

    if(ref $thingy) {                          # if it is $obj->lineWidth
	$width && $thingy->SUPER::lineWidth($width);         # if you need to, set lineWidth for object
	return $thingy->SUPER::lineWidth() || $lineWidth;    # return the width, or the default width
    } else {                                   # if it is class->lineWidth
	$width && ($lineWidth = $width);       # if you need to, set lineWidth for the class
	return $lineWidth;                     # return the width
    }
}

1;  # says use was ok
__END__

=head1 NAME

EMBL::CC

=head1 DESCRIPTION

B<EMBL::CC> represents the CC (comment) lines of an embl entry, as specified in
http://www.ebi.ac.uk/ebi_docs/embl_db/usrman/cc_line.html

B<EMBL::CC> inherits from B<EMBL::_LineGroup>

=head1 AUTHOR

B<Matthew Pocock> Email mrp@sanger.ac.uk

=head1 Member variables

=over

=item comment

This contains a literal string representation of the comment.
It contains all of the text of a single set of comment lines.
On printing, this string is broken up into fragments that fit onto
embl width lines.

If the string contains new lines then curently these are preserved.
However, you can force newlines to be removed by calling something like this:

    $CCline->comment($myCommentText, $stripNL);

where $myCommentText is the text potentialy containing new lines, and $stripNL
is a true/false value which when true causes new lines to be stripped.

=back

=head1 Functions

=over

=item fromEMBLStream

This is a module function that accepts a \*filehandle and \$string as arguments. See
EMBL::_LineGroup for a comment of what they are for. It is designed to be invoked
as 

    $lg = EMBL::CC->fromEMBLStream($fh, \$line);

within the parser. 

=item toEMBLStream

Writes the comment to a stream in embl formatting.
It takes a \*filehandle as its only argument

    $deLine->toStream(\*STDOUT);

=item wrap

Removes all new line information from the comment. This forses all of the text in
the comment to be wrapped to the comment width.

=item lineWidth

Returns and sets the width of the comment. The width is in addition to the five
characters at the beginning of the line. This function can both be applied to
the class and to an object of the class.

    $w = $obj->lineWidth(); # puts the width of $obj into $w
    $w = $obj->lineWidth(60); # set the width to 60 chars, and put that value into $w
    EMBL::XX->lineWidth(70); # sets the default width to 70 chars

If lineWidth is called for a comment that has never had its width set,
then the current default value is returned. e.g.

    EMBL::XX->lineWidth(40);
    $cc1 = new EMBL::CC;
    $cc2 = new EMBL::CC; $cc2->lineWidth(50);
    EMBL::XX->lineWidth(70);
    print $cc->lineWidth(), ",", $cc->lineWidth(), "\n";

will print out '70,50'.
