
#
# Perl Module for EMBL::DT
#
# Cared for by Matthew Pocock <mrp@sanger.ac.uk>
#
# Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE

package EMBL::DT;

use vars qw($AUTOLOAD @ISA @EXPORT_OK %fields);

use Exporter;
use Carp;
use strict;

use EMBL::_LineGroup;

#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = qw();


BEGIN {
    @ISA = ( 'Exporter', 'EMBL::_LineGroup' );

    %fields = (
	       %EMBL::_LineGroup::fields,
	       #Insert field names here as field => undef,
	       'createdDate' => undef,      # creation date
	       'createdRel' => undef,       # release it first appeared in
    	       'updatedDate' => undef,      # last update date
	       'updatedRel' => undef,       # last update release
	       'updatedVer' => undef        # version number
    );
}

sub initialize {}

sub new {
    my $ref = shift;
    my $class = ref($ref) || $ref;
    my $self = new EMBL::_LineGroup;

    $self->rebless($class);

    return $self;
}

sub fromEMBLStream{
    my $class = shift;
    my $fh = shift;
    my $lineRef = shift;
    my $lineGroup = new EMBL::DT;
    my $line;
    
    $line = $$lineRef;
    while(substr($line, 0, 2) eq 'DT') {
	if($line =~ m/DT   (\S+).+?(\d+).*?Created/) {
            # Case of creation date and release
	    my ($date, $rel) = ($1, $2);
	    $lineGroup->createdDate($date);
	    $lineGroup->createdRel($rel);
	}
	elsif($line =~ m/DT   (\S+).+?(\d+).+(\d+)/) {
            # Case of update date and release and version no
	    my ($date, $rel, $ver) = ($1, $2, $3);
	    $lineGroup->updatedDate($date);
	    $lineGroup->updatedRel($rel);
	    $lineGroup->updatedVer($ver);
	}
	$line = <$fh>;
	if(not defined $line) { warn "DT: End of stream"; last; }
    }

    $$lineRef = $line;

    return $lineGroup;
}

sub toEMBLStream {
    my $self = shift;
    my $fh = shift;

    print $fh "DT   ", $self->createdDate, " (Rel. ", $self->createdRel, ", Created)\n";
    print $fh "DT   ", $self->updatedDate || $self->createdDate,
    " (Rel. ", $self->updatedRel || $self->createdRel,
    ", Last updated, Version ", $self->updatedVer || 1, ")\n";

    return $fh;
}

1;  # says use was ok
__END__

=head1 NAME

EMBL::DT

=head1 DESCRIPTION

B<EMBL::DT> represents the DT (date) lines within an embl entry as specified in
http://www.ebi.ac.uk/ebi_docs/embl_db/usrman/dt_line.html

EMBL::DT inherits from B<EMBL::_LineGroup>

=head1 AUTHOR

B<Matthew Pocock> Email mrp@sanger.ac.uk

=head1 Member variables

=over

=item createdDate

The creation date of the entry

=item createdRel

The release number of embl that it first apeared in.

=item updatedDate

The date that this entry was last updated

=item updatedRel

The release number of embl that this update first appeared in.

=item updatedVer

The version number of this update

=back

=head1 functions

=over

=item fromEMBLStream

This is a module function that accepts a \*filehandle and \$string as arguments.
See EMBL::_LineGroup for a description of what they are for.  It is designed to be
invoked as

    $lg = EMBL::DT->fromEMBLStream($fh, \$line);

within the parser.

=item toEMBLStream

Writes the date lines to a stream in embl formatting.  It takes a \*filehandle as
its only argument.  It should behave sensibly when the updated information has not
been provided.
