
#
# Perl Module for EMBL::Feature
#
# Cared for by Matthew Pocock <mrp@sanger.ac.uk>
#
# Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE

package EMBL::Feature;

use vars qw($AUTOLOAD @ISA @EXPORT_OK %fields);

use Exporter;
use Carp;
use strict;

use EMBL::BaseClass;
use Range;

#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = qw();

BEGIN {
    # @ISA has our inheritance.
    @ISA = ( 'Exporter', 'EMBL::BaseClass' );

    %fields = (
	       #Insert field names here as field => value,
	       key => undef,        # feature key
	       location => undef,   # location of feature
               qualifierList => undef, # list of qualifiers
               range => undef,     # bounding range
    );

    # check that you have not tried to define names used by base classes
    EMBL::Feature->check4Clashes();
}

sub initialize {
    my $self = shift;
    $self->{'qualifierList'} = [];
    $self->{'range'} = new Range;
}

sub new {
    my $ref = shift;
    my $class = ref($ref) || $ref;
    my $self = new EMBL::BaseClass;

    $self->rebless($class);
    return $self;
}

sub qualifiers {
    my $self = shift;
    my @qualifiers = @{$self->qualifierList()};
    return @qualifiers;
}

sub fromArray {
    my $class = shift;
    my @qualifierList = @_;
    my $self = new EMBL::Feature;
    @{$self->qualifierList()} = @qualifierList;
    return $self;
}

sub fromEMBLStream {
    my $class = shift;
    my $fh = shift;
    my $lineRef = shift;
    my $feature = new EMBL::Feature;
    my $line;
    my $key;
    my $location;
    my @positions;

    # process key & location
    $line = $$lineRef;
    $key = substr($line, 5, 16);
    chomp($location = substr($line, 21));
    $key =~ s/\s+$//;                # trim trailing white space from keys
    $feature->key($key);

    # the location can span multiple lines
    # add everything up to the first qualifier to the location
    $line = <$fh>;
    while($line =~ m"FT\s{19}[^/]") {
	chomp($line = substr($line, 21));
	$location.=$line;
	$line = <$fh>;
    }
    $feature->location($location);

    while($line =~ m"FT\s{19}/") {
	push @{$feature->qualifierList}, EMBL::Feature::Qualifier->fromEMBLStream($fh, \$line);
	if(not defined $line) { warn "Feature: End of stream"; last; }
    }

    # calculate the bounding box
    $location =~ s/[^,()]*:[^,()]*//g;             # remove alien references
    @positions = sort {$a <=> $b} ($location =~ m/\d+/g); # get sorted list of the numbers in the location list
    $feature->range->start(shift @positions);    # lowest number
    $feature->range->end(pop @positions);        # highest number

    $$lineRef = $line;
    return $feature;
}

sub toEMBLStream {
    my $self = shift;
    my $fh = shift;
    my $key = $self->key;
    my $location = $self->location;
    my @overspill;
    my $line;
    my $qualifier;

    while ($location =~ m/(.{1,57}(,|$))/g) {
	push @overspill, $1;
    }
    @overspill = EMBL::TextFunc->wrapOnCommas($location, 57);
    $location = shift(@overspill);
    print $fh "FT   $key", ' 'x(16-length $key), "$location\n";
    foreach $line (@overspill) {
	print $fh "FT", ' 'x19, $line, "\n";
    }

    foreach $qualifier (@{$self->qualifierList()}) {
	$qualifier->toEMBLStream($fh);
    }

    return $fh;
}

sub filter {
    my $self = shift;
    my $filterFunc = shift;
    my @matches = ();
    my $qualifier;

    foreach $qualifier ($self->qualifiers()) {
        (&$filterFunc($qualifier)) && (push @matches, $qualifier);
    }

    return EMBL::Feature->fromArray(@matches);
}

sub addQualifier {
    my $self = shift;
    my $qualifier = shift;

    warn "Feature $self and qualifier $qualifier:\n\tPlease use addQualifiers() in the future.";
    $self->addQualifiers($qualifier);
}

sub addQualifiers {
    my $self = shift;
    my @qualifiers = @_;
    push @{$self->qualifierList()}, @qualifiers;
}

sub removeQualifiers {
    my $self = shift;
    my @qualifiers = @_;
    my $qualifierList = $self->qualifierList;
    my $qual;
    my $present;
    my $i;
    my $found;

  QUALIFIER:
    foreach $qual (@qualifiers) {
        $i=0;
	$found = 0;
	foreach $present (@$qualifierList) {
#	    print "Comp\t$qual\n\t$present\n\n";
	    if($qual==$present) {
		splice @$qualifierList, $i, 1;
		$found = 1;
		last;
	    }
	    $i++;
	} 
	if(not $found)
	{ carp "No match found for $qual in $self\n"; }
    }
}


1;  # says use was ok

#
# Perl Module for EMBL::Feature::Qualifier
#
# Cared for by Matthew Pocock <mrp@sanger.ac.uk>
#
# Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE

package EMBL::Feature::Qualifier;

use vars qw($AUTOLOAD @ISA @EXPORT_OK %fields);

use Exporter;
use Carp;
use strict;

use EMBL::BaseClass;

#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = qw();


BEGIN {
    # @ISA has our inheritance.
    @ISA = ( 'Exporter', 'EMBL::BaseClass' );

    %fields = (
	       #Insert field names here as field => value,
	       'name' => undef,     # name of qualifier
	       'value' => undef     # value of qualifier
    );

    # check that you have not tried to define names used by base classes
    EMBL::Feature::Qualifier->check4Clashes();
}

sub initialize {
    my $self = shift;
    # place code here to initialize member variables
    # use the syntax $self->{'memberName'} = $val;
    # a good place to set those arrays and hashes to [] & {}
    #

}

sub new {
    my $ref = shift;
    my $class = ref($ref) || $ref;
    my $self = new EMBL::BaseClass;

    $self->rebless($class);
    return $self;
}

sub fromNameValue {
    my $class = shift;
    my $self = new EMBL::Feature::Qualifier;
    my $name = shift;
    my $value = shift;
    $self->name($name);
    $self->value($value);
    return $self;
}

sub fromEMBLStream {
    my $class = shift;
    my $fh = shift;
    my $lineRef = shift;
    my $qualifier = new EMBL::Feature::Qualifier;
    my $line;
    my $name;
    my $value;

    $line = $$lineRef;
    ($name, $value) = $line =~ m"\/([^=]+)=(.*)";
    unless (defined $name) { # handle special case of no value
	($name) = $line =~ m"/(.*)";
	$value = undef;
    }

    $line = <$fh>;
    while( $line =~ m"^FT {19}[^/]" ) {
	chomp($line = substr $line, 21);
#	print "Qualifier: got $line\n";
	$value .= $line. ' ';
	$line = <$fh>;
	unless (defined $line) { warn "Qualifier: End of stream"; last; }
    }

    $qualifier->name($name);
    $qualifier->value($value);

#    print "Name: $name\tValue: $value\n";

    $$lineRef = $line;
    return $qualifier;
}

sub toEMBLStream {
    my $self = shift;
    my $fh = shift;
    my $leader = "FT" . ' 'x19;
    my $line;
    my $value = $self->value();
    my $name = $self->name();
    my $output;

    $line = ($value) ? "/$name=$value" : "/$name";
    $line =~ s/\s+$//;
    while ($line =~ m/(.{1,58}[\",-\s])|(.{1,59})/g) {
	$output = $1 || $2;
#	$output &&
	    print $fh $leader, $output, "\n";
    }

    return $fh;
}

sub getRangeList {
    my $self = shift;
    my @ranges = ();
    my $locationText = $self->location;

    # catch #..()
    if ($locationText =~ m/\d+\.\.\(/) {
	carp "I cannot deal with ranges that look like \#..(\#.\#) or the like";
	return undef;
    }

    # catch ()..#
    if ($locationText =~ m/\)\.\.\d+/) {
	carp "I cannot deal with ranges that look like (\#.\#)..\# or the like";
	return undef;
    }

    # catch acc:location
    if ($locationText =~ m/:/) {
	carp "I cannot deal with locations which contain references to ranges under other accession numbers";
	return undef;
    }

    while ($locationText =~ m/(\d+)\.\.\>?(\d+)/g) {
	push @ranges, new Range($1, $2);
    }

    return @ranges;
}

1;  # says use was ok

# the filter functions for this package
#
# very sparce
#

package Filter;

sub by_name {
    my $pattern = shift;
    return sub {
	my $qualifier = shift;
	return $qualifier->name() =~ m/$pattern/;
    }
}

sub by_value {
    my $pattern = shift;
    return sub {
	my $qualifier = shift;
	return $qualifier->value() =~ m/$pattern/;
    }
}

sub by_nameValue {
    my $name = shift;
    my $value = shift;
    return sub {
	my $qualifier = shift;
	return $qualifier->name() =~ m/$name/ && $qualifier->value() =~ m/$value/;
    }
}

1;
__END__

=head1 Module B<Feature>

=head2 NAME

Feature

=head2 DESCRIPTION

B<Feature> represents a feature within an embl feature table.

Feature inherits from B<EMBL::BaseClass>

=head2 AUTHOR

B<Matthew Pocock> Email mrp@sanger.ac.uk

=head2 FT Member variables

For a complete description of the legal features, their locations and qualifiers
please see http://www.ebi.ac.uk/ebi_docs/embl_db/ft/feature_table.html and the
links from it.

=over

=item key

The key for this feature.

=item location

The location string for this feature

=item qualifierList

A list of qualifier objects associated with this feature

=item range

A Range object that represents the bounds of this object.  Fuzzy ends (i.e.
<# or >#) are localized to the number given, and bounded ends are chosen for
maximum size.

e.g. the location string <250..(423.530) would give a bounding range of 250
to 530.  The location String join(100..230,320..371) would produce a bounding
range of 100 to 371.

Some locations include portions that are outside of the current entry.  For example,

FT   mRNA            join(442..506,X68837:32..2381)

requires you to look at the embl entry with accession B<X68837> and locate the
range B<32..2381> within it.  The bounding box cannot be expected to produce
a meaningfull representation of this location, without being able to do clever
things with multiple entries.  Therefore, I have chosen to B<define> the bounding
range as being the range I<within the current entry> that the feature covers. Thus,
this example's bounding range would be 442 to 506.

In a similar vein, location ranges specified by reference to other feature
locations within the same file are ignored.

If you think it should be different, tell me.

=back

=head2 FT Functions

=over

=item new

Returns a new empty EMBL::Feature object.

=item fromArray

Takes an array of EMBL::Feature::Qualifier objects, and returns a new feature containing
those objects.

=item qualifiers

Returns an array of all of the EMBL::Feature::Qualifier objects within this feature -
analogous to EMBL::Set::Entries and EMBL::FeatureTable::Features.

=item getRangeList

This returns an array of Range objects that represent the ranges
(things looking like #..#) in the location of this object. It is
particularly usefull when you want a set of ranges from a CDS feature.
It will compain if any of the range objects don't look like the
above pattern except if one of the bases is fuzzy.
It will currently also complain if a range is associated with an
accession number.

If for any reason the function falis, it returns C<undef>.  Otherwise,
an array of ranges is returned.  Both the array and the ranges therin
can be edited without changing the values of the feature location.

=item fromEMBLStream

Although this class does not derive from a B<EMBL::_LineGroup>, I have presereved 
the interface, as this object has to read it's self from the embl stream. It is
called from within EMBL::FeatureTable::fromEMBLStream, and returns an Feature
object representing the next feature in the table.

=item toEMBLStream

Writes the feature out to a stream in embl format.

=item addQualifier

Please use addQualifiers(). This function will warn with a message saying use addQualifiers.

=item addQualifiers

Adds a list of qualifiers to this feature. They will be added at the end of the
list of qualifiers. This method is prefered over addQualifier.

e.g.

 $qualifier = EMBL::Feature::Qualifier->fromNameValue('note', 'an interesting thing to know');
 $qualifier2 = EMBL::Feature::Qualifier->fromNameValue('info', 'this is from the documentation');
 $feature->addQualifiers($qualifier);
 $feature2->addQualifiers($qualifier, $qualifier2);

=item filter

This works in an analogus way to EMBL::FeatureTable::filter.  In this case, it
returns a feature composed of the qualifiers that were flagged as matching
the filter function.

There are two filter functions defined as standard.

=over

=item Filter::by_name

Takes a reg-ex. Will cause all of the qualifiers with a name that is matched
by the reg-ex to appear in the returned feature object.

=item Filter::by_value

As above, but matches to the value of the qualifiers.

=item Filter::by_nameValue

Pass it two arguments. The first is a name to
match and the seccond is a value to match.  They
will both be treated as patterns.  Filters in
only the qualifiers that match both name and
value. Usefull when you want all of the note
lines containing a keyword.

=back

=back

=head1 Module B<Feature::Qualifier>

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif"><p>

B<This is documentation for how the qualifiers are stoored within a feature.
You should only read this if you wish to work inimately with the qualifiers.>

=head2 NAME

Feature::Qualifier

=head2 DESCRIPTION

B<Feature::Qualifier> represents the name/value of a feature qualifier.
At the moment, no checking is done to check the legality of either
name or value.

Feature::Qualifier inherits from B<BaseClass>

=head2 AUTHOR

B<Matthew Pocock> Email mrp@sanger.ac.uk

=head2 Member variables

=over

=item name

The name of the qualifier - found by looking for /name=value

=item value

The value of the qualifier - copes with values that span multiple lines.

=back

=head2 Functions

=over

=item new

Returns a new EMBL::Feature::Qualifier object

=item fromNameValue

Constructs a new EMBL::Feature::Qualifier. Accepts two arguments.  The first will be taken to
be the qualifiers name, and the seccond will be it's value.

=item fromEMBLStream

Takes control of parsing the embl stream, and returns an EMBL::Feature::Qualifier object.

=item toEMBLStream

Prints out this embl qualifier.




