
#
# Perl Module for EMBL::FeatureTable
#
# Cared for by Matthew Pocock <mrp@sanger.ac.uk>
#
# Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE

package EMBL::FeatureTable;

use vars qw($AUTOLOAD @ISA @EXPORT_OK %fields %order);

use Exporter;
use Carp;
use strict;

use EMBL::BaseClass;
use SPEM::Iterator;
use EMBL::Feature;

#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = qw();

BEGIN {
    # @ISA has our inheritance.
    @ISA = qw( Exporter EMBL::BaseClass SPEM::Iterator );

    %fields = (
	       #Insert field names here as field => value,
	       'featureList' => undef
    );

    # check that you have not tried to define names used by base classes
    EMBL::FeatureTable->check4Clashes();

    %order = ( # default search order - most negative are sorted to the top
	      'source' => -100,                        # always first
	      'gene' => -10,                           # )
	      'mRNA' => -10,                           # >these together - then tiebreaked
	      'CDS' => -10,                            # )
	      'prim_transcript' => -8,
	      'misc_feature' => -6,
	      'unsure' => -4,
	      'variation' => -2,
	      'repeat_region' => 1,                    # always last
	      );
}

sub initialize {
    my $self = shift;
    $self->featureList([]);
}

sub new {
    my $ref = shift;
    my $class = ref($ref) || $ref;
    my $self = new EMBL::BaseClass;

    $self->rebless($class);
    return $self;
}

sub fromArray {
    my $class = shift;
    my @features = @_;
    my $self = new EMBL::FeatureTable;

    @{$self->featureList()} = @_;

    return $self;
}

sub iterator {
    my( $self ) = @_;
    
    return SPEM::Iterator->new($self, $self->featureList());
}

sub addFeature {
    my( $self,
        $what,
        $where,
        ) = @_;
    my $features = $self->featureList;

    unless(defined $what) {
	confess 'Use: $entry->($what [, $where])';
    }
        
    unless (defined $where) {
        # Add this item at sensibly sorted position
	local $a;
	local $b = $what;
        $where = -1;
	do { 
	    $where++;
	    $a = $features->[$where];
	} while ( defined( $a ) and _defaultSortFunction() < 1 );
    } elsif ($where =~ /^\d+$/) {
        # That's OK
    } elsif (ref($where)) {
        # Add at position occupied by feature given
        $where = $self->where( $where );
    } elsif ($where =~ /^[+-]?\d+$/) {
        # Set to end if off end of array
        $where = @$features if $where > @$features;
    } elsif ($where =~ /start/i) {
        $where = 0;
    } elsif ($where =~ /end/i) {
        $where = @$features;
    } else {
        confess("Invalid position argument ('$where') to addFeature()");
    }

    $self->maintainIteratorList('ADD', $where);
    return splice( @$features, $where, 0, $what );
}

sub removeFeature {
    my $self = shift;
    my $where = shift;

    unless(defined $where) {
	carp 'Use: $entry->($where)';
	return undef;
    }
    
    # Are we given the feature itself?
    if (ref($where)) {
        $where = $self->where( $where );
    }
    
    $self->maintainIteratorList('REMOVE', $where);
    return splice( @{$self->featureList}, $where, 1 );
}

sub get {
    my $self = shift;
    my $where = shift;

    return $self->featureList->[$where];
}


sub where {
    my( $self, $feature ) = @_;
    
    for (my $i = 0; $i < @{$self->featureList}; $i++) {
        return $i if $feature eq $self->featureList->[$i];
    }
    return;
}


sub fromEMBLStream {
    my $class = shift;
    my $fh = shift;
    my $lineRef = shift;
    my $lineGroup = new EMBL::FeatureTable;
    my $line;
    my $tag;
    my $feature;

    $line = $$lineRef;
    $tag = substr($line, 0, 2);
    while(1) {
	if ($tag eq 'FT') {
	    if (substr($line, 6, 1) ne ' ') { # if there is one then add it to the list so far
		push @{$lineGroup->featureList}, EMBL::Feature->fromEMBLStream($fh, \$line);
		next;
	    }
	    $line = <$fh>;                    # get the next line
	    next;                             # not found a key, so suck in the next line
	}
	if ($tag eq 'FH') {                   # heading line
            $line = <$fh>;                    # get the next line
	    next;                             # discard it
	}
	last;                                 # don't understand the line so finnish parsing stream
    } continue {
	if(not defined $line) { warn "FeatureTable: End of stream"; last; }
	$tag = substr($line, 0, 2);
    }

    $$lineRef = $line;
    return $lineGroup;
}

sub toEMBLStream {
    my $self = shift;
    my $fh = shift;
    my $feature;

    print $fh "FH   Key             Location/Qualifiers\n";
    print $fh "FH\n";

    foreach $feature (@{$self->featureList}) {
	$feature->toEMBLStream($fh);
    }

    return $fh;
}

sub features {
    my $self = shift;
    return @{$self->featureList()};
}

sub sort {
    my $self = shift;
    my $sortFunc = shift || \*_defaultSortFunction;
    my @features = @{$self->featureList()};
    
    @features = sort $sortFunc @features;
    $self->featureList(\@features);
    return $self;
}

sub _defaultSortFunction {
    my $aKey = $a->key();
    my $bKey = $b->key();
    my $val = $order{$aKey} <=> $order{$bKey};
    ($val) && return $val;  # clear order on the basis of their keys

    if($aKey eq 'repeat_region') {
	my $aIsRep = $a->filter(Filter::by_nameValue('note', 'consensus'))->qualifiers();
	my $bIsRep = $b->filter(Filter::by_nameValue('note', 'consensus'))->qualifiers();
	$val = $aIsRep <=> $bIsRep;
	($val) && return $val;
    }

    # same keys so sort by location start
    my $aRange = $a->range();
    my $bRange = $b->range();
    $val = $aRange->start() <=> $bRange->start();
    ($val) && return $val;  # clear order on the basis of location start

    return $aRange->length() <=> $bRange->length();
}

sub getFeaturesOverlapping {
    my $self = shift;
    my $range = shift;
    my @overlap = ();
    my $feature;
    my $start = $range->start;
    my $end = $range->end;
    my $featRange;
    my $fstart;
    my $fend;

    foreach $feature ($self->features() ) {
	$featRange = $feature->range();
	$fstart = $featRange->start();
	$fend = $featRange->end();
	unless(
	       ($fstart <=> $start) == ($fend <=> $start) && # start and end same side of query start
	       ($fstart <=> $end) == ($fend <=> $end) &&     # start and end same side of query end
	       ($fstart <=> $start) == ($fend <=> $end)      # the two 'sides' are the same side
	       ) {
	    push @overlap, $feature;
	}
    }

    return @overlap;
}

sub getFeaturesWithin {
    my $self = shift;
    my $range = shift;
    my $start = $range->start;
    my $end = $range->end;
    my @within = ();
    my $feature;
    my $frange;

    foreach $feature (@{$self->features()}) {
	if( ($frange->start > $start) && ($frange->end < $end) ) {
	    push @within, $feature;
	}
    }

    return @within;
}

sub filter {
    my $self = shift;
    my $filterFunc = shift;
    my @matches = ();
    my $feature;

    foreach $feature ($self->features()) {
	(&$filterFunc($feature)) && (push @matches, $feature);
    }

    return EMBL::FeatureTable->fromArray(@matches);
}

1;

package Filter;

# convenient name space for filter functions

sub by_key {
    my $pattern = shift;
    return sub {
	my $feature = shift;
	return $feature->key() =~ m/$pattern/;
    }
}

sub within_range {
    my $range = shift;
    return sub {
	my $feature = shift;
	my $bounds = $feature->range();
	return $bounds->within($range);
    }
}

sub overlapping_range {
    my $range = shift;
    return sub {
	my $feature = shift;
	my $bounds = $feature->range();
	return $bounds->overlapping($range);
    }
}

1;  # says use was ok
__END__

=head1 NAME

FeatureTable

=head1 DESCRIPTION

B<FeatureTable> represents the entire feature table (FH and FT lines).
It contains a list of B<Feature> objects, which represent each feature.
There will be an API for editing the feature list.

FeatureTable inherits from B<EMBL::BaseClass>

=head1 AUTHOR

B<Matthew Pocock> Email mrp@sanger.ac.uk

=head1 Member variables

=over

=item featureList

An array of features that this feature table comprises of.

=back

=head1 Functions

=item new

Creates a new empty feature table.

=item fromArray

Takes an array of feature objects, and returns a new feature table containing
those features.

=item fromEMBLStream

Your standard fromEMBLStream function.

    $ftLine = FeatireTable->fromEMBLStream(\*FH, \$line);

=item toEMBLStream

Prints the feature table to a stream in embl format.
This includes the FH lines as well as the FT body.

    $ftLine->toEMBLStream(\*STDOUT);

=item addFeature

Add a feature object to this embl group.  Optionaly specify where to add it.

    $ftable->addFeature($what [, $where]);

If where is absent, then the feature will be added at a position based upon
its sort order.  If the feature table is not sorted, this will be the first
position where the feature before has a lower precidence than it, and the
one after has the higher precidence - as calculated by the module's default
sort function.

If it is START or END, it will be added accordingly.  If it is a number, then
this is taken to be an index (starting at 0) for where to add it.  Should the
index be outside the range of the entry, it will be added at the end.

=item removeFeature

Removes the feature at the specified index. Returns the object removed,
or C<undef>.

    $ftable->removeFeature($where);

=item get

Returns the feature object at the location you specify.

    $ftable->get(4);

=item features

Returns an array conaining a reference to each feature object in the table.
You can edit this array (splice, shift, etc.) and the original feature table
will remain intact.  If you change any of the features contained within it
using thier member variables, then they will be altered in the feature table.

    @features = $ft->features();


=item sort

Sorts the feature table.  Supply it with a standard sort comparrison
function, and it will use that to sort the feature table.  Otherwise, it
will use my default sort rutine.

If the default sort is consistently not doing what people want, then
it can always be changed to be more reasonable.  I think it does a reasonable
job at the moment.

    $ft->sort();     # just sorts using the default sorting routine
    $ft->sort( sub { $a->key cmp $b->key } ); # sorts by the key name

=item getFeaturesOverlapping

Takes a Range object, assumed to refer to a range within the co-ordinates
of the feature table. Returns an array of features (possibly empty) of all the
features within this table that overlap the range.

    $range = new Range(230, 312);  # make a range for the query
    @hits = $featureTable->getFeaturesOverlapping($range); # get a new feature table
                                   # containing the features overlaping $range

You should consider using the filter function to do this, as it returns a
feature table object that can be queried again.

=item getFeaturesWithin

Tages a Range object, assumed to refer to a range within the co-ordinates
of the feature table. Returns an array (possibly empty) of all the
features within this table that contained within the range.

    $range = new Range(230, 312);
    @hits = $featureTable->getFeaturesWithin($range);

You should consider using the filter function to do this, as it returns a
feature table obejct that can be queried again.

=item filter

Returns a new feature table containing all of the features that match a search
criteria.  It takes a function as the only parameter.  This function should expect
a feature, and return true or false depending on whether the feature matches
the criteria.

    # e.g. imagine $source returns true if the feature has 'source' as its key
    $hits = $ft->filter($source); # $hits now contains a feature table containing the source feature

This is quite static in this form, but imagine a call like this:

    $hits = $ft->filter(Filter::by_key('source'));

Here something clever is happening.  This manages to return the same result as the
previous example, but does it in a more flexible way using closure and anonymous
sub-routines.

Here is a list of current filter functions.  See Roll your own filter functions later in
this documentation to see how to make your own.

=over

=item Filter::by_key

Takes a string which is assumed to be a reg-ex to be applied to the key of a feature.
If the reg-ex matches, C<Filter::by_key> returns true, otherwise false.

    $exons = $ft->filter(Filter::by_key('exon'));  # all features where the key contains 'exon'
    $misc = $ft->filter(Filter::by_key('misc'));   # all features where the key contains 'misc'

=item Filter::within_range

Takes a range object. Will cause all features with a range within the one specified to be
filtered into the return object.

    $within = $ft->filter(Filter::within_range($range)); # all features within the range $range

=item Filter::overlapping_range

Takes a range obejct. Will cause all features that overlap the range to tbe filtered into the
return set.

    $overlap = $ft->filter(Filter::overlapping_range($range)); # all features overlapping $range

=back

=item _defaultSortFunction

This is the internal function used to impliment sort when called without
specifying a comparrison function.  It does the following comparisons:

=over

=item 1

Looks up the keys for a & b in %FeatureTable::order{} and use the return
value to order the keys. This hash contains
values for several of the common keys which will sort them sensibly. Ask
me for details and/or modifications.

=item 2

If the keys gave no ordering, then the key is checked to see if the feature
is 'repeat_region'. If it is, then both features are checked for the text
'consensus' in their 'note' qualifiers. This indicates that the repeat is an
element repeat, not a low complexity repeat. The pattern repeats will be
sorted above the low complexity repeats

=item 3

If an order has still not been found, then the features are sorted on the
basis of the beginning of their bounding box.

=item 4

In the unlikely event that two features are still unsorted, they are ordered
by the size of their bounding box

=item5

Should they still be tied, a value of 0 is returned to indicate that there
is no way of sorting the feature pair.

=back

=back

=head1 Roll your own filter functions

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif"><p>

For most purposes, the usual filter functions should surfice. However, in keeping with
the spirit of extencibility, here is how to write your own.

=head2 A simple example

This is a simple filter function.  It takes a feature as it's parameter, and returns true
if the feature key is equal to 'source'.

    sub is_source {
	my $feature = shift;
	my $key = $feature->key();
	return $key eq 'source';
    }

This is called as C<is_source($feature)> in the function C<FeatureTable::filter>.

=head2 A more hairy example

This is the code that makes the Filter::by_key work.

    sub by_key {
	my $pattern = shift;
	return sub {
	    my $feature = shift;
	    return $feature->key() =~ m/$pattern/;
	}
    }

This probably needs explaining. You would say $hits = $ft->filter(Filter::by_key($myPattern)).
The expression C<Filter::by_key($myPattern)> returns an anonymous subroutine. Within this sub,
the value of $pattern is set to $myPattern by closure (deep!), but the value of $feature is
not yet determined.  Then, in C<FeatureTable::Feature> the anonymous function is executed for
each feature, and $feature becomes the feature being tested. In this way, it apears to the
user that somehow the function C<Filter::by_key> is used to discriminate between features
with or without $myPattern in their keys. Neat!
