
#
# Perl Module for EMBL::ID
#
# Cared for by Matthew Pocock <mrp@sanger.ac.uk>
#
# Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE

package EMBL::ID;

use vars qw($AUTOLOAD @ISA @EXPORT_OK %fields);

use Exporter;
use Carp;
use strict;

use EMBL::_LineGroup;

#
# @ISA has our inheritance.
#

#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = qw();

BEGIN {
    @ISA = ( 'Exporter', 'EMBL::_LineGroup' );

    %fields = (
	       #Insert field names here as field => undef,
	       entryname => undef, # the ID number
	       dataclass => undef, # type of sequence
	       molecule => undef,  # type of molecule
	       division => undef,  # division of embl
	       seqlength => undef  # length of the sequence
	       );
  EMBL::ID->check4Clashes();
}

sub initialize {}

sub new {
    my $ref = shift;
    my $class = ref($ref) || $ref;
    my $self = new EMBL::_LineGroup;

    $self->rebless($class);
    return $self;
}

sub fromEMBLStream {
    my $class = shift;
    my $fh = shift;
    my $lineRef = shift;
    my $LG = new EMBL::ID;
    my $line = $$lineRef;

    @_ = $line =~ m/ID   (\w+)\s+(\w+);\s+(\w+);\s+(\w+);\s+(\w+)/;

    $LG->entryname(shift);
    $LG->dataclass(shift);
    $LG->molecule(shift);
    $LG->division(shift);
    $LG->seqlength(shift);

    $$lineRef = <$fh>;
    return $LG;
}

sub toEMBLStream {
    my $self = shift;
    my $fh = shift;

    print $fh "ID   ",
    $self->entryname, " " x (11 - length $self->entryname),
    $self->dataclass, "; ",
    $self->molecule, "; ",
    $self->division, "; ",
    $self->seqlength, " BP.\n";

    return $fh;
}

1;  # says use was ok
__END__

=head1 NAME

EMBL::ID

=head1 DESCRIPTION

B<EMBL::ID> represents the ID (identification) line that starts an EMBL entry
as specified in http://www.ebi.ac.uk/ebi_docs/embl_db/usrman/id_line.html

=head1 AUTHOR

B<Matthew Pocock> Email mrp@sanger.ac.uk

=head1 Member variables

The variable names are taken directly from the description at
http://www.ebi.ac.uk/ebi_docs/embl_db/usrman/id_line.html

=over

=item entryname

=item dataclass

The type of sequence

=item molecule

=item division

=item seqlength

=back

=head1 Functions

=over

=item fromEMBLStream

This is a module function that accepts a \*filehandle and \$string as arguments and
returns an EMBL::ID object.
See EMBL::_LineGroup for a description of what they are for.  It is designed to be
invoked as

    $lg = EMBL::ID->fromEMBLStream($fh, \$line);

within the parser.

=item toEMBLStream

Writes the ID line out to a stream in embl format.
It takes a \*filehandle as its only argument.

    $idLine->toStream(\*STDOUT);
