
#
# Perl Module for EMBL::Parser
#
# Cared for by Matthew Pocock <mrp@sanger.ac.uk>
#
# Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE

package EMBL::Parser;

use vars qw($AUTOLOAD @ISA @EXPORT_OK);
use Exporter;
use Carp;
use strict;
use EMBL::Set;
use EMBL::Entry;

use SPEM::Parser;

#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = qw();

#
# @ISA has our inheritance.
#

@ISA = ( 'Exporter', 'SPEM::Parser' );

sub new {
    my $ref = shift;
    my $class = ref($ref) || $ref;
    my $self = new SPEM::Parser;
    bless $self, $class;

    $self->{'functionTable'} = {};

    return $self;
}

sub newEntry {
    return new EMBL::Entry;
}

sub newSet {
    return new EMBL::Set;
}

sub newLine {
    my $self = shift;
    my $value = shift;
    my $fh = shift;
    my $lineRef = shift;
    my $line = $value->fromEMBLStream($fh, $lineRef);

    if($line->can('toEMBLStream')) {
	return $line;
    } else {
	carp "The object $line returned from $value ->toStream does not have a 'toEMBLStream function";
    }
}

1;  # says use was ok
__END__

=head1 NAME

EMBL::Parser

=head1 DESCRIPTION

Description for B<EMBL::Parser>

=head1 AUTHOR

B<Matthew Pocock> Email mrp@sanger.ac.uk

=head1 Call back functions, modules and objects

You use the parser object to register a function, module or object as
being interested in a particular line type.  It provides you with great
flexibility in extending the parser to do different things with the
embl files it reads.

If you chose to register a function, it should expect to be called with
two arguments only.  The first will be a \*FH, and the second a reference
to a scalar.

If you chose to register a module or object, a function C<fromEMBLStream>
will be invoked on it with arguments \*FH and a reference to a scalar.
However, as it will be invoked with the arrow operator, the first argument
that it actualy recieves will be a class name, or a blessed reference.

In either case, the function invoked should return a line group object.
This should forfill the interface conditions for a line group, as
described in the section L<Roll your own line groups>

=head1 Functions

=over

=item setFromStream

Parses a stream, and returns an EMBL::Set object containing the embl
entries from the stream.  Parse it a glob to a filehandle as it's only
argument.

=item entryFromStream

Parse it a stream, and a referece to a buffer containing the current line of text.
It will consume one entrys worth of the stream, and return a new EMBL::Entry object.

=item register

Used to register a tag and function or module with an instance of EMBL::Parser.
Use it like this:

    $parser->register($tag, $module);
    $parser->register($tag, $object);
 or
    $parser->register($tag, \&function);

If you try to register a tag that has already been used, then a warning
will be printed, but the new item will replace the old one.

If you register a function, then it must be the fully qualified function name.
It will be assumed to take only the reference to a glob to a file handle, and
the reference to a line of text.  Of course, in this case you can specify any
function name you wish, allowing any function in any package to produce the
line groups of the type you want. It will be invoked as &$funcName(\*FH, \$line).

If you register a module, then that module must provide a fromEMBLStream function.
It will be assumed to take a class name, the reference to a glob to a file handle,
and the reference to a line of text.
It will be invoked as $module->fromEMBLStream(\*FH, \$line)

Likewise, if you register an object, then that object must provide a
fromEMBLStream function.  It will be assumed to take an object, the reference to
a glob to a file handle, and the reference to a line of text.
It will be invoked as $object->fromEMBLStream(\*FH, \$line)

=item unregister

Removes a tag and it's associated function from the Parser.

Use:

    $parser->unregister($tag);

=item registerDefault

Registers a function to call when no other function is interested in a
line.

Use:

    $parser->registerDefault(\&function);
 or
    $parser->registerDefault($module);
    $parser->registerDefault($object);

See register above for a description of the two syntaxes.

=back

=head1 Example of modifying the parser

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif">This information is
intended only for developers who wish to make custom parsers.

Let us assume that you prety much like what the default does, but want to let the
default behavure be to image lines.

=for html
<PRE>
$parser = Embl->defaultParser();
$parser->registerDefault(EMBL::Image);
</PRE>

Now, you wish to unregister the parsing of the date lines.

=for html
<PRE>
$parser->unregister('DT');
</PRE>

Imagine that you are reading EMBL type files from a source where there is a
systematic error in the organism lines OS.  You wish to put your own function
which knows how to correct this, called C<myPackage::correctOS>. You would now
do the following, to change the OS line parsing, without getting warnings.

=for html
<PRE>
# unregister previous interested functions and modules
$parser->unregister('OS');
$parser->unregister('OC');
$parser->unregister('OG');
# now register my function
$parser->register('OS', \&myPackage::correctOS() );
</PRE>

You are now ready to go parse - with a command like this.

    $set = $parse->fromStream(\*FH);

=head1 Roll your own line groups

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif">This information is
intended only for developers who wish to make custom line types.  If this
is not you then you realy do not need to read this.<p>

For an object to function as a line group that can be created from a stream
and writen to a stream, there are two pieces of functionality it must have.
Firstly, it must have a way of being read from a stream.  Secondly it must
have a function toEMBLStream that allows it to be writen to a stream. Sounds
obvious, doesn't it.

The following sections describe both of these steps, together with code
examples, and I hope it explores the major possible different ways to go
at each step.

=head2 Reading your object from the EMBL formatted stream

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif"><p>

There are three possible ways that a line group can be created during parsing.
They are listed here

=over

=item 1

Registering a function to the parser (as C<\&module::funcName>)

Invoked as &module::funcName(\*FH, \$line)

=item 2

Registering a module with a C<fromEMBLStream> function with the parser

Invoked as $module->fromEMBLStream(\*FH, \$line)

=item3

Registering an object (a factory etc.) with a C<fromEMBLStream> function
with the parser

Invoked as $object->fromEMBLStream(\*FH, \$line)

=back

In each case, the return value of the function will be assumed to have
a C<toEMBLStream> function, and will be added to the entry currently
being built.

So here are some code outlines for how to start these functions.

=head3 Absolute function

This will be invoked directly.  No class parameters will be given to it,
or anything like that, so it will look something like

=for html
<PRE>
sub myLineGroupCreatingFunction {
    my $fh = shift;         # the stream being read from
    my $lineRef = shift;    # a reference to the next line to be parsed
    my $lineGroup = new lineOfMyType();
#
# code here to process the file and update $lineRef
#
    return $lineGroup;
}
</PRE>

=head3 Module function

Remeber - this will be invoked using the arrow operator.

=for html
<PRE>
sub fromEMBLStream {
    my $class = shift;      # given my class type
    my $fh = shift;         # the stream being read from
    my $lineRef = shift;    # a reference to the next line to be parsed
    my $lineGroup = new $class;  # ok - so this is a little bit risky
                                 # as we may be invoked by a derived
                                 # class, and end up constructing an
                                 # object with differrent behavure to
                                 # the expected.  See later for how to
                                 # deal with this.
#
# code here to process the file and update $lineRef
#
    return $lineGroup;
}
</PRE>

=head3 Object method

This will be invoked as $object->fromEMBLStream, but may return objects
of a different type to it's self.  This could impliment a class factory,
or just return an object of the same type as it.

=for html
<PRE>
sub fromEMBLStream {
    my $self = shift;       # given my class type
    my $fh = shift;         # the stream being read from
    my $lineRef = shift;    # a reference to the next line to be parsed
    my $class = something decided by $self
    my $lineGroup = new $class;  # ok - so this is a little bit risky
                                 # as we may be invoked by a derived
                                 # class, and end up constructing an
                                 # object with differrent behavure to
                                 # the expected.  See later for how to
                                 # deal with this.
#
# code here to process the file and update $lineRef
#
    return $lineGroup;
}
</PRE>

=head2 the fromEMBLStream function

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif"><p>

There are essentialy two types of fromEMBLStream functions.  The first eats one
line of text.  The seccond eats multiple lines.  I shall give an example of the
code for both, as taken from real examples.

=over

=item Single line eater

=for html
The source code for this is taken from <a href = "Image.html">EMBL::Image</a><p>

=for html
<PRE>
sub fromEMBLStream {
    my $class = shift;           # our class
    my $fh = shift;              # the file handle
    my $lineRef = shift;         # the line reference
    my $line = new EMBL::Image;  # make the image

    $line->text($$lineRef);      # fill the image

    $$lineRef = <$fh>;           # prime the next line into $$lineRef
    return $line;                # return the new image
}

=for HTML
</PRE>

=item Multiple line eater

=for html
The source code for this example is from <a href = "CC.html">EMBL::CC</a><p>

=for html
<PRE>
sub fromEMBLStream {
    my $class = shift;                   # our class
    my $fh = shift;                      # file handle
    my $lineRef = shift;                 # line reference
    my $lineGroup = new EMBL::CC;        # make a new comment
    my $line;                            # to hold a line of input
    my $text = undef;                    # the text so far

    $line = $$lineRef;                   # prime $line
    while(substr($line, 0, 2) eq 'CC') { # while we are a comment
        chomp($line = substr $line, 5);  # remove leading text
        defined($text) && ($text .= " ".$line) || ($text = $line); # add this line to previous text
        $line = <$fh>;                   # get the next line
        if(not defined $line) { warn "CC: End of stream"; last; } # end of file check
    }
    $lineGroup->comment($text);          # set the comment to $text

    $$lineRef = $line;                   # prime lineRef
    return $lineGroup;                   # return the new comment
}

=for html
</PRE>

=back

=head2 Writing your object to an EMBL formatted stream

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif"><p>

When an EMBL::Set is asked to write to a stream, it calles toEMBLStream on all of
it's entries. Each entry in turn calles toEMBLStream on eacch of it's line groups
which then write themselves to the stream. For your object to fit into this
scheim, it must have a toEMBLStream function with the correct signature and
functionality.

Here is a skeletal outline of what it will look like.

=for html
<PRE>
sub toEMBLStream {
    my $self = shift;
    my $fh = shift;

    print $fh $self->asEMBLFormattedText; # may well do something more fancy than this

    return $fh;   # the return value is not currently used but I prefer this
                  # notation, because it allows chaining of toEMBLStream calls.
}

=for html
</PRE>
