
#
# Perl Module for EMBL::Reference
#
# Cared for by Matthew Pocock <mrp@sanger.ac.uk>
#
# Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE

package EMBL::Reference;

use vars qw($AUTOLOAD @ISA @EXPORT_OK %fields);

use Exporter;
use Carp;
use strict;

use EMBL::_LineGroup;
use EMBL::TextFunc;

#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = qw();


BEGIN {
    @ISA = ( 'Exporter', 'EMBL::_LineGroup' );

    %fields = (
        #Insert field names here as field => value,
	       'refNo' => undef,              # RN the reference number
	       'comment' => undef,            # RC multiline comment
	       'position' => undef,           # RP string representing the sequence ranges from this reference
	       'crossReferences' => undef,    # RX cross reference list e.g. to medline
	       'authorList' => undef,         # RA list of authors
	       'title' => undef,              # RT multiline title
	       'locationLines' => undef,      # RL list of all the reference location lines
    );

    # check that you have not tried to define names used by base classes
    EMBL::Reference->check4Clashes();
}

sub initialize {
    my $self = shift;
    $self->{'crossReferences'} = [];
    $self->{'authorList'} = [];
    $self->{'locationLines'} = [];
}

sub new {
    my $ref = shift;
    my $class = ref($ref) || $ref;
    my $self = new EMBL::_LineGroup;

    $self->rebless($class);
    return $self;
}

sub fromEMBLStream {
    my $class = shift;
    my $fh = shift;
    my $lineRef = shift;
    my $lineGroup = new EMBL::Reference;
    my $line;
    my $tag;
    my $comment = "";
    my $title = "";

    $line = $$lineRef;
#    print "First line\n$line";
    while(1) {
	$tag = substr($line, 0, 2);
	$line = EMBL::TextFunc->cleanLine($line);
#	print "got line $tag, $line\n";
	if ($tag eq 'RN') {
#	    print "Got line $line\n";
	    $lineGroup->refNo($line =~ m/(\d+)/);
	    next;
	}
	if ($tag eq 'RC') {
	    $comment .= $line." ";
	    next;
	}
	if ($tag eq 'RP') {
	    $lineGroup->position($line);
	    next;
	}
	if ($tag eq 'RX') {
	    push @{$lineGroup->crossReferences()}, $line;
	    next;
	}
	if ($tag eq 'RA') {
	    push @{$lineGroup->authorList()}, EMBL::TextFunc->listFromText($line, ',\s*', ';');
	    next;
	}
	if ($tag eq 'RT') {
	    $title .= $line." ";
	    next;
	}
	if ($tag eq 'RL') {
	    push @{$lineGroup->locationLines()}, $line;
	    next;
	}
#       print "Matched nothing\n";
	last; # not a tag that I can deal with
    } continue {
	$line = <$fh>;
	if(not defined $line) { warn "Reference: End of stream"; last; }
    }

    # set the comment & title if apropreate
    $comment && $lineGroup->comment($comment);
    $title && $lineGroup->title($title);

    $$lineRef = $tag;
    $line && ($$lineRef.="   ".$line."\n");
    return $lineGroup;
}

sub toEMBLStream {
    my $self = shift;
    my $fh = shift;
    my $line;

    my $comment = $self->comment();
    my $position = $self->position();

    # RN Reference number - if not present print # inplace of the number
    print $fh "RN   [", $self->refNo() || '#', "]\n";

    # RC Comment
    if($comment) {
	foreach $line (EMBL::TextFunc->wrapText($comment, 75)) {
	    print $fh "RC   $line\n";
	}
    }

    # RP range in sequence
    if($position) {
	print $fh "RP   $position\n";
    }

    # RX cross references
    foreach $line (@{$self->crossReferences()}) {
	print $fh "RX   $line\n";
    }

    # RA authors
    # could not get the TextFunc->textFromList to work here, so I have done the wraping myself
    my $authorLine = join(", ", @{$self->authorList}) . ";";   # make one long string
    while (  $authorLine =~ m/(.{1,74}((,\s+)|$))/g ) {        # get each fragment within our size range
	$line = $1; if(length $line > 75) { $line = substr $line, 0, 75; }  # make sure that it is our size
	$line =~ s/\s+$//;                                     # remove any trailing spaces
	print $fh "RA   $line\n";                              # print it
    }

    # RT title
    foreach $line (EMBL::TextFunc->wrapText($self->title(), 75)) {
	print $fh "RT   $line\n";
    }

    # RL reference location
    foreach $line (@{$self->locationLines()}) {
	print $fh "RL   $line\n";
    }

    return $fh;
}

1;  # says use was ok
__END__

=head1 NAME

EMBL::Reference

=head1 DESCRIPTION

B<EMBL::Reference> represents the reference lines of an embl entry as specified in
and links from http://www.ebi.ac.uk/ebi_docs/embl_db/usrman/ref_line.html

EMBL::Reference inherits from B<EMBL::_LineGroup>

=head1 AUTHOR

B<Matthew Pocock> Email mrp@sanger.ac.uk

=head1 Member variables

=over

=item refNo

RN the reference number of this reference

=item comment

RC multiline comment - stoored as a single string

=item position

RP string representing the sequence ranges from this reference

=item crossReferences

RX cross reference e.g. to medline - literal text

=item authorList

RA list of authors - as an array of individual authors

=item title

RT multiline title - stoored as a single string

=item locationLines

RL list of all the reference location lines - stoored as an array of lines of text
as the RL lines are so variable that I could not be bothered to robustly deal with 
the many formats.

=back

=head1 Functions

=over

=item fromEMBLStream

Reads in all of the lines associated with a reference block, returning control when
it reaches a line it can not deal with (i.e. not RN, RC, RP, RX, RA, RT, RL).

    $rl = EMBL::Reference->fromEMBLStream(\*FH, \$line);

=item toEMBLStream

Prints the reference block out to a stream.  Should behave reasonably if some of the
lines are not present, but feel free to prove me wrong.

    $rl->toEMBLStream(\*STDOUT);

