
#
# Perl Module for EMBL::_LineGroup
#
# Cared for by Matthew Pocock <mrp@sanger.ac.uk>
#
# Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE

package EMBL::_LineGroup;

use vars qw($AUTOLOAD @ISA @EXPORT_OK %fields);

use Exporter;
use Carp;
use strict;
use EMBL::BaseClass;

#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = qw();

BEGIN {
    @ISA = ( 'Exporter', 'EMBL::BaseClass' );

    %fields = (
	       #Insert field names here as field => undef,
	       );
  EMBL::_LineGroup->check4Clashes();
}

sub initialize {}

sub new {
    my $ref = shift;
    my $class = ref($ref) || $ref;
    my $self = new EMBL::BaseClass;

    $self->rebless($class);
    return $self;
}

sub toEMBLStream {
    confess "You have called the virtula function EMBL::_LineGroup::toEMBLStream(@_)\n";
}

sub fromEMBLStream {
    confess "You have called the virtual function EMBL::_LineGroup::toEMBLStream(@_)\n";
}

1;  # says use was ok
__END__

=head1 NAME

EMBL::_LineGroup

=head1 DESCRIPTION

B<EMBL::_LineGroup> represents the base class for all logical groupings of lines.  If you
accidentaly forget to override the fromEMBLStream or toEMBLStream functions, the program
will fall over with a (fairly) helpfull error.  It found several of my buggs that way.

=head1 AUTHOR

B<Matthew Pocock> Email mrp@sanger.ac.uk

=over

=item toEMBLStream

Writes the _LineGroup to a stream.  The only argument should be a reference
to a glob of a file handle.  It is a virtual function.  Override it to write
your lines out.  The return falue will not currently be used, although I tend to return
the ref to the glob of a stream.

=fromEMBLStream

These functions are assumed to take two arguments - a reference to a filehandle glob,
and a reference to a line of text.  They are expected to return a new EMBL::_LineGroup
derived object, representing the lines.  The toEMBLStream function of the object should be
able to reproduce the lines.

If they are package functions, (as all of mine are) then they would look like this:
    package myLineGroup;

    sub fromEMBLStream {
        my $class = shift;
	my $fh = shift;
	my $lineRef = shift;
	my $lineGroup = new myLineGroup;
	my $line;
	
	$line = $$lineRef;
	while(can deal with $line) {
#
#	    do things with $line, and add that information to $lineGroup;
#
	    $line = <$fh>;
	}

	$$lineRef = $line;
	return $lineGroup;
    }

And you would register it as

    $parserInstance->register($tagOfInterest, 'EMBL::_LineGroupDerivedObject');

If you were rash enough to inherit from this class, you would have a stub that looked like this:

    package myDerived;

 sub fromEMBLStream {
    $self = myLineGroup::fromEMBLStream(@_);
    $self->rebless(myLineGroup);

    # any other stuff you want to do with the new line group

    return $self;
}

Alternatively, if you do not wish to use a package function then they would look like this:

    package LG2;

    sub fromEMBLStream {
	my $fh = shift;
	my $lineRef = shift;
	my $self = new LG2;
	my $line;

	$line = $$lineRef;
	while(can deal with $line) {
#
#	    do things with $line, and add that information to $lineGroup;
#
	    $line = <$fh>;
	}

	$$lineRef = $line;
	return $lineGroup;
    }

It would be registered with the parser like this:

    $parserInstance->register($tagOfInterest, \&LG2::fromEMBLStream);

And to produce a derived line group object using this fromEMBLStream function
you would have to do something like this

    package LG2Derived;

    sub fromEMBLStream {
	my $fh = shift;
	my $lineRef = shift;
	my $self = LG2::fromEMBLStream $fh, $lineRef;

	bless $self, $LG2Derived;
	$self->initializeTheObjectToMakeItLookLikeA_LG2Derived();
	return $self;
    }

=item toEMBLStream

Override this to produce class specific behavure for printing the linegroup to
a stream.  It will be invoked as $object->toEMBLStream(\*FH), and will be
expected to print all new lines associated with the lines entered.
