
#
# Perl Module for Embl
#
# Cared for by Matthew Pocock <mrp@sanger.ac.uk>
#
# Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE

package Embl;

use vars qw($AUTOLOAD @ISA @EXPORT_OK);
use Exporter;
use Carp;
use strict;

# get the EMBL::* guts

use EMBL::Parser;
use EMBL::Set;
use EMBL::Entry;
use EMBL::_LineGroup;

# get the EMBL::* line types
use EMBL::Image;
use EMBL::Discard;
use EMBL::ID;
use EMBL::XX;
use EMBL::NI;
use EMBL::SV;
use EMBL::DT;
use EMBL::DE;
use EMBL::AC;
use EMBL::KW;
use EMBL::Organism;
use EMBL::Reference;
use EMBL::DR;
use EMBL::FeatureTable;
use EMBL::CC;
use Sequence;

#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = qw();

#
# Global variables (ouch)
#

use vars qw ( $_defaultParser );

sub defaultParser {
    my $class = shift;
    my $parser = new EMBL::Parser;

    $parser->registerDefault( \&defaultErrorProc );
    $parser->register( 'ID', 'EMBL::ID' );
    $parser->register( 'AC', 'EMBL::AC' );
    $parser->register( 'KW', 'EMBL::KW' );
    $parser->register( 'XX', 'EMBL::XX' );
    $parser->register( 'NI', 'EMBL::NI' );
    $parser->register( 'SV', 'EMBL::SV' );
    $parser->register( 'DT', 'EMBL::DT' );
    $parser->register( 'DE', 'EMBL::DE' );
    $parser->register( 'OS', 'EMBL::Organism' );
    $parser->register( 'RN', 'EMBL::Reference' );
    $parser->register( 'RC', 'EMBL::Reference' );
    $parser->register( 'RP', 'EMBL::Reference' );
    $parser->register( 'RX', 'EMBL::Reference' );
    $parser->register( 'RA', 'EMBL::Reference' );
    $parser->register( 'RL', 'EMBL::Reference' );
    $parser->register( 'RT', 'EMBL::Reference' );
    $parser->register( 'DR', 'EMBL::DR' );
    $parser->register( 'FH', 'EMBL::FeatureTable' );
    $parser->register( 'FT', 'EMBL::FeatureTable' );
    $parser->register( 'SQ', 'Sequence' );
    $parser->register( '  ', 'Sequence' );
    $parser->register( 'CC', 'EMBL::CC' );
    
    return $parser;
}

sub Read {
    my $class = shift;
    my $fh = shift;

    return $_defaultParser->setFromStream( $fh );
}

sub entryFromStream {
    my( $class, $fh ) = @_;
    
    return $_defaultParser->nextEntryFromStream( $fh );
}

sub defaultErrorProc {
    my( $fh, $rLine ) = @_;
    chomp( my $line = $$rLine );
    warn "Error: Default line handler EMBL::Image will be called with '$line'\n",
         "       as nothing is registered to deal with this line type.\n";
    return EMBL::Image->fromEMBLStream(@_);
}

#
# @ISA has our inheritance.
#

BEGIN {
    @ISA = ( 'Exporter' );
    $_defaultParser = Embl->defaultParser();
}

1;  # says use was ok
__END__

=head1 Embl module documentation

=head2 NAME

Embl

=head2 DESCRIPTION

This module acts as a simple interface to the EMBL parsing system. 
For most uses you will be happy simply to use the functions detailed
in this documentation.

For more advanced users, basically when you want to change the
behaviour of the parser, you'll need to dig around in the EMBL
heirarchy. EMBL::Parser is a good place to start if you are
interested. If you do want to modify the parser behaviour it is
probably worth talking to Matt (mrp@sanger.ac.uk) or Ewan
(birney@sanger.ac.uk).

=head2 AUTHOR

B<Matthew Pocock> Email mrp@sanger.ac.uk

=head2 A simple example to get you started.

 # ** include this ***
 use Embl;


 # read in some embl files from a stream, and write them to stdout.
 #
 open(FH, "emblFile") || die "oops - could not open emblFile"; # open the stream
 $set = Read Embl(\*FH);          # read all of the embl entries from that stream.


 # set is an EMBL:Set object.
 $set->toStream(\*STDOUT);    # print the entries all out to stdout.


 # read in an embl entry from efetch using the default parser
 #
 $myParser = Embl->defaultParser();
 open(EFETCH, "efetch M31544 |") || die "Could not open the file";
 $efetch = $myParser->setFromStream(\*EFETCH);
 close EFETCH;


 # now add this to the entries read from file handle earlier
 $set->addSet($efetch);


 # could print out again with $set->toEMBLStream


 # get an array of all of the EMBL::Entry objects within $set,
 # and print each one out to a different file
 @entries = $set->entries();
 while ($entry = pop @entries) {
    if( ! defined ($id = $entry->ID()) ) {
	warn("Got no id for this entry. Skipping...\n");
    }
    open(OUT, ">$id.embl") || die "could not open this entry file";
    $entry->toStream(\*OUT);
    close OUT;
 }	

=head2 Functions

Not many functions are defined in this module, as basically they are provided
in the EMBL::Set and EMBL::Entry modules.

=over

=item Read

Takes a reference to a glob of a filehandle.  Returns an B<EMBL::Set> representing the
entries read from that stream.  It uses a parser constructed by B<defaultParser>.

    $set = Read Embl(\*STDIN);    # quite neat syntax - reads like English.

or

    $set = Embl->Read(\*STDIN);   # the other way to write the same thing.

=item defaultParser

Returns a parser object with the default associations between tags and line types.
Use this whenever you need a parser, unless you specificaly want to construct your own.

    $parser = Embl->defaultParser();

If this parser reaches a line that it can not deal with, it will print a warning message and stoor the line as an EMBL::Image.

=back

=head2 Variables

=over

=item $_defaultParser

This is a private package variable.
Forget about it.
It is here for eficiency, not beauty.

=back

=head1 How to use the core classes

=for html
This is a set of links to the modules that you as a user will be manipulating.
If you find any sections which are marked by the magic symbol
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif">
then you can safely ignore that part. <p>

=over

=item EMBL::Set

Entries can always be produced in sets.  The set may represent all of the entries
that were hit in a blast search, a set of overlapping genomic fragments, or just
some files that you want to work with.  However, as a stream can always contain
multiple EMBL entries seperated by //, the functions that read streams always
return a set of entries. This set will usualy only contain one entry. For example,
parsing an efetch query will produce one EMBL::Entry object representing the
result.  The result will always be delivered to you inside an EMBL::Set.

=over

=item entries

There is usualy just one thing that you want to do to a set of entries - get at
them all.

C<entries> returns an array of entries contained within the set.  You will often
wish to loop over each member of a set and perform some action on them.  For
example,

 print "Set $set contains entries for accession numbers\n";
 foreach $entry ($set->entries) {
    print "\t", $entry->AC(), "\n";
 }

=item addSet

Sometimes, you will have gathered entries from multiple queries.  How do you make
a single set out of them?  Easy - use the C<addSet> function to add one set into
another.

 # get our first set from a file
 open(FH, "myEmblDatabase") || die "Could not open 'myEmblDatabase'";
 $set = Read Embl(\*FH);
 close FH;
 # get another entry from an efetch
 open(EFETCH, "efetch M31544 |") || die "Could not open the file";
 $newSet = Read EMBL(\*EFETCH);
 close EFETCH;
 # now add the efetch result to $set
 $set->addSet($efetch);

=back

=for html
More extencive documentation for
<a href = "EMBL/Set.html">EMBL::Set.pm</a> is available.<p>

=item EMBL::Entry

You will definitely meet EMBL::Entry obejcts.  They represent a single EMBL entry
complete with simple commands to query entries for basic information like their
Accession number, or sequence.  The following is a list of the four functions
you will probably use most frequently.  They are invoked as usual by

    $entry->func();

=over

=item ID

Returns the id of this entry - the first value on the ID line.
It is equivelent to calling entryname on the EMBL::ID line object. 

=item AC

Returns the primary accession number for this embl entry.

=item FT

Returns the EMBL::FeatureTable obect that represents the feature table for this entry.
Edits to this table will alter the entry.

=item SQ

Returns the Sequence obejct that represents the sequence associated with this
entry. Edits to the sequence will alter the entry.

=back

=for html
More extencive documentation for
<a href = "EMBL/Entry.html">EMBL::Entry.pm</a> is available.<p>

=item EMBL::FeatureTable

You will probably be interested in features within the feature table of an embl
entry. The features are held within a FeatrueTable object, which knows how to
read and write its self, and how to arrange its features.

=over

=item features

This is the function that someone inspecting a feature table is going to use most
frequently. It returns an array of the features contained within the feature
table, in much the same way as C<entries> returns an array of the entries within
an C<EMBL::Set>.

 # get the feature table from $entry
 $ft = $entry->FT();
 # loop over each feature
 foreach $feature ($ft->features()) {
    # do something;
 }

=back

=for html
More extencive documentation for
<a href = "FeatureTable.html">FeatureTable.pm</a>
is available.<p>

=item EMBL::Feature

These are the real point of EMBL entries.  They contain anotation of the sequence
held below them.  You can access all of the available information within a
feature using the member functions listed below.

=over

=item key

The feature key.  It will be something like C<CDS> or C<misc_signal>.

=item location

The location string.  This is returned to you as text.  It is up to you what you
do with it.

=item qualifiers

This returns a reference to an array of the qualifiers associated with this
feature.  There I<should> be a cleaner API to this information.  I will make it
soon, but in the mean time, humour me.

=item range

This is a usefull function that returns a C<Range> object which is a bounding
range for the feature.  If the feature is composed of multiple ranges, C<range>
will contain all of the regions.

=item getRangeList

This is a little more advanced.  It is intended for use on CDS lines.  It returns
an array of C<Range> objects which make up the join statement.  Thus, if the
location was join(32..65,160..231) then getRangeList would return an array
containing a range 32 to 665 and a range 160 to 231.

There are problems associated with ranges that are outside the current feature -
either referrenced by another feature, or via an accession.  If these things
are found within the location of a feature, a warning will be printed, and
C<undef> returned.

=back

Here is an example of the sort of thing you can do.

 foreach $feature ($ft->features()) {
    print "Feature ", $feature->key(), " has bounding range ", $feature->range->toText(), "\n";
 }

=for html
More extencive documentation for
<a href = "Feature.html">Feature.pm</a>
is available.<p>

=back

=head1 I<Development> and I<advanced user> information.

This section includes pointers to all of the information that you will need to
extend the EMBL module capabilities, and extract every last piece of information
from an embl file. Each set of links indicate which category they fall into.

=head2 Base Classes for developers of other database parsers

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif"><img src = "http://www.sanger.ac.uk/~mrp/magic.gif"><P>
This is super deep stuff for those of you who wish to write a complete set of modules to deal with a different database. The base classes presume that the records are seperated by //\n and have two character line type identifiers. Other than that, they require very little. There is complete documentation available at the following locations:<P>
<a href = "SPEM/Spem.html">SPEM::Spem.pm</a> is an overview of the whole thing.<p>
<a href = "SPEM/Parser.html">SPEM::Parser.pm</a> tells you how to construct a parser object that knows what database it is parsing<p>
<a href = "SPEM/Parser.html">SPEM::Set.pm</a> makes it easy to make your own set of entries object<p>
<a href = "SPEM/Parser.html">SPEM::Entry.pm</a> provides all of the basic functionality of a single database entry<p>

=head2 Core modules for developers and advanced users

This is information which will be mandatory for developers who create new line
group types, or wish to extend the old ones.  It also explains how to alter the
default parser's behavure.

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif">
<a href = "EMBL/Parser.html">EMBL::Parser.pm</a>
describes the parser, how to modify the default one, and how to roll your own.
Required reading if you wish to change which classes are interested in given
EMBL lines.  It also explains how to write your own custom line types, and
extend already writen ones.<p>

=head2 Embl Line Types for developers and advanced users

And then there are all of the current line groups.  Each one lists it's data
members, and it's functions. They will either give you ideas, or provide you
with interesting base classes! Feel free to use them as templates when you write
your own.

The advanced user will be able to use the member variables of each line type to
find out extra information about entries which is not available via the standard
Embl interface.

=over

=item Line types within the EMBL hierachy

=for html
<a href = "EMBL/AC.html">EMBL::AC.pm</a> Accession line<p>
<a href = "EMBL/KW.html">EMBL::KW.pm</a> Represent the keyword lines<p>
<a href = "EMBL/NI.html">EMBL::NI.pm</a> Nucleotide identifier line<p>
<a href = "EMBL/CC.html">EMBL::CC.pm</a> A comment block<p>
<a href = "EMBL/XX.html">EMBL::XX.pm</a> An XX spacer<p>
<a href = "EMBL/DE.html">EMBL::DE.pm</a> Description<p>
<a href = "EMBL/DR.html">EMBL::DR.pm</a> Database cross reference<p>
<a href = "EMBL/ID.html">EMBL::ID.pm</a> The ID line<p>
<a href = "EMBL/DT.html">EMBL::DT.pm</a> The date line pair<p>
<a href = "EMBL/Reference.html">EMBL::Reference.pm</a>
A complete reference block<p>
<a href = "EMBL/Organism.html">EMBL::Organism.pm</a>
The organism information lines<p>

=item Line types outside of the EMBL hieracy - foreighn objects

=for html
<a href = "Sequence.html">Sequence.pm</a> The Sequence associated with an entry<p>
<a href = "FeatureTable.html">Feature.pm</a> The feature table<p>

=back

=head2 Special line types for developers and advanced users

Advanced users may use these for modifying the default parser - for example,
you may wish to discard on parsing everything that is not sequence or feature
table.  Developers may wish to use these as base classes, or templates.

=over

=item

=for html
<a href = "EMBL/Discard.html">EMBL::Discard.pm</a>
Discards one line of the input stream<p>
<a href = "EMBL/Image.html">EMBL::Image.pm</a>
A literal image of the text of a line<p>

=back

=head2 Utility modules for developers

These modules provide basic functionality for you to use.

=over

=item

=for html
<a href = "EMBL/BaseClass.html">EMBL::BaseClass.pm</a>
Contains functionality for all of the classes<p>
<a href = "EMBL/TextFunc.html">EMBL::TextFunc.pm</a> Common text functions<p>

=back
