=pod

=head1 NAME - GFF::Graph.pm

Perl graph drawing library for 'Gene Finding Format' (GFF) analysis routines
using the GFF.pm and Curve_plot.pm) Perl Object libraries

=head1 SYNOPSIS

use GFF::Graph qw(New_Plot Set Range_Score_Plot, Scatter_Score_Plot Print_Plot);  # include what functions you need... contains an implicit 'use GFF ;'

=head1 DESCRIPTION

GFF::Analysis (derived from GFF) is a utility
library for for Gene Finding Format, built upon
the GFF.pm module library.

Exports: 
    New_Plot()
    Set()
    Range_Score_Plot()
    Scatter_Score_Plot()
    Print_Plot()

The module methods use Raphael Leplae's 'Curve_plot.pm' module
to draw plots of GFF data (scores).

=head1 AUTHORSHIP

Copyright (c) 1999 
Created by Richard Bruskiewich <rbsk@sanger.ac.uk>

Curve_plot.pm is a graph drawing module by Raphael Leplae,
which uses the GD.pm graphics modules (from Tim Boutell's GIF package?).

Sanger Centre, Wellcome Trust Genome Campus, Cambs, UK
All rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation

=head1 SOURCE CODE

The most current release of the Perl source code 
for this module is available through a link at 
http://www.sanger.ac.uk/Software/GFF/
All bug reports may be submitted to Richard 
Bruskiewich (rbsk@sanger.ac.uk). Future releases may
be issued as Bioperl archived modules.

=cut

package GFF::Graph ;

use Carp;
use strict;
use vars qw($AUTOLOAD @ISA @EXPORT @EXPORT_OK $VERSION);
require Exporter;
use GFF ;
use Curve_plot ;

$VERSION = '2.06' ; # Based upon Version 2.0 GFF
#
# @ISA has our inheritance.
#
@ISA = qw( Exporter GFF );
#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#
@EXPORT    = qw();
@EXPORT_OK = qw(New_Plot Set Range_Score_Plot Scatter_Score_Plot Print_Plot);
#
# @ISA has our inheritance.
#
@ISA = ( 'Exporter' );


=pod

=head1 METHODS

Parameters and data are passed as hash elements.

=head2 New_Plot

Utility to create a GIF canvas (GD image) handle for drawing.

=head3 Arguments:

=over 4

=item C<width>

Overrides default (800) pixel width of canvas.

=item C<height>

Overrides default (600) pixel height of canvas.

=item C<palette>

This optional argument can initialize the GD colour palette 
with user defined colours. The argument should be a reference 
to a hash, indexed on colour labels and pointing to array 
references of (R,G,B) byte intensity values. For example:

        my %ColourPalette = (    # [ R  , G  , B  ]
              'salmon'          => [0xFA,0x80,0x72],
              'tomato'          => [0xFF,0x63,0x47],
              'orange'          => [0xFF,0xA5,0x00],
        ) ; 

Inside New_Plot, the array references are replaced with 
the GD::colorAllocate() references, for use by the calling 
script. 

=item C<bgcolour>

Background colour,either as a 'string' label into a defined
C<palette> or as a reference to an array of RGB colour values
(default: white => [255,255,255]). If the optional 'transparent'
argument flag is defined, then bgcolour is also made transparent.

=back

=cut

sub New_Plot {
    my %args=(
	      width    => 800,
	      height   => 600,
	      bgcolour => [255,255,255],
	      @_,
	      ) ;

    my $width       = $args{'width'} ;
    my $height      = $args{'height'} ;
    my $palette     = $args{'palette'} ;
    my $bgcolour    = $args{'bgcolour'} ;
    my $transparent = $args{'transparent'} ; # $bgcolour made transparent when defined

    # Create GD object
    my $gd = new GD::Image($width, $height);

    &Set(image=>$gd,palette=>$palette) if defined($palette) ;

    if(ref($bgcolour) =~ /ARRAY/) {
	$bgcolour = $gd->colorAllocate($bgcolour->[0],$bgcolour->[1],$bgcolour->[2]) ;
    } elsif(defined($palette)) {
	$bgcolour = $palette->{"$bgcolour"} ;
    } else {
	carp "GFF::Graph::New_Plot() Warning: bgcolour not defined? using 'white'!\n" ;
	$bgcolour = $gd->colorAllocate(255,255,255) ;
    }
    $gd->transparent($bgcolour) if(defined($transparent)) ;
    $gd->interlaced('true');
    $gd->filledRectangle(0, 0, $width, $height, $bgcolour);

    return $gd ;
}

=pod

=head2 Set

This method modifies the global graph drawing environment,
(e.g. assigns palette colours to an image).

=head3 Arguments:

=over 4

=item C<image>

GD::Image reference. Required argument when a GD::Image 
characteristic is being (re)set, eg. the palette.

=item C<palette>

Modifies an existing palette. See New_Plot() above.

=back

=cut

sub Set {
    my %args = @_;

    foreach my $key (keys %args) {
	if($key eq 'palette') {
	    my $gd = $args{'image'} ;
	    croak "GFF::Graph::Set() - Need a GD::Image to set the 'palette' to?" if !defined($gd) ;
	    my $palette=$args{'palette'} ;
	    if(defined($palette)) {
		croak "### GFF::Graph::New_Plot()Error: 'palette' must be a reference to a hash, \n".
		      "### indexed on colour labels and pointing to array references of GD colour allocations!\n" 
			if (ref($palette) !~ /HASH/) ;
		foreach my $colour (keys %{$palette}) {
		    my $RGB = $palette->{"$colour"} ;

		    croak "### GFF::Graph::New_Plot()Error: 'palette' must point to array references of GD colour allocations!\n" 
			if (ref($RGB) !~ /ARRAY/) ;

		    $palette->{"$colour"} = $gd->colorAllocate($RGB->[0],$RGB->[1],$RGB->[2]);
		}
	    }
	}
    }
}



=pod

=head2 Range_Score_Plot

This procedure draws a GIF graph of a GeneFeatureSet scores against
a sequence range (position in kilobases).

=head3 Arguments:

=over 4

=item C<filename>

The GIF image may be plotted to the specified file $filename.

=item C<image>

Alternative. the graph may be written onto a pre-existing
GD::Image object passed as an $image argument.

=item C<series>

The series is a reference to a hash of references
to GFF::GeneFeatureSet objects, keyed on data series names.
These GFFF::GeneFeatureSet objects should generally consist
of uniform features. If one special GeneFeauture record marked
with <feature> = 'Total', then it is ignored in the data plot
itself, but its <score> field is assumed to contain the total
number of features in the GFF set, hence is used as the source
for this feature number (in parenthesis) in the legend (if specified).
This may be suppressed with the C<no_totals> argument.

=item C<font>

Sets the default GD font for text in the graph to one of 'tiny' (5 x 8), 
'small' (6 x 12), 'medium' (7 x 13) or 'giant' (9 x 15). Default: 'medium'.

=item C<width>

=item C<height>

=item C<x_min>

=item C<x_max>

=item C<x_step>

=item C<y_min>/C<y2_min>

=item C<y_max>/C<y2_max>

=item C<y_step>/C<y2_step>

=item C<y_title>/C<y2_title>

The image dimensions are $width x $height. The graph
may have two Y axes (Y2 is on the right hand side). All the
comments for Y apply to Y2 with arg label transformation. 

The Y axis $y_title can be given.

The bounds of values ($x_min-$x_max,$y_min-$ymax) and
major scale label increment ($x_step,$y_step) may be given
as arguments or determined by the content of the GeneFeatureSets,
if undefined. 

Note that the units of the X axis are kilobases. the $y_title
labels the Y axis units.

=item C<x_offset>

=item C<y_offset>

The upper left hand plot area origin is (0,0) by default, but may
be reset by  C<x_offset> and C<y_offset>, respectively.

=item C<title>

A $title can also be given to the graph.

=item C<legend>/C<legend2>

Set the legend location. Can be 'top', 'bottom', 'left' and 'right', or a pairwise 
combination of two compatible specifications, e.g. 'top,left', where the first specification
is taken as the primary location, to which the second specification is added as an 
alignment adjustment. To suppress the legend, set to undef/0. C<legend2> controls the 
legend for the second Y axis associated data series.

=item C<legend2>

Same as C<legend> except that it specifies the right hand Y axis specified data series.

=item C<colours>

The colour of plotted curves are normally automatically assigned, but
specific colour values for specific curves may be assigned by
providing a 'Red Green Blue' (RGB) 3-tuple of intensities, valued [0..255],
in arrays, given as reference value in a hash, indexed by $series keys, 
and itself provided as a reference as the C<colours> argument, e.g.

my %colours = (
	       'Series A' => [255,0,0], # red
	       'Series B' => [0,255,0], # green
	       'Series C' => [0,0,255], # blue
	       ) ;
.
.
.
Range_Score_Plot( colours => \%colours,...)

=item C<gaps>

GFF defining data 'gaps' may be passed as a set. The score field for
such GFF should be undef ('.') as a signal to clip line plots across
such gaps.


=item C<no_x_axis>

When defined, suppresses the x axis plotting.

=item C<no_totals>

When defined, suppresses printing of GFF 'Total' record 
information in the legend.

=back

=head3 Returns:

The method returns the graph relative coordinates for the
graph plot area: the top left hand origin (x,y) of the 
plot area plus the width (w) and height (h) of the plot area,
as the list '(x,y,w,h)'.

NOTE: The specification of 'vertical' mode does *NOT* 
change these coordinates. They remain as the would have
if the plot was plotted horizontally (even though,
in display terms, GD plotted the image mirrored about
the diagonal axis running through the aforementioned
(x,y) 'top left hand' point of origin.

=cut

sub my_xformat { my $val = shift; return sprintf("%d",$val); } 
sub my_yformat { my $val = shift; return sprintf("%7.3f",$val); } 

my $total_filter = sub {
       my $self = shift ;
       if($self->feature() eq 'Total') {
	   1 ;
       } else {
	   0 ;
       }
} ;

sub y_step { 
   my $max = shift ; 
   return 
     (($max <= 100)? 
       (($max <= 50)? 
         (($max <= 20)? 
	   (($max <= 10)?
	      (($max <= 5)? 
	         (($max <= 2)?
	            (($max <= 1)?
		       (($max <= 0.5)? 
		          (($max <= 0.2)?
			     (($max <= 0.1)? 0.01 
			       : 0.02 ) 
			   : 0.05 ) 
		       : 0.1 ) 
		    : 0.2 ) 
		 : 0.5 )
	       : 1)
	    : 2)
	  : 5)
       : 10) 
    : 20) ; 
}

sub Range_Score_Plot {

    my %args=(
	      x_step   => 100,
	      y_factor => 1,
	      font => 'medium',
	      @_,
	      ) ;

    my $filename = $args{'filename'}; # GIF file
    my $image    = $args{'image'}; # predefined GD::Image() file
    my $transparent = $args{'transparent'}; # make background transparent

    # this graph's plot origin in predefined GD::Image() file
    my $x_origin  = $args{'x_origin'};
    my $y_origin  = $args{'y_origin'};

    # The data is a reference to a hash of references to
    # GFF::GeneFeatureSet objects keyed on data series name
    # If a special 'Total' <source> field object is seen, 
    # then these are filtered out and their score data taken
    # as the total number of items in the series...
    my $series     = $args{'series'} ;    

    my $width      = $args{'width'} ;    # minimum width (x_axis may extend)
    my $x_offset   = $args{'x_offset'} ; # in pixels
    my $height     = $args{'height'} ;
    my $y_offset   = $args{'y_offset'} ; # in pixels

    my $font       = $args{'font'} ;
    my $vertical   = $args{'vertical'} ;

    my $x_min      = $args{'x_min'} ;  # in kilobases
    my $x_max      = $args{'x_max'} ;  # in kilobases
    my $x_step     = $args{'x_step'} ; # in kilobases
    my $y_min      = $args{'y_min'} ;
    my $y_max      = $args{'y_max'} ;
    my $y_step     = $args{'y_step'} ;
    my $y_factor   = $args{'y_factor'} ; # multiplicative normalization factor for Y score value
    my $y_title    = $args{'y_title'} ;

    # master switch for second (right hand) axis
    # if defined, should be a reference to a hash 
    # of $series indices to be placed on the second y axis  
    my $y2_axis    = $args{'y2_axis'} ; 
    $y2_axis = 0 if !defined($y2_axis) ; 

    my $y2_min     = $args{'y2_min'} ;
    my $y2_max     = $args{'y2_max'} ;
    my $y2_step    = $args{'y2_step'} ;
    my $y2_factor  = $args{'y2_factor'} ; # multiplicative normalization factor for Y score value
    my $y2_title   = $args{'y2_title'} ;

    my $title      = $args{'title'} ;
    my $legend     = $args{'legend'} ;
    my $legend2    = $args{'legend2'} ;
    my $no_x_axis  = $args{'no_x_axis'} ;
    my $gapgff     = $args{'gaps'};
    my $no_totals  = $args{'no_totals'};

    my $colours    = $args{'colours'} ;
    $colours = 0 if !defined($colours) ;
    croak "GFF::GeneFeatureSet::Range_Score_Plot(): 'colours' argument must be a reference to a hash!" 
	if($colours && ref($colours) !~ /HASH/) ;

    my $debug      = $args{'debug'} ; # debug flag

    if( defined($debug) and $debug ) {
	print "<P>GFF::Graph::Range_Score_Plot(), arguments:</P><UL>"  ;
	foreach my $arg (sort keys %args) {
	    print "<LI>$arg => $args{$arg}</LI>\n"  ;
	    
	}
	print "</UL>\n"  ;
    }
 
    #print '<pre>' ;
    #print "Args:\n",join "\n",%args ;
    #print '</pre>' ;

    if(!( ((defined($filename) and $filename) or 
	   (defined($image) and $image and ref($image))) 
	                    and
	  (defined($series) and $series and ref($series) =~ /HASH/) and
          (defined($width)  and $width  and $width  =~ /^(\d+)$/ ) and
          (defined($height) and $height and $height =~ /^(\d+)$/ )
         )
        ) {
       croak "GFF::GeneFeatureSet::Range_Score_Plot((filename: $filename|image: $image), series: $series, ref(\$series) is HASH? ",
             ref($series) =~ /HASH/,",width: $width, height: $height):".
	     " needs at least a GIF filename (or image handle), dataset, width and height!"
    }

    $x_min = 100000000 if !(defined($x_min)) ;
    $x_max = 1         if !(defined($x_max)) ;
    $y_min = 100000000 if !(defined($y_min)) ;
    $y_max = 0         if !(defined($y_max) and $y_max) ;
    $y_factor = 1   if !(defined $y_factor and $y_factor) ;
    $y_title  = "Gene Features" if !(defined $y_title and $y_title) ;

    if($y2_axis) {
	$y2_min = 100000000 if !(defined($y2_min)) ;
	$y2_max = 0         if !(defined($y2_max) and $y2_max) ;
	$y2_factor = 1      if !(defined $y2_factor and $y2_factor) ;
	$y2_title  = "Gene Features" if !(defined $y2_title and $y2_title) ;
    }

    my $graph = new Curve_plot ( $width, $height, $image, $transparent );

    $x_step   = 100 if !(defined $x_step and $x_step) ; # default to 100 kilobases

    $graph->x_offset ($x_offset) if defined($x_offset) ;
    $graph->y_offset ($y_offset) if defined($y_offset) ;

    $graph->set (y_label_format  => \&my_yformat,
		 x_label_format  => \&my_xformat,
		 #auto_colour     => 1,
		 y_grid          => '192,192,192',
		 );    
    $graph->set (legend => $legend)  if defined($legend) ;
    # define (h,k) origin first...
    $graph->set ('x'=> $x_origin) if defined($x_origin) ;
    $graph->set ('y'=> $y_origin) if defined($y_origin) ;
    # before setting 'vertical' plotting mode
    $graph->set ('vertical' => 1) if defined($vertical) ;
    $graph->set ('font' => $font) ;
    $graph->set(plot_xlabels => 0,
		plot_xticks  => 0,
		plot_xbase   => 0)   if defined($no_x_axis) && $no_x_axis ;

    $graph->set(y_label_format  => \&my_yformat,
                plot_x2labels   => 0,
		plot_x2ticks    => 0,
		plot_x2base     => 0) if($y2_axis) ;
    $graph->set (legend2 => $legend2)  if($y2_axis) && defined($legend2) ;

    $graph->title ($title)        if(defined $title and $title);
    $graph->y_title ($y_title)    if defined($y_title);
    $graph->x_title ("kilobases") if !(defined($no_x_axis) && $no_x_axis)  ;

    $graph->y2_title ($y2_title)  if defined($y2_title);

    my $colour_filter = sub {
	my $gf = shift ;
	return
    } ;
    print '<P><U>Range plot data points</U>:</P>'  if(defined($debug) and $debug) ;
    my $curve_nb1 = -1 ; my $curve_nb2 = -1 ;
    foreach my $dataset (keys %{$series}) {
        my $gffs   = $series->{$dataset} ;
	next if(!$gffs->count()) ; # ignore empty datasets
        my $gfft   = $gffs->filter($total_filter) ;
	my $total  = $gfft->sumScore(0,1) ; # unweighted sum of subtotals
        my $gffdp  = $gffs->exclude($total_filter) ; # datapoints

	# add gaps to dataset if defined (these have no scores, so will be passed to
	# graph as undef).
	$gffdp->addGFF($gapgff) if $gapgff;
	
	my $dataset2 = ($y2_axis && exists($y2_axis->{$dataset})) ;
        my (@x,@y,@x2,@y2) ;
        foreach my $gf ($gffdp->order_gf()) {
            # value of attribute
	    my $score = $gf->score() ;
	    printf "<P>(%5.1f,%7.3f)</P>\n",((($gf->start+$gf->end-1)/2)/1000), $score  if(defined($debug) and $debug>2) ;

	    if($dataset2) {
		# weighted centre of the gene feature, in kilobases
		unshift(@x2,(($gf->start+$gf->end-1)/2)/1000) ;

		if(defined($score) and $score ne '.') {
		    $score *= $y2_factor ;
		} else {
		    $score = undef ; # signal to Curve_plot to skip gap?
		}
		unshift(@y2,$score) ; # may be undef'd for gaps in the plot?

		$y2_min = $score if(defined($score) and $score < $y2_min) ;
		$y2_max = $score if(defined($score) and $score > $y2_max) ;

	    } else {
		# weighted centre of the gene feature, in kilobases
		unshift(@x,(($gf->start+$gf->end-1)/2)/1000) ;

		if(defined($score) and $score ne '.') {
		    $score *= $y_factor ;
		} else {
		    $score = undef ; # signal to Curve_plot to skip gap?
		}
		unshift(@y,$score) ; # may be undef'd for gaps in the plot?


		$y_min = $score if(defined($score) and $score < $y_min) ;
		$y_max = $score if(defined($score) and $score > $y_max) ;
	    }
        }
	my $label = $dataset ; 
        $label .= ($total ? " ($total)" : '' ) unless($no_totals) ;
	if($dataset2) {
	    $graph->add_dataset2 ($label, \@x2, \@y2);
	    $graph->add_legend2($label) if defined($legend2) ;
	} else {
	    $graph->add_dataset ($label, \@x, \@y);
	    $graph->add_legend($label) if defined($legend) ;
	}

	if($dataset2) {
	    $curve_nb2++  ;
	} else {
	    $curve_nb1++ ;
	}

	if($gapgff){
	    if($dataset2) {
		$graph->dataset_define2 ($curve_nb2, 'ignore_gaps' => 0 ) ;
	    } else {
		$graph->dataset_define ($curve_nb1, 'ignore_gaps' => 0 ) ;
	    }
	}
	if( $colours && exists($colours->{$dataset}) ) {
	    my $RGB = $colours->{$dataset} ;
	    croak "GFF::GeneFeatureSet::Range_Score_Plot(): 'colours' entry is not an array ref?"
		if( ref($RGB) !~ /ARRAY/ ) ;
	    if($dataset2) {
		$graph->dataset_define2 ($curve_nb2, 'colour' => $RGB ) ;
	    } else {
		$graph->dataset_define ($curve_nb1, 'colour' => $RGB ) ;
	    }
        }
	if($dataset2) {
	    $x_min = ($x2[0]   <  $x_min ? $x2[0]   : $x_min) ;
	    $x_max = ($x2[$#x] >  $x_max ? $x2[$#x] : $x_max) ;
	} else {
	    $x_min = ($x[0]   <  $x_min ? $x[0]   : $x_min) ;
	    $x_max = ($x[$#x] >  $x_max ? $x[$#x] : $x_max) ;
	}

    }
    $x_min = ($x_max <= $x_min ? $x_max-1 : $x_min) ;
    $y_min = ($y_max <  $y_min ? $y_max-1 : $y_min) ;
    if($y2_axis) { 
	$y2_min = ($y2_max <  $y2_min ? $y2_max-1 : $y2_min) ;
    }

    my $value ;
    my @xaxis ; 
    $value = int($x_min/$x_step) * $x_step ; # ignore the fraction!
    while($value <= $x_max) { # minimal major X ticks
	unshift(@xaxis,$value) ;
	$value += $x_step ;
    }
    $graph->x_axis  (\@xaxis);
    $graph->x2_axis (\@xaxis) if($y2_axis);

    if( defined($y_step) and $y_step ) {
	$y_step *= $y_factor ;
    } else {
	$y_step = y_step($y_max-$y_min)  ;
    }

    my(@yaxis) ;
    $value = int($y_min/$y_step) * $y_step ;
    while($value <= ($y_max + $y_step)) { # add extra major Y tick
        unshift(@yaxis,$value) ;
	$value += $y_step ;
    }
    $graph->y_axis (\@yaxis);

    if($y2_axis) {
	if( defined($y2_step) and $y2_step ) {
	    $y2_step *= $y2_factor ;
	} else {
	    $y2_step = y_step($y2_max-$y2_min)  ;
	}

	my(@y2axis) ;
	$value = int($y2_min/$y2_step) * $y2_step ;
	while($value <= ($y2_max + $y2_step)) { # add extra major Y tick
	    unshift(@y2axis,$value) ;
	    $value += $y2_step ;
	}
	$graph->y2_axis (\@y2axis) ;
    }

    $graph->plot ();

    $graph->save_gif ("$filename.gif") if defined($filename) ;

    # Get pos of plot area
    my $PAx = $graph->get('x', 'plot_area');
    my $PAy = $graph->get('y', 'plot_area');
    my $PAw = $graph->get('w', 'plot_area');
    my $PAh = $graph->get('h', 'plot_area');
    return( $PAx, $PAy, $PAw, $PAh ) ;
}
=pod

=over 4

=item Scatter_Score_Plot( $filename, 
			  $gffX, $title_X, \
			  $gffY, $title_Y, \
			  $width, $height, \
                          $x_min, $x_max, $x_step, $x_factor, \
                          $y_min, $y_max, $y_step, $y_factor, \
                          $title )

This method plots a XY scatter plot of scores of two sets, 
$gffX and $gffY,  of GFF records, scores aligned by aligned 
interval overlap, returning (in order) r, Mx, Sx, My, Sy and Sxy 
as a list of values, where:

    r   is the correlation coefficient,
    Mx  is the mean value of X,
    Sx  is the standard deviation of X, 
    My  is the mean value of Y,
    Sy  is the standard deviation of Y and 
    Sxy is the covariance of X and Y.

The method assumes an exact pairwise geometric range overlap 
between the members of each dataset (i.e. that analysis windows
for the scores match up one-to-one) and that eachset to be 
compared have homogenous records with uniform source/feature semantics.

The bounds of values ($x_min-$x_max,$y_min-$ymax) and
major scale label increment ($x_step,$y_step) may be given
as arguments or determined by the content of the GeneFeatureSets,
if undefined. 

Note that the $title_X and $title_Y label the X and Y axis units, 
respectively, while the corresponding $x_factor and $y_factor
may be used as a scaling factor on the X and Y data point (scores) 
(e.g. window size in kilobases, if datapoints are given as score/kilobase).
These latter values default to 1.

A $title can also be given to the graph.

=back

=cut

my $axis_title = sub {
    my $self = shift ;
    return (($self->feature()).($self->source())) ;
};

my $LC_Threshhold = 0.5 ;

sub Scatter_Score_Plot {
    my %args=(
	      x_factor => 1,
	      y_factor => 1,
	      font => 'medium',
	      @_,
	      ) ;

    my $filename = $args{'filename'}; # GIF file
    my $image    = $args{'image'};    # predefined GD::Image() file
    my $transparent = $args{'transparent'}; # make background transparent

    # this graph's plot origin in predefined GD::Image() file
    my $x_origin  = $args{'x_origin'};  # predefined GD::Image() file
    my $y_origin  = $args{'y_origin'};  # predefined GD::Image() file

    # The data are references to two GFF::GeneFeatureSets
    my $gffX     = $args{'gffX'} ; 
    my $titleX   = $args{'titleX'} ;

    my $gffY     = $args{'gffY'} ; 
    my $titleY   = $args{'titleY'} ;

    my $font     = $args{'font'} ;
    my $vertical = $args{'vertical'} ;

    my $width    = $args{'width'} ; 
    my $height   = $args{'height'} ;

    my $x_min    = $args{'x_min'} ; 
    my $x_max    = $args{'x_max'} ; 
    my $x_step   = $args{'x_step'} ;
    my $x_factor = $args{'x_factor'} ; 

    my $y_min    = $args{'y_min'} ;
    my $y_max    = $args{'y_max'} ;
    my $y_step   = $args{'y_step'} ;
    my $y_factor = $args{'x_factor'} ; 

    my $title    = $args{'title'} ; 
    my $legend   = $args{'legend'} ;

    my $debug    = $args{'debug'} ; # debug flag

    # print STDOUT "<P>GFF::Graph::Scatter_Score_Plot(Ymin:$y_min,Ymax:$y_max,Ystep:$y_step)</P>\n" ;
 
    if(!(((defined($filename) and $filename) or ((defined($image) and $image and ref($image)))) and
	 defined($gffX) and ref($gffX) and $gffX =~ /HASH/ and
	 defined($gffY) and ref($gffY) and $gffY =~ /HASH/ and
         defined($width)  and $width  and $width  =~ /^(\d+)$/ and
         defined($height) and $height and $height =~ /^(\d+)$/ )) {
       croak "GFF::GeneFeatureSet::Scatter_Score_Plot() needs at least a GIF filename (or image handle), 2 GFF datasets, width and height!"
    }

    $x_min    = 100000000 if !(defined($x_min)) ;
    $x_max    = 0         if !(defined($x_max)) ;

    $y_min    = 100000000 if !(defined($y_min)) ;
    $y_max    = 0         if !(defined($y_max) and $y_max) ;

    my $graph = new Curve_plot ($width, $height, $image, $transparent);

    $graph->set ('y_label_format'  => \&my_yformat,
		 'x_label_format'  => \&my_yformat, # make fine point format too
		 'x_grid'          => '192,192,192',
                 'y_grid'          => '192,192,192',
		 );
    $graph->set ('legend' => $legend)  if defined($legend) ;

    # define (h,k) origin first...
    $graph->set ('x'=> $x_origin) if defined($x_origin) ;
    $graph->set ('y'=> $y_origin) if defined($y_origin) ;
    # before setting 'vertical' plotting mode
    $graph->set ('vertical' => 1) if defined($vertical) ;
    $graph->set ('font' => $font) ;

    $graph->title ($title) if(defined $title and $title);

    $graph->x_title ($titleX);
    $graph->y_title ($titleY);


    my $gffXf = $gffX->exclude($total_filter) ; # datapoints only
    my $gffYf = $gffY->exclude($total_filter) ; # datapoints only

    if($GFF::TRACE > 1) {
	print "<P>Dump of Scatter Plot GFF inputs:<pre>\n" ;
	$gffXf->dump(\*STDOUT) ;
	$gffYf->dump(\*STDOUT) ;
	print "</pre>\n" ;
    }

    # First, order the gf data by size...
    my @gffX = $gffXf->order_gf() ;
    my @gffY = $gffYf->order_gf() ;

    my (@x, @y, @string, $gfX, $gfY) ;
    my ($N, $X, $Y, $X2, $Y2, $XY) ;
    $N = $X = $Y = $X2 = $Y2 = $XY = 0 ;
    print '<P><U>Scatter plot data points</U>:</P>' if(defined($debug) and $debug) ;
    DATA: foreach $gfX  (@gffX) {
	do {

	    $gfY = shift @gffY ;
	    last DATA if !defined($gfY) ;

	    if($GFF::TRACE > 1) {
		print '<P>Attempting match of:<P><pre>' ;
		$gfX->dump(\*STDOUT) ;
		print '</pre><br><B>to</B><br><pre>' ;
		$gfY->dump(\*STDOUT) ;
		print '</pre>' ;
	    }

	    @string = $gfX->match($gfY);

	    print '<P>Match string:',join (',',@string),'</P>'  if(defined($debug) and $debug) ;
	} until($string[0]) ;

	print '<P>until Matched!</P>' if($GFF::TRACE > 1) ;

        #
	# Assume a matched pair of gene feature ranges at this point
        # so tabulate their corresponding scores (with scaling factored in)
	#
	my $Xscore = $gfX->score() * $x_factor ;
	my $Yscore = $gfY->score() * $y_factor ;

	$x_min = $Xscore if $x_min > $Xscore ;
	$y_min = $Yscore if $y_min > $Yscore ;
	$x_max = $Xscore if $x_max < $Xscore ;
	$y_max = $Yscore if $y_max < $Yscore ;

	push(@x,$Xscore) ;
	push(@y,$Yscore) ;

	printf "<P>(%7.3f,%7.3f)</P>\n",$Xscore,$Yscore if(defined($debug) and $debug) ;

        # Compile some statistical values?
	$N++ ;
	$X  += $Xscore ;
	$Y  += $Yscore ;
	$X2 += $Xscore**2 ;
	$Y2 += $Yscore**2 ;
	$XY += $Xscore*$Yscore ;
    }

    confess "x min $x_min > x max $x_max" if $x_min > $x_max ;
    confess "y min $y_min > y max $y_max" if $y_min > $y_max ;

    return undef if !(scalar(@x) and scalar(@y)) ;

    # Calculate a few statistics on these values?
    my $Mx  = $X/$N ;
    my $My  = $Y/$N ;
    my $Sx  = ($X2/$N - $Mx**2)**0.5 ;
    my $Sy  = ($Y2/$N - $My**2)**0.5 ;
    my $Sxy = ($XY/$N - $Mx*$My) ;
    my $D   = $Sx*$Sy ;
    my $r   = ($D ? $Sxy/$D : 0) ;

    my $label = "$titleX versus $titleY" ;
    $graph->add_dataset ($label, \@x, \@y);
    $graph->dataset_define (0,  'line_style' => 'none', 'dot_style' => 'O' );
    $graph->add_legend($label) if defined($legend) ;

    # Draw a linear regression line if this is sensible?
    if($r > $LC_Threshhold) {

	my (@lx, @ly) ;

	$lx[0] = $x_min ;
	$lx[1] = $x_max ;

	my $m = (($r*$Sy)/$Sx) ;

	$ly[0] = $My + $m*($x_min-$Mx) ;
	$ly[1] = $My + $m*($x_max-$Mx) ;

	$label = "Linear Regression of $titleX on $titleY" ;
	$graph->add_dataset ($label, \@lx, \@ly);
	$graph->dataset_define (1,  'colour' => [192,0,0], 'line_style' => 'dotted' );
        $graph->add_legend($label) if defined($legend) ;
    }

    # Plot the axes

    my @xaxis ; 

    if( defined($x_step) and $x_step ) {
	$x_step *= $x_factor ;
    } else {
	$x_step = y_step($x_max-$x_min)  ;
    }

    my $value = int($x_min/$x_step) * $x_step ; # ignore the fraction!
    while($value <= ($x_max + $x_step)) {
        unshift(@xaxis,$value) ;
	$value += $x_step ;
    }
    $graph->x_axis (\@xaxis);

    my(@yaxis) ;

    if( defined($y_step) and $y_step ) {
	$y_step *= $y_factor ;
    } else {
	$y_step = y_step($y_max-$y_min)  ;
    }

    $value =  int($y_min/$y_step) * $y_step ; # ignore the fraction!
    while($value <= ($y_max + $y_step)) {
        unshift(@yaxis,$value) ;
	$value += $y_step ;
    }
    $graph->y_axis (\@yaxis);
    $graph->plot ();
    $graph->save_gif ("$filename.gif") if defined($filename) ;
    return ($r,$Mx,$Sx,$My,$Sy,$Sxy) ;
}

=head2 Print_Plot

Utility to print out a GD image canvas as a GIF.

=head3 Arguments:

=over 4

=item C<image>

Mandatory GD::Image handle (such as created by New_Plot()).

=item C<filename>

Optional filename to which to write the Gif image.

=item C<filehandle>

Optional file glob to which to write the Gif image.

=back

=cut

sub Print_Plot {
    my %args=(
	      filehandle => \*STDOUT,
	      @_,
	      ) ;

    my $image      = $args{'image'} ;
    if( !(defined($image) and $image and ref($image)) ) {
	carp "GFF::Graph::Print_Plot(): invalid image argument: $image?" ;
	return ;
    }
    my $filename   = $args{'filename'} ;
    my $filehandle = $args{'filehandle'} ;
    local (*GIF);
    if( defined($filename) and $filename ) {
	if(! open (GIF, "+>$filename") ) {
	    carp "GFF::Graph::Print_Plot() could not open Gif file $filename: $!!\n" ;
	    return ;
	}
	$filehandle = \*GIF ;
    }
    binmode $filehandle ;
    print $filehandle $image->gif ;
    close GIF if $filehandle == \*GIF;
}

1;  # says use was ok
__END__

=head1 Revision History

    2.07 (24/11/99)- rbsk: Set() method added; xorigin,yorigin renamed to x_origin,y_origin

2.06(4/11/99)  - rbsk: Y2 axis in Range_Score_Plot plus numerous bug fixes for CWA
               - handling Range_Score_Plot 'gaps' in plots, using C<gapgff>
               - C<no_x_axis> suppression of x axis

2.05(17/10/99) - rbsk: colour 'palette' argument added to New_Plot

2.04(28/9/99)  - rbsk: convert to hash argument passing paradigm added 
                       'image', 'xorigin' and 'yorigin' arguments to permit 
                       plotting of graphs into an existing GD canvas.
                       - added New_Plot() and Print_Plot().
                       - 'no_x_axis' argument for Range_Score_Plot()

2.03(8/9/99)   - rbsk:  created Scatter_Score_Plot().

2.02(3/9/99)   - rbsk: added concept of 'Total' <source> records for legend.
                   if such a total exists for a given series, then it is
                   returned as a second (comma delimited) number in the legend hash

2.01(16/7/99)  - rbsk: transferred Range_Score_Plot() (formerly draw_graph) from Analysis.pm
                   to here and generalized to multiple GFF plots (API changed)
