#!/usr/local/bin/perl

=head1 NAME

GetAce.pm - Perl module for information retrieval from ACEDB.

=head1 AUTHOR

Richard Bruskiewich - Email B<rbsk@sanger.ac.uk>

Copyright (c) 1999  Richard Bruskiewich (rbsk@sanger.ac.uk)
Sanger Centre, Wellcome Trust Genome Campus, Cambs, UK
All rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation.

=head1 SYNOPSIS

B<use GetAce;>

=cut

####################################################################

package GetAce ;

use Carp;
use strict;
use Sequence ;
use Cwd ;
use MyUtilLib ;
use GFF ;

use vars qw($AUTOLOAD @ISA @EXPORT @EXPORT_OK $VERSION $TRACE);

require Exporter;
#
# @ISA has our inheritance.
#
@ISA = qw( Exporter );
#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#
@EXPORT    = qw();
@EXPORT_OK = qw(trace_on trace_off dumpDNA dumpGFF dumpACE dumpGenes);

$VERSION = '2.0' ;

my $IsaGene = "Predicted_gene OR Method = GD_mRNA OR ".
              "Method = GD_CDS OR Method = GD_Not_Confirmed OR ".
              "Method = supported_* OR Pseudogene OR Method = Pseudogene" ;

# e.g. 130N4 or BRCA2 or dJ102G20.C1.1 or dJ102G20.1 or dJ102G20.CX.1 or dJ102G20.CX.1b (isoform)
my $NameFilter = "\\w*((\\.C(\\d+|X|Y))?\\.\\d+[a-z]?)?" ;

=pod

=head1 DESCRIPTION

GetAce is a Perl module of useful subroutines which use AcePerl to 
retrieve specific objects information from an ACEDB database.

All the routines take an AcePerl database ($db) handle as their
first argument, and generally, a sequence name ($seqname) as 
their second argument. For methods which create files, optional 
filename $path, $prefix and $suffix arguments can be provided.
These default to specific values if omitted: path to Cwd(), 
prefix to empty, and suffix to a routine specific file extension.

All the routines return 1 if successful, 0 if the operation fails.
Null $db or $seqname throws an exception (croak!).

=head1 Subroutines

Note: $filename arguments default to $seqname if unspecified.

=over 4

=item trace_on

Turn run tracing on.

=item trace_off

Turn run tracing off.

=back

=cut

my $TRACE = 0 ;

sub trace_on  { $TRACE = 1 ; }
sub trace_off { $TRACE = 0 ; }
sub trace { print STDERR @_ if $TRACE ; } # debugging function in -verbose mode

=pod

=over 4

=item dumpDNA( $db, $seqname, I<$path, $filename, $prefix, $suffix> )

Routine to dump a FASTA DNA sequence from the database. Suffix defaults to '.dna'.

=back

=cut

sub dumpDNA {
    my  $db       = shift ;
    my  $seqname  = shift ;
    my  $path     = shift ;
    my  $filename = shift ; # defaults to seqname if undef
    my  $prefix   = shift ;
    my  $suffix   = shift ;

    croak "GetAce::dumpDNA() needs at least a db & seqname!\n" 
	if !(defined $db and $db and defined $seqname and $seqname) ; 

    $path   = Cwd::cwd() if !(defined($path)   && $path) ;
    $filename = $seqname if !(defined($filename) && $filename) ;
    $prefix = ""         if !(defined($prefix) && $prefix) ;
    $suffix = ".dna"     if !(defined($suffix) && $suffix) ;

    if (my $DNA = $db->fetch('DNA',$seqname)) {
        my $seq = Sequence->new();
        $seq->name( "$DNA" );
        $seq->seq( $DNA->at() );
        my $DNAFilename = filAbsPath("$prefix$filename$suffix",$path) ;
	trace("GFF::GetAce::dumpDNA() file: $DNAFilename\t\n") ;
        if(!open(DNA,"+> $DNAFilename")) {
             carp( "Cannot open DNA output file $DNAFilename: $!!") ;
	     return 0 ;
	}
        $seq->write_fasta(\*DNA);
        close(DNA) ;
        return 1 ;
    } else {
        carp("*** Warning: \"$seqname\" DNA object not found?\n");
        return 0 ;
    }
}

=pod

=over 4

=item dumpGFF( $db, $seqname, I<$path, $filename, $prefix, $suffix, $GFFvers> )

Routine to dump the GFF associated with a given sequence, from the database. 
Suffix defaults to '.gff'. GFF version ($GFFvers) defaults to 2.

=back

=cut

sub ReverseGFF {
    my $file = shift ;

    print STDERR "ReverseGFF($file) called\n" ;
    if(!open(_GFF_REV_IN,"<$file")) {
	carp "Cannot open GFF file $file for input: $!!" ;
	return 0 ;
    }

    my $gffi = new GFF::GeneFeatureSet(2) ;
    $gffi->read(\*_GFF_REV_IN) ;
    close(_GFF_REV_IN) ;

    if(!$gffi->count()) {
	carp "Empty GFF file?\n\n" ;
	return 0 ;
    }

    my ($name,$start,$end) = $gffi->region() ;
    if(!defined($name)) {
	$name = $file ;
	$name = s|^~?(\w+)?/?(\w*/)*|| ; # clip off path prefix
	$name = s|(\.\w*)+$|| ;          # clip off file extensions 
	print STDERR "Name from filename: $name\n" ;
    }

    if(!(defined($start) and defined($end))) {
	($start,$end) = $gffi->min_max_range() ;
	print STDERR "Range from min_max_range(): $start..$end\n" ;
    }

    $gffi->region($name,$start,$end) ; # reset just in case

    foreach my $gf ($gffi->order_gf) {

	# reverse complement the coordinates
	my $gfstart = $gf->start ;
	my $gfend   = $gf->end ;
	$gf->start($end-$gfend+$start) ;
	$gf->end($end-$gfstart+$start) ;

	# reverse the strand
	my $strand = $gf->strand() ;
	if($strand eq '+') {
	    $gf->strand('-') ;
	} elsif($strand eq '-') {
	    $gf->strand('+') ;
	} # otherwise, do nothing
    }

    # overwrite input file?
    if(!open(_GFF_REV_OUT,"+>$file")) {
	carp "Cannot open GFF file $file for output: $!!" ;
	return 0 ;
    }
    $gffi->dump_header(\*_GFF_REV_OUT) ;
    $gffi->dump(\*_GFF_REV_OUT) ;
    close(_GFF_REV_OUT) ;

    return 1 ;
}

sub dumpGFF {
    my  $db       = shift ;
    my  $seqname  = shift ;
    my  $path     = shift ;
    my  $filename = shift ; # defaults to seqname if undef
    my  $prefix   = shift ;
    my  $suffix   = shift ;
    my  $GFFvers  = shift ;
    my  $source   = shift ; # <source> restrictions on dumping, of form 'a|b|c'
    my  $feature  = shift ; # <feature> restrictions on dumping, of form 'a|b|c'

    croak "GetAce::dumpGFF() needs a defined and non-NULL db!\n" 
	if !(defined $db and $db) ; 

    croak "GetAce::dumpGFF() needs a defined and non-empty seqname!\n" 
	if !(defined $seqname and $seqname) ; 

    $path    = Cwd::cwd() if !(defined($path)   && $path) ;
    $filename = $seqname if !(defined($filename) && $filename) ;
    $prefix  = ""         if !(defined($prefix) && $prefix) ;
    $suffix  = ".gff"     if !(defined($suffix) && $suffix) ;
    $GFFvers = "2"        if !(defined($GFFvers) and $$GFFvers) ;

    if(defined($source)) {
	$source = "-source $source" ;
    } else {
	$source = '' ;
    }

    if(defined($feature)) {
	$feature = "-feature $feature" ;
    } else {
	$feature = '' ;
    }

    my $GFF_Filename = filAbsPath("$prefix$filename$suffix",$path) ; 
    trace("GFF::GetAce::dumpGFF() file: $GFF_Filename\t\n") ;
    my $command  = "gif seqget \"$seqname\" ; seqfeatures $source $feature -version $GFFvers -file $GFF_Filename";
    my $results = $db->raw_query($command);
    my $reverse = 0 ;
    if($results =~ /Can\'t\s+GFF\s+dump\s+from\s+(a\s+)?reversed/) {
	# then try getting the reverse complement GFF.
	$command  = "gif seqget \"$seqname\" ; seqactions -rev_comp;  seqfeatures $source $feature -version $GFFvers -file $GFF_Filename";
	$results = $db->raw_query($command);
	$reverse = 1 ;
    }
    if($results !~ /FMAP_FEATURES +\"$seqname\"/){
	carp("*** Warning: \"$seqname\" GFF dump failure: $results!\n") ;
        return 0 ;
    }

    ReverseGFF($GFF_Filename) if $reverse ;

    return 1 ;
}

=pod

=over 4

=item dumpACE( $db, $seqname, I<$path, $filename, $prefix> )

Routine to dump the ?Sequence ACEDB objects record of a given sequence and
all its Subsequences, from the database. File suffix is '.ace'.

=back

=cut

=pod

=over 4

=item dumpGenes( $db, $seqname, I<$path, $filename, $prefix, $suffix, $gene, $namepat > )

Routine to dump 'true genes' list in the given sequence 
from the database. Suffix defaults to '.tgl'.
=back

=cut

sub dumpGenes {
    my $db       = shift ;
    my $seqname  = shift ; 
    my $path     = shift ;
    my $filename = shift ; # defaults to seqname if undef
    my $prefix   = shift ;
    my $suffix   = shift ;
    my $gene     = shift ;
    my $namepat  = shift ;

    croak "GetAce::dumpACEObject() needs at least a database handle and a sequence name!\n" 
	if !(defined $db and $db and defined $seqname and $seqname) ; 

    $path   = Cwd::cwd()   if !(defined($path)   && $path) ;
    $filename = $seqname if !(defined($filename) && $filename) ;
    $prefix = ""           if !(defined($prefix) && $prefix) ;
    $suffix = ".tgl"       if !(defined($suffix) && $suffix) ;
    $gene   = $IsaGene     if !(defined($gene) and $gene) ;
    $namepat = $NameFilter if !(defined($namepat) and $namepat) ;

    # Query for the names of true genes in clone itself
    my $query  = "find sequence \"$seqname\" ; follow Subsequence ; $gene" ; 
    my @genelist = $db->fetch( -query => $query ) ;

    # Query to include the names of true genes in any Source (link) sequence
    $query  = "find sequence \"$seqname\"; follow Source ; follow Subsequence ; $IsaGene" ; 
    my @srcgenelist = $db->fetch( -query => $query ) ;
    if(scalar(@srcgenelist)) {
	push(@genelist,@srcgenelist) ;
    }

    if ( scalar(@genelist) == 0 ) {
       carp( "*** Warning: no \"true genes\" found in \"$seqname\"?\n") ;
       return 0 ;
    }

    # parse gene list for unique root names (ignoring functional suffixes)
    my ($name, %namelist) ;
    foreach $name (@genelist) { 
	$name =~ /$namepat/i ;
        $namelist{$&} = 1 ;
    }
         
    my $GeneFilename = filAbsPath("$prefix$filename$suffix",$path) ;
    trace("GFF::GetAce::dumpGenes() file: $GeneFilename\t\n") ;
    if( !open(GENES,"+> $GeneFilename") ) {
         carp "Cannot open \'true gene\' output file $GeneFilename: $!!" ;
         return 0 ;
    }

    foreach $name (keys %namelist) {
       print GENES "$name\n" ;
    }     
    close(GENES) ;
    return 1 ;
}


sub dumpACEObject {
    my $db       = shift ;
    my $seqname  = shift ;
    my $file     = shift ;

    croak "GetAce::dumpACEObject() needs at least a database handle and a sequence name!\n" 
	if !(defined $db and $db and defined $seqname and $seqname) ; 

    $file = \*STDOUT if !(defined $file and $file) ;

    my $command  = "find Sequence \"$seqname\"";
    my $result = $db->raw_query($command) ;
    $result =~ s/\0//g;
    if( $result =~ m|// Found 0 objects in this class| ) {
	print $file   "// No ACEDB object found for Sequence \"$seqname\"?\n\n" ;
	carp("*** No ACEDB object found for Sequence \"$seqname\"?\n") ;
        return 0 ;
    } else {
        $command  = "show -a"; # get the dump of the object...
        my $result = $db->raw_query($command) ;
        $result =~ s/\0//gs ;     # nulls filtered out
        $result =~ s|^//.*$||gm ; # comments filtered out
        $result =~ s|\n\n|\n|gs ; # blank lines filtered out
        print $file $result, "\n" if defined $result and $result ;
        return 1 ;
    }
}

sub dumpACE {
    my $db       = shift ;
    my $seqname  = shift ; 
    my $path     = shift ;
    my $filename = shift ; # defaults to seqname if undef
    my $prefix   = shift ;
    my $suffix   = shift ;

    croak "GetAce::dumpACE() needs at least a database handle and a sequence name!\n" 
	if !(defined $db and $db and defined $seqname and $seqname) ; 

    $path   = Cwd::cwd() if !(defined($path)   && $path) ;
    $filename = $seqname if !(defined($filename) && $filename) ;
    $prefix = ""         if !(defined($prefix) && $prefix) ;
    $suffix = ".ace"     if !defined($suffix) ;

    # Retrieve the sequence objects here...
    my $parent = $db->fetch('Sequence',$seqname);
    if (! defined($parent) ) {
       carp("*** Warning: \"$seqname\" ACEDB sequence object not found?\n") ;
       return 0 ;
    }
    my $query  = "find sequence \"$seqname\" ; follow Subsequence" ;
    my @subsequences = $db->fetch( -query => $query );

    my $AceFilename = filAbsPath("$prefix$filename$suffix",$path) ;
    trace("GFF::GetAce::dumpACE() file: $AceFilename\t\n") ;
    if(!open(ACE,"+> $AceFilename")) {
         carp "Cannot open ACEDB output file $AceFilename: $!!" ;
         return 0 ;
    }

    print ACE "// ACE dump for Sequence $parent and associated Subsequences\n" ;
    dumpACEObject($db,$parent,\*ACE) ; 
    foreach my $subseq (@subsequences) {
	dumpACEObject($db,$subseq,\*ACE) ;
    }    
    close(ACE) ;
    return 1 ;
}


1; 

__END__

=pod

=head1 REVISIONS

28/9/99 - $source and $feature arguments added to DumpGFF()

22/9/99 - explicit filename argument

13/9/99 - trace_on and off!

5/7/99  - Created: from GetAce.pl subroutines

=cut



