
#
# Perl Module for SPEM::Entry.pm
#
# Cared for by Matthew Pocock <mrp@sanger.ac.uk>
#
# $Header: /nfs/humace2/CVS_master/PerlModules/SPEM/Iterator.pm,v 1.1 1999/06/10 15:17:58 jgrg Exp $
#
#Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE

package SPEM::Iterator;

use strict;
use Carp;


### ITERATOR METHODS ###

sub iterator {
    confess("iterator() method must be provided by inheriting package");
}

# Gets called by iterator method in Parent
sub new {
    my( $pkg, $parent, $arrayRef ) = @_;
    
    return bless {
        '_index' => $parent->freeIndex(),
        '_array' => $arrayRef,
    }, $pkg;
}

sub next {
    my( $iterator ) = @_;
    
    my $index = $iterator->{'_index'};
    my $i = $$index;
    if (defined $i) {
        $$index++;
    } else {
        confess("Use of stale iterator - you must make a new one with the iterator() method");
    }
    
    if (my $element = $iterator->{'_array'}[$i]) {
        # We'll get an element if we're within the array
        return $element;
    } else {
        # Delete the iterator and return nothing
        undef $$index;
        return;
    }
}


### THESE METHODS ACT ON THE PARENT'S "_ITERATORS" ARRAY ###
### "Friend" methods? ###

# Return ref to next free position in the '_ITERATORS'
# array, which it sets to zero
sub freeIndex {
    my( $parent ) = @_;
    
    # Make a new empty iterators hash in the parent object,
    # unless it already exists.
    $parent->{'_ITERATORS'} ||= [];
    
    my $index = 0;
    
    # Get index postion left by an old iterator, if it
    # exists.  (Avoids growing '_ITERATORS' array to
    # satisfy people who are paranioid about memory
    # consumption.)  Else will just return first index
    # off the end of the array.
    while (defined $parent->{'_ITERATORS'}[$index]) {
        $index++;
    }

    # Get ref to position in _ITERATORS array
    my $indexRef = \$parent->{'_ITERATORS'}[$index];
    $$indexRef = 0; # Set to zero
    # Return ref to scalar in _ITERATORS array
    return $indexRef;
}

sub maintainIteratorList {
    my( $parent, $action, $where ) = @_;
        
    # Just return if there's no list to maintain
    return unless $parent->{'_ITERATORS'};
    
    # Check we got an integer
    confess("Non-integer location ('$where') provided to maintainIndexList()")
        unless $where =~ /^\d+$/;
    
    if ($action eq 'ADD') {
        foreach my $i (@{$parent->{'_ITERATORS'}}) {
            if (defined $i) {
                $i++ if $i >= $where;
            }
        }
    } elsif ($action eq 'REMOVE') {
        foreach my $i (@{$parent->{'_ITERATORS'}}) {
            if (defined $i) {
                $i-- if $i >= $where;
            }
        }
    } else {
        confess("Unknown action : ('$action')");
    }
}

1;

__END__

=head1 NAME

SPEM::Iterator.pm

=head1 DESCRIPTION

Provides methods for iterating accross the arrays
of objects found in SPEM objects, in, for example
the EMBL::FeatureTable or EMBL::Entry objects.

=over 4

=head2 iterator

This method shold be provided by the inheriting
package, and calls the new method.

=head2 new (OBJECT, ARRARREF)

Expects a reference to the parent object, and a
reference to the array to be iterated over. 
Returns a new iterator object, which has a
reference to a scalar in the _ITERATORS array of
the parent object.

=head2 next

Returns the next object from the array, or undef
if it goes off the end of the array

=back maintainIteratorList ("ADD|REMOVE", WHERE)

Should be called by methods in the Parent object
which add or remove elements from the object
array.  It increments or decrements the scalars
in the _ITERATORS array, so that the next element
returned by the next method on any of the
iterators is always the same, unless this element
has been removed.  The WHERE argument is the
array index affected by the add or remove action.

=head1 AUTHOR

B<James Gilbert> Email jgrg@sanger.ac.uk

