
#
# Perl Module for SPEM::Parser
#
# Cared for by Matthew Pocock <mrp@sanger.ac.uk>
#
# Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE

package SPEM::Parser;

use vars qw($AUTOLOAD @ISA @EXPORT_OK %fields);
use Exporter;
use Carp;
use strict;
use SPEM::Set;
use SPEM::Entry;

#
# Place functions/variables you want to *export*, i.e. be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = qw();

# @ISA has our inheritance.
@ISA = ( 'Exporter' );

my %fields = (
	      #Insert field names here as field => undef,
	      functionTable => undef,  # the function table
	      _default => undef,        # the default function
	      );

sub new {
    my $ref = shift;
    my $class = ref($ref) || $ref;
    my $self = { '_permitted' => \%fields,
		 %fields };

    $self->{'functionTable'} = {};
    bless $self, $class;

    return $self;
}

#
# Autoload up my field name set/get
#
sub AUTOLOAD {
    my $self = shift;
    my $type = ref($self) || carp "Ok, $self is not an object of mine!";
    my $name = $AUTOLOAD;

    # don't propagate DESTROY messages...

    $name =~ /::DESTROY/ && return;

    $name =~ s/.*://; #get only the bit we want
    unless (exists $self->{'_permitted'}->{$name} ) {
	carp "In type $type, can't access $name - probably passed a wrong variable into Sequence  ";
    }
    if (@_) {
	return $self->{$name} = shift;
    } else {
	return $self->{$name}
    }
}

sub newEntry {
    my $self = shift;
    confess (
	     "You have tried to call the pure virtual function newEntry for object $self.\n".
	     "This function mush be overriden by all base classes.\n".
	     "See the SPEM::Parser documentation for further details."
	     );
}

sub newSet {
    my $self = shift;
    confess (
	     "You have tried to call the pure virtual function newSet for object $self.\n".
	     "This function mush be overriden by all base classes.\n".
	     "See the SPEM::Parser documentation for further details."
	     );
}

sub newLine {
    my $self = shift;
    confess (
	     "You have tried to call the pure virtual function newLine for object $self.\n".
	     "This function mush be overriden by all base classes.\n".
	     "See the SPEM::Parser documentation for further details."
	     );
}

# parses an entire stream, returning a set containing all of the entries contained within
sub setFromStream {
    my $usage = 'Usage: $parser->setFromStream(\*FH)';
    my $self = shift || carp $usage;
    my $fh = shift || carp $usage;                         # our stream
    my $set = $self->newSet();                             # make a new set using the factory method newSet
    my $entry;                                             # a single entry

    #no strict 'refs';

    while($entry = $self->nextEntryFromStream($fh)) {       # get the next entry from the stream
	@{$entry->lineEntries()} &&                           # if this entry contains any lines
	    $set->addEntry($entry);                           # add it to the set
    }

    return $set;                                           # return this set
}

# parses a stream up to and including the end of the first entry encountered, returning that entry
sub nextEntryFromStream {
    my $usage = 'Use: $parser->nextEntryFromStream(\*fh, \$currentLine)';
    my $self = shift || carp $usage;
    my $fh = shift || carp $usage;                         # stream to read from
    my $line = <$fh>;                                      # buffere for the current line of text
    my $entry = $self->newEntry();                         # make a new entry using the factory method newEntry
    my $ft = $self->functionTable();                       # get the function mapping hash
    my $default = $self->_default();                       # and the default function
    my $tag;                                               # the two letter line tag
    my $value;                                             # the thing associated with a given tag

    return undef unless defined $line;                     # if we are trying to read from an ended stream, return nothing

    while((defined $line) && ($line ne "//\n")) {          # while the stream is open
	if($line =~ m/\S/) {                               # if this line contains any non-space characters
	    $tag = substr $line, 0, 2;        # get the tag
	    $value = $ft->{$tag} || $default; # look up what to do with the line

	    if (ref $value eq 'CODE') { 
                # entry under this tag is code
	        # - presume it to be a function returning a line group
                $entry->addLineGroup(&$value($fh, \$line))
            } else {
	        # entry under this tag is not code
	        # - presume it to be a module or object to be handled with newLine
		$entry->addLineGroup($self->newLine($value, $fh, \$line));
            }
	} else {                                           # the line contained nothing
	    $line = <$fh>;                                 # so get the next line
	}
    }

    return $entry;                                         # and return what we have read in.
}

# registers a function, module or object with this parser with the given tag
sub register {
    my $usage = 'Usage: $parser->register($tag, $thing)';
    my $self = shift || carp $usage;
    my $tag = shift || carp $usage;
    my $func = shift || carp $usage;
    my $ft = $self->functionTable();

    unless(length $tag == 2) {
	carp "The tag must be two characters long. '$tag' is not this length\n";
    }

    if(exists $ft->{$tag}) {
	carp "$tag is already registered to another funtion.  Previous assignment will be overwriten.";
    }

    $ft->{$tag} = $func;

    return 1;
}

# unregisters a tag, and it's associated callback
sub unregister {
    my $usage = 'Usage: $parser->unregister($tag)';
    my $self = shift || carp $usage;
    my $tag  = shift || carp $usage;

    my $ft = $self->functionTable();

    if (exists $ft->{$tag}) {
	delete $ft->{$tag};
	return 1;
    } else {
        carp "$tag was not present in the function table";
	return undef;
    }
}

# registers the thing to deal with the default line type.
sub registerDefault {
    my $usage = 'Usage: $parser->registerDefault($thing)';
    my $self = shift || carp $usage;
    my $func = shift || carp $usage;

    if (defined $self->_default) {
	carp "Overriding the default parsing behavior for this parser";
    }

    $self->_default($func);

    return 1;
}

1;  # says use was ok
__END__

=head1 NAME

SPEM::Parser

=head1 DESCRIPTION

B<SPEM::Parser> is part of the core modules that deal with parsing and
representing SWISSPROT and EMBL formatted information. See <SPEM::Spem>
for an overview. Throughout this documentation, the concrete example
of L<EMBL::Parser> will be used for illustration.

B<SPEM::Parser> is the baseclass for parser objects that parse SWISSPROT
and EMBL type streams into a set of SPEM::Entry and SPEM::Set derived
objects.

You can largely 'inherit' this documentation into a derived class.

=head1 AUTHOR

B<Matthew Pocock> Email mrp@sanger.ac.uk

=head1 Introduction

This module is meant to be used purely as a base class. You can make objects
of type SPEM::Parser, but if you try to use them they will complain that you
have not overridden some member functions.  This class encapsulates the
funcitonality necisary to parse a stream, looping over each line of text
and calling some function for that line. It does not define the actual
objects that will be created from the stream.  This is up to you to
implement through subclassing.

The functions within this module can be broken down into three types. The
first are the stream parsing methods (C<setFromStream> and C<nextEntryFromStream>).
These are idealy the only methods that someone using a derived module
should have to worry about. The seccond set deal with registering items
as being interrested in a given line type (C<register>, C<unregister> and
C<registerDefault>).  They allow you to configure each parser to recognize
different line types. The third group are the factory methods (C<newEntry>,
C<newSet> and C<newLine>). These are designed allow
the parsing routines to encapsulate parsing without knowing what type of
objects they are interacting with. Oh, and of course, there is a C<new>
constructor that returns a virgin parser object.

Happy parsing.

=head1 Things you can register with the parser.

You may wish to read the parsing modle described in L<SPEM::Spem> before
reading this section. The long and short of it is that you can register
just about anything as being interested in any two letter code. This
describes what those anythings are likely to be.

You use the parser object to register a function, module or object as
being interested in a particular line type.  It provides you with great
flexibility in extending the parser to do different things with the
files it reads.

If you chose to register a function, it should expect to be called with
two arguments only.  The first will be a \*FH, and the second a reference
to a scalar.

If you chose to register a module or object, the Parser member function C<makeLine>
will be invoked with the registered value, \*FH and a reference to a scalar.
C<makeLine> should then call a function on value using the arrow operator,
and passing it \*FH and the reference. The resulting return value should be returned.

In either case, the function invoked should return a line group object.
This should forfill the interface conditions for a line group, as
described in the section L<Roll your own line groups>

=head1 Functions

=head2 Stream parsing functions

=over

=item setFromStream

Parses a stream, and returns an SPEM::Set derived object containing the
entries from the stream.  Parse it a glob to a filehandle as it's only
argument.

It will consume the entire stream, creating as many SPEM::Entry derived
objects from it as it can.  These will all be put into a single SPEM::Set
derived object, and parsed back to you. The file handle will then be
exhausted, so you should close it. E.G.

 open(STREAM, 'fileName');
 $set = $parser->setFromStream(\*STREAM);
 close STREAM;

=item nextEntryFromStream

This function consumes one entry's worth of the stream, returning a single
SPEM::Entry derived object representing that consumed text. It is used
by C<setFromStream> to read in each entry within the stream. It takes
two arguments.  The first is a ref to a glob of a file handle. The second
is a ref to the current line of text. When the function has finnished
constructing the entry from the stream, the current line of text will
represent the first line that was not added to the entry. Thus, this
will either be the first line of the next entry, or off the end of the
stream. So here is an example of how to use it.

 $line = <STREAM>;
 $e1 = $parser->nextEntryFromStream(\*STREAM, \$line);   # read in 1st entry
 $e2 = $parser->nextEntryFromStream(\*STREAM, \$line);   # read in 2nd entry
 if($line) {                                             # if the stream hasn't ended
	$e3 = $parser->nextEntryFromStream(\*STREAM, \$line); # read 3rd entry
 } else {
	print "End of stream encountered\n";             # otherwise print warning.
 }

=back

=head2 Registration functions

=over

=item register

Used to register a tag and function or module with an instance of a
SPEM::Parser derived object. Use it like this:

    $parser->register($tag, $module);
    $parser->register($tag, $object);
 or
    $parser->register($tag, \&function);

If you try to register a tag that has already been used, then a warning
will be printed, but the new item will replace the old one.

If you register a function, then it must be the fully qualified function name.
It will be assumed to take only the reference to a glob to a file handle, and
the reference to a line of text.  Of course, in this case you can specify any
function name you wish, allowing any function in any package to produce the
line groups of the type you want. It will be invoked as &$funcName(\*FH, \$line).

If you register a module, then that module must provide a function that can be called
by your C<newLine> function.
It will be assumed to take a class name, the reference to a glob to a file handle,
and the reference to a line of text.

Likewise, if you register an object, then that object must provide the function
called from C<newLine>.
It will be assumed to take an object, the reference to
a glob to a file handle, and the reference to a line of text.

=item unregister

Removes a tag and it's associated function from the Parser.

Use:

    $parser->unregister($tag);

=item registerDefault

Registers a function to call when no other function is interested in a
line.

Use:

    $parser->registerDefault(\&function);
 or
    $parser->registerDefault($module);
    $parser->registerDefault($object);

See register above for a description of the two syntaxes.

=back

=head2 Factory Methods

The functions must be overridden on all instantiatable subclasses. They allow the
algorithm of parsing to be encapsulated without worrying about what specific types
of objects are being produced and manipulated.

If you instantiate and attempt to use an object where these have not been overridden,
a long error message will be printed out and the programm will terminate.

You may well wish to leave these functions undocumented in a derived class.

=over

=item newSet

In a subclass, you should override this to return an empty C<SPEM::Set> derived object.
This function will be called during passing to make a set to put the entries in. It is
called with no arguments, and should return just one C<SPEM::Set> derived object. It is
a virtual function and will die with a warning if you try to call SPEM::Set::newSet.

This is an example of what I envisige the derived function looking like.

 package EMBL::Parser;

 ...

 sub newSet {
   return new EMBL::Set();
 }

 ...

=item newEntry

In a subclass, you should override this to return an empty C<SPEM::Entry> derived object.
This function will be called during passing to make a new entry for each entry in the stream.
It is called with no arguments and should return just one C<SPEM::Entry> derived object.
It is a virtual function and will die with a warning message if you try to call
SPEM::Set::newEntry.

This is an example of what I envisige the derived function looking like.

 package EMBL::Parser;

 ...

 sub newEntry {
   $self = shift;
   return new EMBL::Entry();
 }

 ...

=item newLine

In a subclass, you should override this to call the apropreate constructor or factory method
to create a new line from a stream when presented with the module name or an object.
It is called as follows:

    $parser->newLine($value, $fh, $lineRef);

$value is the object that was associated in the parser with a line type that matched the
current line.  It will be either a string representing a module name, or an object implimenting
a class factory. I envisig the function looking like this:

 package EMBL::Parser;

 sub newLine {
    my $self = shift;
    my $val = shift;
    my $fh = shift;
    my $lineRef = shift;
    return $val->fromEMBLStream($fh, $lineRef);
 }

This framework allows you to substitute fromEMBLStream with fromSWISSStream etc., or whatever
other standard interface you prefer.

=back

=head1 Example of modifying the parser

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif">This information is
intended only for developers who wish to make custom parsers.

This section assumes that you are working with the derivation implimented in
L<EMBL::Parser> module, but all of the behavure described here should be identical
for any other parser derived from C<SPEM::Parser> unless you have overridden some
of the non virtual members.

Let us assume that you prety much like what the default does, but want to let the
default behavure be to image lines.

=for html
<PRE>
$parser = Embl->defaultParser();       # uses the Embl module to create a default parser
$parser->registerDefault(EMBL::Image); # replacies the default line type with EMBL::Image
</PRE>

Now, you wish to unregister the parsing of the date lines.

=for html
<PRE>
$parser->unregister('DT');
</PRE>

Imagine that you are reading EMBL type files from a source where there is a
systematic error in the organism lines OS.  You wish to put your own function
which knows how to correct this, called C<myPackage::correctOS>. You would now
do the following, to change the OS line parsing, without getting warnings.

=for html
<PRE>
# unregister previous interested functions and modules
$parser->unregister('OS');
$parser->unregister('OC');
$parser->unregister('OG');
# now register my function
$parser->register('OS', \&myPackage::correctOS() );
</PRE>

You are now ready to go parse - with a command like this.

 $set = $parse->fromStream(\*FH);

=head1 Roll your own line groups

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif">This information is
intended only for developers who wish to make custom line types.  If this
is not you then you realy do not need to read this.<p>

For an object to function as a line group that can be created from a stream
and writen to a stream, there are two pieces of functionality it must have.
Firstly, it must have a way of being read from a stream.  Secondly it must
have a function that allows it to be writen to a stream. Sounds
obvious, doesn't it.

The following sections describe both of these steps, together with code
examples, and I hope it explores the major possible different ways to go
at each step.

=head2 Reading your object from the SPEM formatted stream

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif"><p>

There are three possible ways that a line group can be created during parsing.
They are listed here

=over

=item 1

Registering a function to the parser (as C<\&module::funcName>)

Invoked as &module::funcName(\*FH, \$line). This option is employed by L<Embl> in
which a funciton is called that prints a warning message and then defers
responsibility for parsing to EMBL::Image. In this way, an image of the line
is made and added to the entry, but warining that the default parser had to
be used is given.

=item 2

Registering a module with a function recognized by your C<newLine> with the parser

Invoked within C<newLine> presumably as something like 
$module->fromEMBLStream(\*FH, \$line) where for argument sake, each linegroup
module provides a constructor named fromEMBLStream which is called from within
EMBL::Parser::newLine.

=item3

Registering an object (a factory etc.) with a function recognized by your
C<newLine> function with the parser

Invoked within C<newLine> presumably as $object->fromEMBLStream(\*FH, \$line)
where for argument sake, the object has a fromEMBLStream factory method which
is called from within EMBL::Parser::newLine.

=back

In each case, the return value of the function will be assumed to have
some way to write it's self out - for the embl example this will be
the toEMBLStream function, and will be added to the entry currently
being built. This function will be called by the funciton that overrided
L<SPEM::Entry#writeLine> - in this case L<EMBL::Entry#writeLine>. You may wish
to inforse this within your C<newLine> function using the C<can> construct as
is done in L<EMBL::Parse#newLine>.

So here are some code outlines for how to start these functions.

=head3 Absolute function

This will be invoked directly.  No class parameters will be given to it,
or anything like that, so it will look something like

=for html
<PRE>
sub myLineGroupCreatingFunction {
    my $fh = shift;         # the stream being read from
    my $lineRef = shift;    # a reference to the next line to be parsed
    my $lineGroup = new lineOfMyType();
#
# code here to process the file and update $lineRef
#
    return $lineGroup;
}
</PRE>

=head3 Module function

Remember - this will be invoked using the arrow operator.

=for html
<PRE>
sub fromEMBLStream {
    my $class = shift;      # given my class type
    my $fh = shift;         # the stream being read from
    my $lineRef = shift;    # a reference to the next line to be parsed
    my $lineGroup = new $class;  # OK - so this is a little bit risky
                                 # as we may be invoked by a derived
                                 # class, and end up constructing an
                                 # object with different behaviour to
                                 # the expected.  See later for how to
                                 # deal with this.
#
# code here to process the file and update $lineRef
#
    return $lineGroup;
}
</PRE>

This is the primary approach adopted in the EMBL modules.

=head3 Object method

This will be invoked as $object->fromEMBLStream, but may return objects
of a different type to it's self.  This could implement a class factory,
or just return an object of the same type as it.

=for html
<PRE>
sub fromEMBLStream {
    my $self = shift;       # given my class type
    my $fh = shift;         # the stream being read from
    my $lineRef = shift;    # a reference to the next line to be parsed
    my $class = something decided by $self
    my $lineGroup = new $class;  # OK - so this is a little bit risky
                                 # as we may be invoked by a derived
                                 # class, and end up constructing an
                                 # object with different behaviour to
                                 # the expected.  See later for how to
                                 # deal with this.
#
# code here to process the file and update $lineRef
#
    return $lineGroup;
}
</PRE>

This may be applicable for cases where there are few possible items on the line -
such as the OC lines. Here, you could have an object which contains a tree of all
current organism classifications, and returns an object that is just an index to the
one applicable for this entry.

=head2 Patterns used in the fromEMBLStream functions

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif"><p>

There are essentially two types of fromEMBLStream functions.  The first eats one
line of text.  The second eats multiple lines.  I shall give an example of the
code for both, as taken from real examples.

=over

=item Single line eater

=for html
The source code for this is taken from L<EMBL::Image>

=for html
<PRE>
sub fromEMBLStream {
    my $class = shift;           # our class
    my $fh = shift;              # the file handle
    my $lineRef = shift;         # the line reference
    my $line = new SPEM::Image;  # make the image
    
    $line->text($$lineRef);      # fill the image
    
    $$lineRef = <$fh>;           # prime the next line into $$lineRef
    return $line;                # return the new image
}
=for HTML
</PRE>

=item Multiple line eater

=for html
The source code for this example is from L<EMBL::CC>

=for html
<PRE>
sub fromEMBLStream {
    my $class = shift;                   # our class
    my $fh = shift;                      # file handle
    my $lineRef = shift;                 # line reference
    my $lineGroup = new SPEM::CC;        # make a new comment
    my $line;                            # to hold a line of input
    my $text = undef;                    # the text so far

    $line = $$lineRef;                   # prime $line
    while(substr($line, 0, 2) eq 'CC') { # while we are a comment
        chomp($line = substr $line, 5);  # remove leading text
        defined($text) && ($text .= " ".$line) || ($text = $line); # add this line to previous text
        $line = <$fh>;                   # get the next line
        if(not defined $line) { warn "CC: End of stream"; last; } # end of file check
    }
    $lineGroup->comment($text);          # set the comment to $text

    $$lineRef = $line;                   # prime lineRef
    return $lineGroup;                   # return the new comment
}

=for html
</PRE>

=back

=head2 Writing your object to an EMBL formatted stream

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif"><p>

When an SPEM::Set is asked to write to a stream, it calls the member toStream for all of it's entries. Each entry in turn calls its L<SPEM::Entry#writeLine> on each of its line groups which then write themselves to the stream. For your object to fit into this scheme, it must have a function with the correct signature and name to be called from your writeLine function.

Here is a skeletal outline of what it will look like.

=for html
<PRE>
package EMBL::Entry;
 
...
 
sub writeLine {
    my $self = shift;
    my $fh = shift;
    my $what = shift;

    $what->toEMBLStream($fh);
}
 
...
 
package EMBL::ficticiousLineType;
 
sub toEMBLStream {
    my $self = shift;
    my $fh = shift;
 
    print $fh $self->asEMBLFormattedText; # may well do something more fancy than this
 
    return $fh;   # the return value is not currently used but I prefer this
                  # notation, because it allows chaining of toEMBLStream calls.
}
</PRE>
