package Spangle::Analysis::DotPlot;

use GD;

=pod

  =head1 NAME
  
  Spangle::Analysis::DotPlot - creates a coloured dotplot gif from a list
                               of matches

  =head1 SYNOPSIS
  
  my $dp = new Spangle::Analysis::DotPlot(-length1 => $len1,
					  -length2 => $len2,
					  -size    => $size,
					  -start1  => $start1,
					  -start2  => $start2,
					  -end1    => $end1,
					  -end2    => $end2,
					  -clones  => $clones,
					  -col     => $col,
					  );


  $dp->parse_match  ($fh1,1);
  $dp->dotplot      ($giffh);
  $dp->print_matches($out);

  
=head1 DESCRIPTION

A module to create a coloured dotplot from
a series of ungapped match segments

=head1 CONTACT

Michele Clamp (michele@sanger.ac.uk)

=head1 APPENDIX

The rest of the documentation details each of the object methods. Internal methods are usually preceded with a _

=cut

use strict;  
use vars   qw(@ISA);

# Object preamble - inherits from Bio::SeqFeature::Generic

use Bio::Root::RootI;

@ISA = qw(Bio::Root::RootI);

sub new {
  my($pkg,@args) = @_;

  my $self = bless {}, $pkg;


  my ($length1,$length2,$size,$start1,$start2,$end1,$end2,$clones,$col) 
      = $self->_rearrange([qw(LENGTH1
			      LENGTH2
			      SIZE
			      START1
			      START2
			      END1
			      END2
			      CLONES
			      COL)],@args);

  $size = 500 unless $size;

  $self->length1($length1);
  $self->length2($length2);
  $self->size   ($size);
  $self->start1 ($start1);
  $self->start2 ($start2);
  $self->end1   ($end1);
  $self->end2   ($end2);
  $self->col    ($col);
  # Initialize the variables

  $self->{_start1}   = [];
  $self->{_start2}   = [];
  $self->{_length}   = [];
  $self->{_score}    = [];

  return $self; # success - we hope!
}

sub add_match {
  my ($self,$start1,$start2,$length,$score) = @_;

  if (!defined($start1) || !defined($start2) ||!defined($length)) {
    $self->throw("Must enter start1 start2 length for a match. Input was \"$start1,$start2,$length\"");
  }

  push(@{$self->{_start1}},$start1);
  push(@{$self->{_start2}},$start2);
  push(@{$self->{_length}},$length);
  push(@{$self->{_score}} ,$score);

}

sub print_matches {
  my ($self,$fh) = @_;

  my $count = 0;

  while ($count < $self->count) {
    print($fh $self->{_start1}[$count] . " " . $self->{_start2}[$count] . " " .   
$self->{_length}[$count] . " " . $self->{_score}[$count] . "\n"); 
    $count++;
  }
}
    


    
sub length1 {
  my ($self,$length1) = @_;

  if (defined($length1)) {
    $self->{_length1} = $length1;
  }
  return $self->{_length1};
}

sub length2 {
  my ($self,$length2) = @_;

  if (defined($length2)) {
    $self->{_length2} = $length2;
  }
  return $self->{_length2};
}

sub start1 {
  my ($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_start1} = $arg;
  }
  return $self->{_start1};
}
    
sub start2 {
  my ($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_start2} = $arg;
  }
  return $self->{_start2};
}


sub end1 {
  my ($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_end1} = $arg;
  }
  return $self->{_end1};
}
        
sub end2 {
  my ($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_end2} = $arg;
  }
  return $self->{_end2};
}
        
sub size {
  my ($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_size} = $arg;
  }
  return $self->{_size};
}
        
sub clones {
  my ($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_clones} = $arg;
  }
  return $self->{_clones};
}
        
sub col {
  my ($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_col} = $arg;
  }
  return $self->{_col};
}
         
sub count {
  my ($self) = @_;

  my $count1 = $#{$self->{_start1}};
  my $count2 = $#{$self->{_start2}};
  my $count3 = $#{$self->{_length}};

  if ($count1 != $count2  || $count2 != $count3) {
    $self->throw("ERROR: Number of start1 start2 and lengths disagree");
  } else {
    return $count1;
  }
}

sub draw_box_and_scale {
  my ($self,$im,$width,$height,$col,$start1,$end1,$start2,$end2,$clone) = @_;
  
  my $xoffset   = $width * 0.2;
  my $yoffset   = $width * 0.2;

  my $newim = new GD::Image($width + 2 * $xoffset, $height + 2 * $yoffset);

  my $white  = $newim->colorAllocate(255,255,255);
  my $black  = $newim->colorAllocate(0,0,0);
  my $red    = $newim->colorAllocate(255,0,0);

  $newim->rectangle($xoffset,$yoffset,$width+$xoffset+1,$height+$yoffset+1,$black);

  my $tmp = int(log($end1-$start1)/log(10));

  print("Tmp is $tmp\n");

  my $tick = 10**($tmp-1);

  print("Tick is $tick\n");

  my $start = int($start1);  
  my $end =   int($end1);
  

  print("start is $start\n");
  print("end is $end\n");

  my $ticklen = 10;

  my $i = $start;
  my $tickcount = 0;
  while ($i <= $end) {


    $newim->line($xoffset + ($i-$start1)*$width/($end1-$start1),$yoffset-$ticklen,
		 $xoffset + ($i-$start1)*$width/($end1-$start1),$yoffset,$black);
    

    my $val = int($i/1000);

    if ($tickcount % 5 == 0) {
      $newim->stringUp(gdSmallFont,$xoffset + ($i-$start1)*$width/($end1-$start1) - 5,$yoffset - 4*$ticklen,$val,$black);

      $newim->line($xoffset + ($i-$start1)*$width/($end1-$start1),$yoffset-2* $ticklen,
		   $xoffset + ($i-$start1)*$width/($end1-$start1),$yoffset,$black);

      $tickcount = 0;
    }
    $tickcount++;
    $i += $tick;
  }

  $start = int($start2);
  $end =   int($end2);

  print("start is $start\n");
  print("end is $end\n");

  $i = $start;
  $tickcount = 0;

  while ($i <= $end) {
    my $val = int($i/1000);
    $newim->line($xoffset-$ticklen, $yoffset + ($i-$start2)*$height/($end2-$start2),
		 $xoffset    ,$yoffset + ($i-$start2)*$height/($end2-$start2),$black);
    
    

    if ($tickcount % 5 == 0) {
      $newim->string(gdSmallFont,$xoffset - 5*$ticklen, $yoffset + ($i-$start2)*$height/($end2-$start2) - 5,$val,$black);

      $newim->line($xoffset-2*$ticklen, $yoffset + ($i-$start2)*$height/($end2-$start2),
		   $xoffset    ,$yoffset + ($i-$start2)*$height/($end2-$start2),$black);

      $tickcount = 0;
    }
    $tickcount++;

    $i += $tick;
  }
  
  my @xclones;
  my @yclones;

  if ($clone) {
    foreach my $cl (@$clone) {
      if (!($cl->end < $start1 || $cl->start > $end1)) {
	push(@xclones,$cl);
      }
      if (!($cl->end < $start2 || $cl->start > $end2)) {
	push(@yclones,$cl);
      }
    }
    
    my $y   = 5;
    my $dir = 1;

    if ($#xclones <= 25) {
      foreach my $cl (@xclones) {
	print("Clone is " . $cl->id . " " . $cl->start . " " . $cl->end . "\n");
	my $clstart = $cl->start;
	my $clend   = $cl->end;
	
	if ($clstart < $start1) {$clstart = $start1;}
	if ($clend   > $end1) {$clend   = $end1;}
	
	my $xstart = $xoffset + ($clstart-$start1)*$width/($end1-$start1);
	my $xend   = $xoffset + ($clend  - $start1)*$width/($end1-$start1);
	
	$newim->rectangle($xstart,$y,
			  $xend  ,$y+12,$black);
	
	my $id = $cl->id;
	$newim->string(gdSmallFont, ($xstart + 3) , $y,$id,$black);
	
	$y = $y + $dir * 12;
	$dir = $dir * -1;
      }
    }  
      my $x   = 15;
      $dir = 1;
    

    if ($#yclones <= 25) {
      foreach my $cl (@yclones) {
	print("Clone is " . $cl->id . " " . $cl->start . " " . $cl->end . "\n");
	my $clstart = $cl->start;
	my $clend   = $cl->end;
	
	if ($clstart < $start2) {$clstart = $start2;}
	if ($clend   > $end2) {$clend   = $end2;}
	
	my $ystart = $yoffset + ($clstart-$start2)*$height/($end2-$start2);
	my $yend   = $yoffset + ($clend  - $start2)*$height/($end2-$start2);
	
	$newim->rectangle($x,   $ystart,
			  $x+12,$yend,$black);
	
	my $id = $cl->id;
	$newim->stringUp(gdSmallFont,  $x,$yend - 3,$id,$black);
	
	$x = $x + $dir * 12;
	$dir = $dir * -1;
      }
      
    }
  }
  $newim->copy($im,$xoffset+1,$yoffset+1,0,0,$width,$height);

  
  return $newim;
}

sub dotplot  {
  my ($self,$fh) = @_;

  $fh   = \*STDOUT unless $fh;

  my $size   = $self->size;

  my $start1 = $self->start1;
  my $start2 = $self->start2;
  
  my $end1   = $self->end1;
  my $end2   = $self->end2;

  my $len1   = ($end1 -  $start1 + 1 );
  my $len2   = ($end2 -  $start2 + 1 );
  
  if ($len2 > $len1) {
      $size = $size* $len1/$len2;
  }

  my $height = $size;
  my $width  = $size * $len2/$len1;

  print(STDERR "Gif width/height = $width/$height\n");

  my $im     = new GD::Image($width,$height);

  my $white  = $im->colorAllocate(255,255,255);
  my $black  = $im->colorAllocate(0,0,0);       
  my $red    = $im->colorAllocate(255,0,0);
  my $green  = $im->colorAllocate(0,255,0);
  my $blue   = $im->colorAllocate(0,0,255);
  my $orange = $im->colorAllocate(255,128,0);
  my $yellow = $im->colorAllocate(255,255,0);

  if ($colour) {
    $im->filledRectangle(0,0,$width,$height,$black);
  }

  my $count = 0;

 POINT: while ($count < $self->count) {
     my $s1coord = $self->{_start1}->[$count];
     my $s2coord = $self->{_start2}->[$count];
      
     $count++;

     next POINT if ($s1coord < $start1 || $s2coord < $start2 || $s1coord > $end1 || $s2coord > $end2);

     my $s1     = ($s1coord - $start1)* $height/$len1;
     my $s2     = ($s2coord - $start2)* $width /$len2;
      
     my $len   = $self->{_length}->[$count];
     my $score = $self->{_score}->[$count]/($len);
     
     if ($colour == 2) {
	 $score = $self->{_score}->[$count]/100;
     }
     
     my $col = $black;
     
     if ($colour) {
	 if ($score > 4) {
	     $col = $red;
	 } elsif ($score > 3) {
	     $col = $orange;
	 } elsif ($score > 2) {
	     $col = $yellow;
	 } elsif ($score > 1) {
	     $col = $green;
	 } else {
	     $col = $blue;
	 }
     }
     if ($len > 0 ) {
	 
	 $im->line($s1,$s2,$s1+$len*$height/$len1,$s2+$len*$width/$len2,$col);
	 $im->filledRectangle($s1-2,$s2-2,$s1+$len*$height/$len1+1,$s2+$len*$width/$len2+1,$col);
	 #$im->line($s1+1,$s2,$s1+$len*$height/$len1 + 1,$s2+$len*$width/$len2,$col);
	 
     } else {
	 
	 #  $im->line($width - $s1,$s2,$width - $s1+$len*$height/$len1,$s2-$len*$width/$len2,$black);
	 
	 $im->line($s1,$s2,$s1+$len*$height/$len1,$s2-$len*$width/$len2,$col);
	 $im->filledRectangle($s1-1,$s2-1,$s1+$len*$height/$len1+1,$s2-$len*$width/$len2+1,$col);
     }
     $count++;
 }
  
  my $newim = $self->draw_box_and_scale($im,$width,$height,$black,$start1,$end1,$start2,$end2,$clones);
  
  print $fh $newim->gif;
}

sub parse_match {
  my ($self,$fh) = @_;
  my $line;
  my $count = 0;

  while ($line = <$fh>) {
      chomp($line);
      my ($start,$end,$length,$score) = split(' ',$line);
      $self->add_match($start,$end,$length,$score);
      $count++;
  }
}

sub parseMSP {
    my ($self,$fh,$gcoord,$chr) = @_;
    my $line;
    my $clones;
    
    $self->throw("ERROR: No length set for sequence 1") unless $self->length1;
    $self->throw("ERROR: No length set for sequence 2") unless $self->length2;
    
    print("Starting parse\n");
    
    while ($line = <$fh>) {
	my ($score,$pid,$qstart,$qend,$id1,$hstart,$hend,$id,$rest) = split(' ',$line);
	if ($id1 =~ /(.*)\/(.*)-(.*)/) {
	    my $clstart1 = $2;
	    my $clend1   = $3;
	    $qstart     += $clstart1;
	    $qend       += $clstart1;
	    
	}
	if ($id=~ /(.*)\/(.*)-(.*)/) {
	    my $clstart2 = $2;
	    my $clend2   = $3;
	    $hstart     += $clstart2;
	    $hend       += $clstart2;
	}

	$clones->{$id}++;
	
	my $foundchr = 0;
	
	if ($gcoord->{$id} && $gcoord->{$id}{chr} eq $chr) {
	    $foundchr = 1;
	    $hstart += $gcoord->{$id}{start};
	    $hend   += $gcoord->{$id}{start};
	}
	
	if ($foundchr == 1) {
	    print("Adding $qstart $hstart " . (($qend - $qstart) + 1) . "\n");
	    if ($qstart > $qend) {
		$self->add_match($qend,$hstart,($qend - $qstart + 1),$score);
	    } else {
		$self->add_match($qstart,$hstart,($qend - $qstart + 1),$score);
	    }
	}
    }
    
    print("Done parsing MSP output\n");
}

1;
