package Spangle::BlastEnd;

=pod

  =head1 NAME
  
  Spangle::BlastEnd  - shoddy clone overlap finding object

  =head1 SYNOPSIS
  
    $blast = new Spangle::BlastEnd($seq1,$seq2);

=head1 DESCRIPTION

Runs blast on the ends of sequences only to find any overlap between
them.  Used for finding overlap between clones

=head1 CONTACT

Describe contact details here

=head1 APPENDIX

The rest of the documentation details each of the object methods. Internal methods are usually preceded with a _

=cut

use strict;  
use vars   qw(@ISA);

# Object preamble - inherits from Bio::SeqFeature::Generic
use Bio::Tools::Blast;
use Bio::Tools::Blast::Run::Webblast;
use Bio::Root::RootI;
use Bio::SeqIO;
@ISA = qw(Bio::Root::Object);


sub new {
    my( $pkg, @args ) = @_;

    my $self = bless {}, $pkg;
    if (@args == 2) {
        if (ref($args[0]) eq "Bio::Seq") {
  #	  print("Adding sequence 1\n");
	    $self->{_seq1} = $args[0];
        } else {
	    $self->throw("ERROR: Arguments must be Bio::Seq");
        }
        if (ref($args[1]) eq "Bio::Seq") {
	    $self->{_seq2} = $args[1];
  #	  print("Adding sequence 2\n");
        } else {
	    $self->throw("ERROR: Arguments must be Bio::Seq");
        }
    }
    elsif (@args) {
        $self->throw("Unexpected arguments [@args]");
    }
    return $self;
}


sub find_overlap {
    my ($self) = @_;

    if (!defined($self->{_seq1})) { $self->throw("ERROR: Sequence 1 not defined");}
    if (!defined($self->{_seq2})) { $self->throw("ERROR: Sequence 2 not defined");}

    # First try head of 2 against all of 1

    print("Sequence 1 is " . $self->{_seq1}->id . " " . length($self->{_seq1}->seq) . "\n");
    print("Sequence 2 is " . $self->{_seq2}->id . " " . length($self->{_seq2}->seq) . "\n");

    my ($file1)       = $self->make_seqfile($self->{_seq1});

    my ($headseq2,$headfile2) = $self->make_head_sequence($self->{_seq2});
    my ($tailseq2,$tailfile2) = $self->make_tail_sequence($self->{_seq2});

    my $headdb = $self->make_database($headfile2);
    my $taildb = $self->make_database($tailfile2);

    my $blast = $self->run_blast($file1,$headfile2,"h");

    my @headhits = $self->filter_hits($blast,0.95);

    my $tophit;
    my $type;
    my $len1;
    my $len2;

    if ($#headhits >= 0) {
	$tophit = $self->find_end_match(@headhits);
	$type = "head";
	$len1 = $self->{_seq1}->length;
	$len2 = $headseq2->length;
    } else {
	$blast = $self->run_blast($file1,$tailfile2,"t");

	my @tailhits = $self->filter_hits($blast,0.95);

	if ($#tailhits >= 0) {
	    $tophit = $self->find_end_match(@tailhits);
	    $type = "tail";
	    $len1 = $self->{_seq1}->length;
	    $len2 = $tailseq2->length;

	}
	
    }

    return unless $tophit;

    my ($orientation,$overlap) = $self->find_orientation($tophit,$type,$len1,$len2);

    return ($orientation,$overlap);
}

sub find_end_match {
    my ($self,@hits) = @_;

    @hits = sort { $b->score <=> $a->score } @hits;

    print("Number of hits found is " . ($#hits+1) . "\n");

    my $tophit = $hits[0];

    return $tophit;

}
sub filter_hits {
    my ($self,$blast,$pid) = @_;
    
    $pid = 0.95 unless $pid;

    my $minlen = 90;

    my @hits;

    return @hits unless (defined($blast));

    foreach my $hit ($blast->hits) {
	foreach my $hsp ($hit->hsps) {
	    my $len = abs($hsp->end('query') - $hsp->start('query')+1);
	    if ($hsp->frac_identical > $pid && $len >= $minlen) {
		push(@hits,$hsp);
	    }

#	    }
	}
    }

    @hits = sort { $b->frac_identical  <=> $a->frac_identical || $a->expect <=> $a->expect } @hits;

    if ($#hits >= 0) {
	printf ("%12s\t %7s\t %12s\t %7s\t %7s\t %7s\t %7s\n", "Expect.","Score","frac_ident","start",
		"end","hstart","hend");
    }
    foreach my $hsp (@hits) {
	    printf "%12.1e\t %7d\t %12.2f\t %7d\t %7d\t %7d\t %7d\n", 
	    $hsp->expect, 
	    $hsp->score, 
	    $hsp->frac_identical, 
	    $hsp->start('query'),
	    $hsp->end  ('query'),
	    $hsp->start('sbjct'),
	    $hsp->end  ('sbjct'),
	}
    return @hits;
}

sub run_blast {
    my ($self,$seq1,$db,$ext) = @_;

    $ext = "h" unless $ext;

    my $blastout = "/tmp/blastend.$$.$ext.blast";
    my $command  = "blastn $db $seq1 B=500 -hspmax 1000 2> /dev/null > $blastout";

#    print ("Running command $command\n");

    my $status = system($command );

#    print("Exit status of blast is $status\n");
    
    my $blast;

    eval {
	$blast = Bio::Tools::Blast->new(-file   => $blastout,
					-parse  => 1,
					-check_all_hits,
#					-signif => 1.0,
					);
    };
    if ($@) {
	print($@);
    }
    return $blast;
} 

    
sub find_orientation {
    my ($self,$hit,$type,$len1,$len2) = @_;

    # First of all find the strandedness of the hit
    my $strand1 = $hit->strand('query');
    my $strand2 = $hit->strand('sbjct');
    
#    print("strand1 is $strand1\n");
#    print("strand2 is $strand2\n");

#    print("len1 is $len1\n");
#    print("len2 is $len2\n");

    my $orient;

    if ($strand1 eq 'Plus' && $strand2 eq 'Plus' ) {
	if ($type eq "head") {
	    $orient = 1;
	} else {
	    $orient = 4;
	}
    } elsif ($strand1 eq 'Plus' && $strand2 eq 'Minus') {
	if ($type eq "head") {
	    $orient = 3;
	} else {
	    $orient = 2;
	}
    } elsif ($strand1 eq 'Minus' && $strand2 eq 'Plus') {
	if ($type eq 'head') {
	    $orient = 3;
	} else {
	    $orient = 2;
	} 
    } elsif ($strand1 eq 'Minus' && $strand2 eq 'Minus') {
	if ($type eq 'head') {
	    $orient = 1;
	} else {
	    $orient = 4;
	}
    }

    print("Orient is $orient\n");
    my $overlap;

    if ($orient == 1) {
	# Case 1.
	# 1) ----------->
	# 2)          --------->
	
	$overlap = $hit->start('query') - ($hit->start('sbjct') - 1);

    } elsif ($orient == 2) {
	# Case 2.
	# 1) ----------->
	# 2)          <---------
	
	$overlap = $hit->start('query') - ($len2 - $hit->end('sbjct'));
	
    } elsif ($orient == 3) {
	# Case 3.
	# 1) <-----------
	# 2)          ---------->
	
	$overlap = $len1 - $hit->end('query') + 1 - ($hit->start('sbjct') - 1);

    } elsif ($orient == 4) {
	# Case 4.
	# 1) <-----------
	# 2)          <---------

 	$overlap = $len1 - $hit->end('query') - ($len2 - $hit->end('sbjct')) + 1;
    }

    return($orient,$overlap);
}
	
	

sub make_head_sequence {
    my ($self,$seq,$chars) = @_;

#    print("Sequence is $self,$seq,$chars\n");

    $chars = 2000 unless $chars;

    my $newstr  = $seq->seq(1,$chars);
    my $newseq  = new Bio::Seq(-seq => $newstr);
       $newseq->id ($seq->id . ".head");
    my $newfile = $self->make_seqfile($newseq);

    return($newseq,$newfile);
}

sub make_tail_sequence {
    my ($self,$seq,$chars) = @_;

    $chars = 2000 unless $chars;

    my $newstr  = $seq->seq($seq->length - $chars + 1);
    my $newseq  = new Bio::Seq(-seq => $newstr);
       $newseq->id ($seq->id . ".tail");
    my $newfile = $self->make_seqfile($newseq);

    return ($newseq,$newfile);
}

sub make_database {
    my ($self,$file) = @_;

    # Make a database out of the sequence
    my $status = system("pressdb $file 2> /dev/null");

#    print("Status of pressdb : $status\n");
    
    return $file;
}

    
sub make_seqfile {
    my ($self,$seq) = @_;

    my $file = "/tmp/blastend_" . $seq->id . ".$$";

    my $seqio = new Bio::SeqIO(-file => ">$file",
			       -format => 'Fasta');

    $seqio->write_seq($seq);
    
    return $file;
}
