package Spangle::BlastWise;

=pod

=head1 NAME
  
Spangle::BlastWise - realigns protein blast hits to genomic sequence

=head1 SYNOPSIS
  
my $blastwise = new  Spangle::BlastWise  ($genefile);           # Initialise the blastwise module

   $blastwise->memory(50000);                                   # Set the memory (in kb) to use
   $blastwise->miniseq($subseq);                                # Set the subsequence to align over
   $blastwise->align_protein($name)                             # Align a protein id $name to the genomic
  

### $blastwise->add_region($start,$end);
### $blastwise->remove_region($start,$end);
### $blastwise->each_region;
### $blastwise->remove_all_regions();

 
my @genes = $blastwise->eachGene;                               # Extract the genes as an array

=head1 DESCRIPTION


=head1 CONTACT

Describe contact details here

=head1 APPENDIX

The rest of the documentation details each of the object methods. Internal methods are usually preceded with a _

=cut

use strict;  
use vars   qw(@ISA);

use Wise2;

# Object preamble - inherits from Bio::Root::RootI

use Bio::Root::RootI;

@ISA = qw(Bio::Root::RootI);

sub new {
  my($pkg, @args) = @_;

  my $self = bless {}, $pkg;

  $self->{_wisegenes}   = [];
  $self->{_genes}       = [];
  $self->{_miniseq}     = undef;
  
  if (@args == 1) {
    $self->_wise_init($args[0]);
  }
  elsif (@args) {
    $self->throw("Unexpected arguments [@args]");
  }

  return $self;
}

sub _wise_init {
  my($self,$genefile) = @_;

  print("Initializing genewise\n");
  
  $self->{_gf}   = &Wise2::read_GeneFrequency21_file($genefile)            || $self->throw( "Could not read $genefile for parameters");
  $self->{_ct}   = &Wise2::CodonTable::read_CodonTable_file("codon.table") || $self->throw("Could not read codon table");
  $self->{_rmd}  = &Wise2::RandomModelDNA_std()                            || $self->throw("Could not make standard random model (weird!)");
  $self->{_alg}  = &Wise2::gwrap_alg_type_from_string("623")               || $self->throw("Could not make algorithm");
  $self->{_gp}   = &Wise2::GeneParameter21_wrap($self->{_gf},0.00001,0.000001,$self->{_rmd},0,1,1,$self->{_ct},0.999,0.99,0.8,0.99,0.4) || $self->throw("Could not make gene parameters");
  
  # sets memory amount for main memory
  $self->memory(100000);  
  
  #
  # these are for the protein part of the comparison
  #
  
  $self->{_cm}   = &Wise2::CompMat::read_Blast_file_CompMat("blosum62.bla");
  $self->{_rm}   = &Wise2::default_RandomModel();
  
}


=pod

=head2 memory

  Title   : memory
  Usage   : $bw->memory(100000);
  Function: Sets the amount of memory to use (kb);
  Returns : Nothing
  Args    : int

=cut

sub memory {
  my($self,$arg) = @_;

  if (defined($arg)) {
     $self->{_memory} = $arg;
     &Wise2::change_max_BaseMatrix_kbytes($arg);
  }
  return $self->{_memory};
}

=pod

=head2 start

  Title   : start
  Usage   : $bw->start($start)
  Function: The start base in the dna to start aligning
  Returns : int
  Args    : int

=cut

sub start {
    my ($self,$arg) = @_;

    if (defined($arg)) {
	$self->{_start} = $arg;
    }
    return $self->{_start};
}


=pod

=head2 end

  Title   : end
  Usage   : $bw->end($end)
  Function: The end base in the dna to start aligning
  Returns : int
  Args    : int

=cut

sub end {
    my ($self,$arg) = @_;

    if (defined($arg)) {
	$self->{_end} = $arg;
    }
    return $self->{_end};
}

 
=pod

=head2 eachGene

  Title   : eachGene
  Usage   : @genes = Spangle::BlastWise->eachGene
  Function: Returns the aligned proteins as genes
  Returns : @Spangle::Subseq
  Args    : none

=cut

sub eachGene {
    my ($self) = @_;

    if (!defined($self->{_genes})) {
	$self->{_genes} = [];
    }
    return @{$self->{_genes}};
}
 
=pod

=head2 deleteGenes

  Title   : deleteGenes
  Usage   : $blastwise->deleteGenes
  Function: Deletes all stored genes.
  Returns : nothing
  Args    : none

=cut

sub deleteGenes {
    my ($self) = @_;

    $self->{_genes}     = [];
    $self->{_wisegenes} = []

}

=pod

=head2 _addWiseGene

  Title   : _addWiseGene
  Usage   : $bw->_addWiseGene (internal use only)
  Function: Adds an aligned protein gene to the object
  Returns : nothing
  Args    : Wise2::umm something

=cut


sub _addWiseGene {
  my ($self,$gene) = @_;

  if (!defined($self->{_wisegenes})) {
      $self->{_wisegenes} = [];
  }

  push(@{$self->{_wisegenes}},$gene);
}

=pod

=head2 _addGene

  Title   : addGene
  Usage   : $blastwise->_addGene($gene)
  Function: Adds a converted wise gene to the real gene array
  Returns : nothing
  Args    : Spangle::Subseq

=cut


sub _addGene {
    my ($self,$gene) = @_;

    if (!defined($self->{_genes})) {
	$self->{_genes} = [];
    }
    push(@{$self->{_genes}},$gene);

}

=head2 set_dna_seq

  Title   : set_dna_seq
  Usage   : Spangle::BlastWise->set_dna_seq($dnastr);
  Function: Generates a wise genomic sequence
  Returns : nothing
  Args    : String

=cut

sub set_dna_seq {
  my ($self,$dnastr) = @_;

  my $tmpseq = "/tmp/pog$$.fa";

  print(STDERR "Length of dna is " . length($dnastr) . "\n");

  open(DNA,">$tmpseq");
  print(DNA ">tmpid\n");
  
  $dnastr =~ tr/a-z/A-Z/;
  $dnastr =~ s/(.{72})/$1\n/g;

  print(DNA "$dnastr\n");
  
  close(DNA);
  
  open(DNA,"<$tmpseq");

  undef $self->{_seq};
  undef $self->{_gen};
  undef $self->{_mother};

  $self->{_seq} = &Wise2::Sequence::read_fasta_Sequence(\*DNA);
  $self->{_gen} = &Wise2::Genomic::Genomic_from_Sequence($self->{_seq});
  
  close(DNA);
  
  unlink $tmpseq;

  if( $self->{_gen} == 0 ) {
    $self->throw("Could not genomic from sequence %s!",$self->{_seq}->name());
  }
  
  $self->{_mother} = &Wise2::GenomicRegion::new_GenomicRegion($self->{_gen});
}

=pod


=head2 miniseq

  Title   : miniseq
  Usage   : Spangle::BlastWise->miniseq($subseq);
  Function: A subsequence of a longer piece of dna
            (which can be several regions) which
            restricts the areas to realign the protein to
  Returns : Spangle::Subseq
  Args    : Spangle::Subseq

=cut

sub miniseq {
    my ($self,$subseq) = @_;
    
    if (defined($subseq)) {
	$self->{_miniseq} = $subseq;

	my $start = 0;
	my $end   = $self->{_miniseq}->seq->length - 1;
    
	$self->start($start);
	$self->end  ($end);

	my $seq = $subseq->dna_seq || $self->throw("No dna sequence defined in miniseq $subseq");
	my $dna = $seq   ->seq     || $self->throw("Can't generate dna sequence for miniseq $subseq");

	$self->set_dna_seq($dna);
    }
    return $self->{_miniseq};
}
    

=pod

=head2 align_protein

  Title   : align_protein
  Usage   : $bw->align_protein($pro,$start,$end)
  Function: aligns a protein to a genomic sequence between the genomic coords $start,$end
  Returns : nothing
  Args    : Bio:Seq,int,int

=cut


sub align_protein {
    my ($self,$name) = @_;

    my $pro = $self->getseq($name);
    
    my $pgl = new Wise2::PotentialGeneList; # empty list.
    my $pg  = new Wise2::PotentialGene;     # empty potential gene

    $pgl->add_pg($pg);
    $pg ->set_homolog($pro);

    $pg->set_guess_start($self->start);
    $pg->set_guess_end  ($self->end);
    
    &Wise2::resolve_PotentialGenes_on_GenomicRegion($self->{_mother},
						    $pgl,
						    $self->{_alg},
						    $self->{_alg},
						    20,
						    20,
						    $self->{_cm},
						    -12,
						    -2,
						    $self->{_gp},
						    $self->{_rmd},
						    $self->{_rmd},
						    " ",0,25);
    
    
    foreach my $gene ( $self->{_mother}->each_gene() ) {
	$self->_addWiseGene($gene);
    }

    my @newgenes = $self->genes2subseq();
    my @exgenes  = $self->expand_genes(@newgenes);

    foreach my $g (@exgenes) {
	$self->_addGene($g);
    }

}


=pod

=head2 genes2ace

  Title   : genes2ace
  Usage   : $bw->genes2ace
  Function: Writes the genes as ace objects
  Returns : @string
  Args    : idstring,methodstring

=cut

sub genes2ace {
  my($self,$id,$method) = @_;

  my $tmpace = "/tmp/pogace$$.ace";
  open(ACE,">$tmpace");

  my $fh = *ACE;
  
  $self->{_mother}->show_pretty_GenomicRegion(*STDOUT);
  $self->{_mother}->show_ace_GenomicRegion($self->{_gen}->baseseq->name(),*ACE);

  close(ACE);

  open(ACE,"<$tmpace");

  my @out = <ACE>;

  foreach my $line (@out) {
    $line =~ s/No_Id_Given\./$id.$method./;
    $line =~ s/No_Id_Given/$id/;
  }

  unlink $tmpace;

  return @out;
}

=head2 genes2subseq

  Title   : genes2subseq
  Usage   : $bw->genes2subseq
  Function: Returns the genes as subsequences
  Returns : @Spangle::Subseq
  Args    : None

=cut

sub genes2subseq {
  my($self) = @_;

  my $tmpsub = "/tmp/pogsub$$.ace";
  open(SUB,">$tmpsub");

  my $fh = *SUB;
  
  $self->{_mother}->show_pretty_GenomicRegion(*SUB);

  close(SUB);

  open(SUB,"<$tmpsub");

  my @out = <SUB>;

  my $subseq;
  my $strand;
  my @genes;

  foreach my $line (@out) {

    if ($line =~ /Gene +(\d+) +(\d+)/) {

       if ($2 > $1) {
         $strand = 1;
       }  else {
         $strand = -1;
       }
       if (ref($subseq) eq "Spangle::Subseq") {
           push(@genes,$subseq);
       }
       
       $subseq = Spangle::Subseq->new;
    }
    if ($line =~ /Exon +(\d+) +(\d+)/) {
        my $ex = Spangle::Exon->new;
        if ($1 > $2) {
        $ex->start ($2);
        $ex->end   ($1);
        } else {
        $ex->start ($1);
        $ex->end   ($2);
        }
        $ex->strand($strand);

        $subseq->add_Exon($ex);
    }
  }
  if (ref($subseq) eq "Spangle::Subseq") {
    push(@genes,$subseq);
  }

  unlink $tmpsub;

  return(@genes);


}

=pod

=head2 getseq

  Title   : getseq
  Usage   : $bw->getseq($id)
  Function: Generates a genewise protein object from a protein id
  Returns : nothing
  Args    : string

=cut


sub getseq {
  my($self,$name) = @_;
  local *PRO;

  my $pro;
  print "Shell command test:\n", qx{echo "OK"};

  print(STDERR "fetching $name\n");

  my $tmpseq = "/tmp/pro$$.fa";

  system("efetch -f $name > /tmp/pro$$.fa");
  open (PRO,"<$tmpseq");

  my $seq = &Wise2::Sequence::read_fasta_Sequence(\*PRO);

  close(PRO)    || $self->throw("Error from efetch\n");
  defined($seq) || $self->throw("Could not efetch '$name'");

  unlink $tmpseq;

  if( $seq ) {
    $pro = &Wise2::Protein::Protein_from_Sequence($seq);
  } else {
    warn("Could not read $name");
    next;
  }
  return $pro;
}


=pod

=head2 expand_genes

  Title   : expand_genes
  Usage   : @newgenes = $bw->expand_genes
  Function: Expands genes on the subseq to gene on the genomic
  Returns : @Spangle::Subseq
  Args    : none

=cut

sub expand_genes {
    my ($self,@genes)  = @_;

    my @newgenes;
    
    foreach my $gene (@genes) {
	my @gex  = $gene->each_Exon;
	my $noex = $#gex;
	
	if ($noex > 0) {
	    $gene = $self->convert_coords($gene);
	    push(@newgenes,$gene);
	}
    }
    return @newgenes;
}


=pod

=head2 convert_coords

  Title   : convert_coords
  Usage   : my $newgene = $bw->convert_coords($gene)
  Function: Converts gene coordinates on the subseq to
            gene coordinates on the genomic
  Returns : Spangle::Subseq
  Args    : Spangle::Subseq

=cut

sub convert_coords {
    my ($self,$gene) = @_;

    my $subseq = $self->miniseq || $self->throw("No miniseq defined");

    my @exons = sort { $a->start <=> $b->start } $gene->each_Exon;

    my $newgene = Spangle::Subseq->new;
  
    foreach my $ex (@exons) {
#	print("Converting exon " . $ex->start . "\t" . $ex->end . "\n");
	my @newex = $self->find_exon_coord($ex);
	
	foreach my $newex (@newex) {
	    $newex  ->strand($subseq->strand);
	    $newgene->add_Exon($newex);
	}
    }
    
    $newgene->sort;
    
    return $newgene;
}


=pod

=head2 find_exon_coord

  Title   : find_exon_coord
  Usage   : my @newexons = $bw->find_exon_coord($exon);
  Function: Converts an exon hit to the subseq dna to an 
            array of hits to the genomic dna.
  Returns : @Bio::EnsEMBL::Exon
  Args    : Bio::EnsEMBL::Exon

=cut

sub find_exon_coord {
    my ($self,$exon) = @_;
    
    if($exon->start > $exon->end) {
	my $tmp = $exon->end;
	$exon->end  ($exon->start);
	$exon->start($tmp);
    }
    
    my $start = $exon->start;
    my $end   = $exon->end;
    
    my @out;
    my @exons = $self->miniseq->each_Exon;
    
    if ($self->miniseq->strand == 1) {
	@exons = sort {$a->start <=> $b->start} @exons;
	
	my $foundstart = 0;
	my $foundend   = 0;
	
	my $count = 0;
	my $dna   = $exons[0]->start - 1;
	
	foreach my $ex (@exons) {
#	    print("Looking in exon " . $ex->start . "\t" . $ex->end . "\n");
	    my $tmpstart;
	    my $tmpend;
	    
#	    print("Transform = " . ($ex->start - $dna)  . "\t" . ($ex->end-$dna) . "\n");
	    
	    if ($foundstart == 0 ) {
		if ($start >= ($ex->start - $dna) && $start <= ( $ex->end - $dna)) {
		    $foundstart = 1;
		    $tmpstart = $dna + $start;
#		    print("Found start $tmpstart\n");
		}
	    }
	    
	    if (!$foundend && $foundstart) {
		$tmpstart = $ex->start unless $tmpstart;
		$tmpend   = $ex->end;
		
		# Before making a new exon check we haven't also
		# Got the end point in the same exon
		
		if ($end <= $ex->end - $dna ) {
		    $tmpend = $end + $dna;
		    $foundend = 1;
#		    print("Found end $tmpend\n");
		}
		
	    }
	    
	    if (defined($tmpstart) && defined($tmpend)) {
#		print("Creating exon $tmpstart - $tmpend\n");
		my $newex = Spangle::Exon->new;
		$newex->start ($tmpstart);
		$newex->end   ($tmpend);
		push(@out,$newex);
		
		if ($foundend == 1) {
		    return @out;
		}
		
	    }
	    
	    $dna = $dna + $exons[$count+1]->start - $ex->end - 1;
	    $count++;
	}
	return @out; 
    } else {
	
	@exons = sort {$b->start <=> $a->start} @exons;
	
	my $foundstart = 0;
	my $foundend   = 0;
	
	my $count = 0;
	my $dna   = $exons[0]->end;
	

	foreach my $ex (@exons) {
#	    print ("DNA is $dna\n");
#	    print("Looking in exon " . $ex->start . "\t" . $ex->end . "\n");
	    my $tmpstart;
	    my $tmpend;
	    
#	    print("Transform = " . ($dna - $ex->start+1)  . "\t" . ($dna - $ex->end+1) . "\n");
	    
	    if ($foundstart == 0 ) {
		if ($start >= ($dna - $ex->end) && $start <= ($dna-$ex->start)) {
		    $foundstart = 1;
		    $tmpstart = $dna - $start + 1;
#		    print("Found start $tmpstart\n");
		}
	    }
	    
	    if (!$foundend && $foundstart) {
		$tmpstart = $ex->end unless $tmpstart;   ### is this right?
		$tmpend   = $ex->start;
		
		# Before making a new exon check we haven't also
		# Got the end point in the same exon
		
		if ($end <= ($dna - $ex->start)) {
		    $tmpend = $dna - $end + 1;
		    $foundend = 1;
#		    print("Found end $tmpend\n");
		}
		
	    }
	    
	    if (defined($tmpstart) && defined($tmpend)) {
#		print("Creating exon $tmpstart - $tmpend\n");
		my $newex = Spangle::Exon->new;
		$newex->start ($tmpend);
		$newex->end   ($tmpstart);
		push(@out,$newex);
		
		if ($foundend == 1) {
		    return @out;
		}
		
	    }
	    
	    if ($count < $#exons) {
		$dna = $dna - ($ex->start - $exons[$count+1]->end) + 1;
	    }

	    $count++;
	}
	return @out; 
    }
}

1;







