package Spangle::GD;

use GD;
use strict;

=pod 

=head1 NAME

Spangle::GD - Utility routines for GD gif generation module

=head1 SYNOPSIS

  use Spangle::GD

Generate some colours to use in the gifs

  my $col1 = Spangle::GD::colour(255,0  ,0);
  my $col2 = Spangle::GD::colour(255,255,255);

Draw a coloured line on an image

  my $im   = GD::Image->new(200,200);
  my $col1 = Spangle::GD::colour(255,255,100);

  Spangle::GD::draw_line($im,$col1,10,10,100,100);
  
Generate a GD image that has a gradient between $col1 and $col2

  my $dir    = 1    #  Gradient in vertical direction (0 for horiz)
  my $width  = 100; #  Width of image
  my $height = 30;  #  Height of image

  my $gif1 = Spangle::GD::draw_gradient_box($col1,$col2,$width,$height,$dir);

Now draw this image on a bigger image
  my $im = new GD::Image(200,200);

  $im->setBrush($gif1);
  $im->line(30,100,30,100,gdBrushed);

And print

  binmode STDOUT;

  print $im->gif;


=head1 Methods in Spangle::GD

=head2 draw_gradient_box

  Title   : draw_gradient_box
  Usage   : $im = Spangle::GD::draw_gradient_box($col1,$col2,$width,$height,$dir)
  Function: Generates a filled rectangle with a colour gradient from col1 to col2
  Returns : GD::Image
  Args    : Spangle::GD::color,Spangle::GD::color,int,int,0 or 1

=cut

sub draw_gradient_box {
  my ($col1,$col2,$width,$height,$dir) = @_;

  my $im     = new GD::Image($width,$height);
  
  # We create a brush with the gradient in the
  # direction specified by $dir
  #
  # $dir = 1 : vertical gradient
  # $dir = 0 : horizontal gradient

  my $len;
  if ($dir == 1) {
    $len = $height;
  } else {
    $len = $width;
  }

  my $brush;
  if ($dir == 1) {
    $brush = new GD::Image(1,     $height);
  } else {
    $brush = new GD::Image($width,1      );
  }

  for (my $i = 0; $i < $len; $i++) {
    # Change the colour smoothly from $col1 to $col2
    
    my $red   = int ($col1->[0]   + $i * ($col2->[0] - $col1->[0]) / $len);
    my $green = int ($col1->[1]   + $i * ($col2->[1] - $col1->[1]) / $len);
    my $blue  = int ($col1->[2]   + $i * ($col2->[2] - $col1->[2]) / $len);

    my $col = $brush->colorAllocate($red,$green,$blue);

    if ($dir == 1) {
      $brush->line(0 ,$i,1   ,$i+1,$col);
    } else {
      $brush->line($i,0 ,$i+1,1   ,$col);
    }

  }

#  return $brush->gif;

  $im->setBrush($brush);
  if ($dir == 1) {
    $im->line(0     ,$height/2,
	      $width,$height/2,
	      gdBrushed);
  } else {
    $im->line($width/2,0,
	      $width/2,$height,
	      gdBrushed);
  }
  return $im;

}


=pod

=head2 draw_line

  Title   : draw_line
  Usage   : Spangle::GD::draw_line($im,$col,$x1,$y1,$x2,$y2)
  Function: Draws a coloured line on an image
  Returns : nothing
  Args    : GD::Image , Spangle::GD::colour,int,int,int,int

=cut

sub draw_line {
  my  ($im,$col,$x1,$y1,$x2,$y2) = @_;

  # test whether colour is allocated - and allocate if nec.
  my $colindex = $im->colorExact($col->[0],$col->[1],$col->[2]);
  print(STDERR "$colindex\n");
  if ($colindex == -1) {
    $colindex = $im->colorAllocate($col->[0],$col->[1],$col->[2]);
  }

  $im->line($x1,$y1,$x2,$y2,$colindex);
}

=pod

=head2 colour

  Title   : colour
  Usage   : $col = Spangle::GD::colour(255,52,100);
  Function: Sets up a colour array ref 
  Returns : Array ref of red,green,blue
  Args    : int,int,int (0-255)

=cut

sub colour {
  my ($red,$green,$blue) = @_;

  my @col = ($red,$green,$blue);
  return \@col;

}

=pod

=head2 colours

  Title   : colours
  Usage   : $col = Spangle::GD::colours($im);
  Function: Allocates a standard set of colours for a GD image
  Returns : ref to hash keyed by colour names
  Args    : GD::Image

=cut

sub colours {
  my ($im) = @_;
  my %colhash;

  $colhash{white}    = $im->colorAllocate(255, 255, 255);        
  $colhash{black}    = $im->colorAllocate(0  , 0  , 0  );
  $colhash{red}      = $im->colorAllocate(255, 0  , 0  );      
  $colhash{green}    = $im->colorAllocate(0  , 255, 0  );
  $colhash{yellow}   = $im->colorAllocate(255, 250, 205);
  $colhash{blue}     = $im->colorAllocate(0  , 0  , 255);

  return \%colhash;

}

1;
