#
# BioPerl module for Exon
#
# Cared for by Ewan Birney <birney@sanger.ac.uk>
#
# Copyright Ewan Birney
#
# You may distribute this module under the same terms as perl itself

# POD documentation - main docs before the code

=pod 

=head1 NAME

Bio::EnsEMBL::Homol - Homology feature

=head1 SYNOPSIS

    $ex = new Bio::EnsEMBL::Homol;

    $ex->start(10);
    $ex->end(100);
    $ex->hstart(5);
    $ex->hend(35);

=head1 DESCRIPTION

Homology feature object - to store results of blast searches, pfam hits
and the like.

=head1 CONTACT

Describe contact details here

=head1 APPENDIX

The rest of the documentation details each of the object methods. Internal methods are usually preceded with a _

=cut


# Let the code begin...


package Spangle::Homol;

use vars qw(@ISA);
use strict;

# Object preamble - inherits from Bio::SeqFeature::Generic

use Bio::SeqFeature::Generic;

@ISA = qw(Bio::SeqFeature::Generic);

sub new {
    my( $pkg, @args ) = @_;

    my $self = $pkg->SUPER::new(@args);

    # add in EnsEMBL tag as 1.

    $self->primary_tag('homol');
    $self->source_tag('EnsEMBL');

    my ($id,$hstart,$hend) = $self->_rearrange([qw(ID
					       HSTART
					       HEND
					      )],@args);

    $self->_parse($id,$hstart,$hend);

    return $self;
}

# Parse routine called from the constructor to
# set the basic variables start,end and strand

sub _parse {
  my ($self,$id,$hstart,$hend) = @_;

  $self->id($id);

  # Swap start and end if they're in the wrong order
  # We assume that the strand is correct and keep the input value.
  if ($self->start > $self->end) {
    my $tmp = $self->end;

    $self->end  ($self->start);
    $self->start($tmp);
  }

  if ($hstart > $hend) {
    my $tmp = $hend;
    $hend    = $hstart;
    $hstart  = $tmp;
  }

  # New methods for the Homol object
  $self->hstart ($hstart);
  $self->hend   ($hend);

}

=pod 

=head1 Methods unique to Homol

=head2 id

 Title   : id
 Usage   : $id = $feat->id
 Function: Returns the hit id
 Returns : string
 Args    : none

=cut

sub id {
  my ($self,$value) = @_;

  if (defined($value)) {
    $self->{_id} = $value;
  }

  return $self->{_id};
}

=head2 title

 Title   : title
 Usage   : $title = $feat->title
 Function: Returns the hit title
 Returns : string
 Args    : none

=cut

sub title {
  my ($self,$value) = @_;

  if (defined($value)) {
    $self->{_title} = $value;
  }

  return $self->{_title};
}

=head2 hstart

 Title   : hstart
 Usage   : $hstart = $feat->hstart
 Function: Returns the hit start coordinate of the feature
 Returns : integer
 Args    : none

=cut

sub hstart {
  my ($self,$value) = @_;

  if (defined($value)) {
    $self->{hstart} = $value;
  }

  return $self->{hstart};
}

=head2 hend

 Title   : hend
 Usage   : $hend = $feat->hend
 Function: Returns the hit end coordinate of the feature
 Returns : integer
 Args    : none

=cut

sub hend {
  my ($self,$value) = @_;

  if (defined($value)) {
    $self->{hend} = $value;
  }

  return $self->{hend};
}

=head2 frame

 Title   : frame
 Usage   : $feat->frame($frame)
 Function: Get/set the frame of the homol.
 Returns : integer (0,1,2)
 Args    : 0,1,2

=cut

sub frame {
  my ($self,$value) = @_;

  if (defined($value)) {
    $self->{frame} = $value;
  }

  return $self->{frame};
}

=head2 length

 Title   : length
 Usage   : $l = $h->length
 Function: Gets/sets the length of the homol.
 Returns : int
 Args    : none or int

=cut

sub length {
  my ($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_length} = $arg;
  }

  return $self->{_length};
}

=head2 score_per_base

 Title   : score_per_base
 Usage   : $score = $homol->score_per_base
 Function: Returns the average score per base
 Returns : int
 Args    : none

=cut
  
sub score_per_base {
  my ($self) = @_;

  my $sc = $self->score;

  $sc = $sc/($self->end - $self->start + 1);
 
  return $sc;

}


=head2 splice

 Title   : splice
 Usage   : @splice = $homol->splice
 Function: Decides whether the homol is spliced.
 Returns : int,int
 Args    : none

=cut
  
sub splice {
  my ($self) = @_;
  
  my $len = $self->length;
  
  $len = $self->hend unless $len;

  # We have found the start/end of the seuqnece
  # if we get to within $tol bases.

  my $tol = 20;

  my $lsplice = 0;
  my $rsplice = 0;
  
  # Forward strand
  if ($self->strand == 1) {
    if ($self->hstart >= $tol) {
      $lsplice = 1;
    }
    if ($self->hend <= ($len - $tol)) {
      $rsplice = 1;
    } 
  }

  # Reverse strand
  if ($self->strand == -1) {
    if ($self->hend <= ($len -  $tol)) {
      $lsplice = 1;
    } 
    if ($self->hstart >= $tol) {
      $rsplice = 1;
    }
  }
  
  return($lsplice,$rsplice);

}    


=head2 match

 Title   : match
 Usage   : @overlap = $feat->match($f)
 Function: Returns an array of 3 numbers detailing
           the overlap between the two features.
           $overlap[0]  = 1 if any overlap, 0 otherwise
           The remaining elements of the array are only set if $overlap[0] = 1
           $overlap[1]  = left hand overlap (-ve if $f starts within $self, +ve if outside)
           $overlap[2]  = right hand overlap (-ve if $f ends within $self, $+ve if outside)
 Returns : @int
 Args    : Bio::SeqFeature::Generic

=cut

sub match {
  my ($f1,$f2) = @_;
  
  my ($start1,
      $start2,
      $end1,
      $end2,
      $rev1,
      $rev2,
     );

  # Swap the coords round if necessary
  if ($f1->start > $f1->end) {
    $start1 = $f1->end;
    $end1   = $f1->start;
    $rev1   = 1;
  } else {
    $start1 = $f1->start;
    $end1   = $f1->end;
  }

  if ($f2->start > $f2->end) {
    $start2 = $f2->end;
    $end2   = $f2->start;
    $rev2   = 1;
  } else {
    $start2 = $f2->start;
    $end2   = $f2->end;
  }

  # Now check for an overlap
  if (($end2 > $start1 && $start2 < $end1) ||
	($start2 < $end1 && $end2 > $start1)) {
	
	#  we have an overlap so we now need to return 
	#  two numbers reflecting how accurate the span 
	#  is. 
	#  0,0 means an exact match with the exon
	# a positive number means an over match to the exon
	# a negative number means not all the exon bases were matched

	my $left = ($start2 - $start1);
	my $right = ($end1 - $end2);
	
	if ($rev1) {
	    my $tmp = $left;
	    $left = $right;
	    $right = $tmp;
	}
	
	my @overlap;

	push (@overlap,1);

	push (@overlap,$left);
	push (@overlap,$right);

	return @overlap;
      }
    
}


=head2 test_overlap

 Title   : test_overlap
 Usage   : ($lo,$ro) = $h->test_overlap($ex)
 Function: Tests the validity of the overlap between
           the homol and another feature
 Returns : int,int
 Args    : Bio::SeqFeature

=cut

sub test_overlap {
    my ($self,$f,$tol) = @_;
 
    $tol = 20 unless $tol;
    
    my @splice = $self->splice;
    my $seqlen = $self->length;
    
    my @overlap = $self->match($f);
    
    my $lo = 0;
    my $ro = 0;
    
    # Test the left overlap

    if ($overlap[0]) {
	if (abs($overlap[1]) < $tol) {                                # the ends match to within tol
	    $lo = 1;
	} elsif ($splice[0] == 0 && $overlap[2] > 0) {                # we haven't found a left splice site
	    $lo = 1;
	} elsif ($overlap[1] < 0 && $self->strand == 1   &&   $self->hstart <= 20 ) { # we have found the start of the seq
	    $lo = 1;
	} elsif ($overlap[1] < 0 && $self->strand == -1  &&  $seqlen <= ($self->hend+20)) {  # we have found the end of the seq
	    $lo = 1;
	}
	#Test the right overlap 
	if (abs($overlap[2]) < ($tol)) {                              # the ends match to within $tol
	    $ro = 1;
	} elsif ($splice[1] == 0 && $overlap[2] > 0 ) {               # we haven't found a splice site yet
	    $ro = 1;
	} elsif ($overlap[2] < 0 && $self->strand == 1 &&  $seqlen <= ($self->hend+20)  ) {  # we have found the end of the seq.
	    $ro = 1;
	} elsif ($overlap[2] < 0 && $self->strand == -1 && $self->hstart <= 20) {            # we have found the start of the seq.
	    $ro = 1;
	} 
    }
    return ($lo,$ro,@overlap);
} 

=head2 add_supported_feature

 Title   : add_supported_feature
 Usage   : $feat->add_supported_feature
 Function: Add a feature (eg Exon) that this feature supports
 Returns : nothing
 Args    : Bio::SeqFeature::Generic

=cut

sub add_supported_feature {
  my($self,$f) = @_;

  if (!defined($self->{_supported_feature})) {
    $self->{_supported_feature} = [];
  }

  push(@{$self->{_supported_feature}},$f);
}

=head2 each_supported_feature

 Title   : each_supported_feature
 Usage   : @f = $feat->each_supported_feature
 Function: Returns all features that this feature supports
 Returns : @Bio::SeqFeature::Generic
 Args    : none

=cut

sub each_supported_feature {
  my($self) = @_;

  if (!defined($self->{_supported_feature})) {
    $self->{_supported_feature} = [];
  }
  
  return @{$self->{_supported_feature}};
}

=head2 toString

 Title   : toString
 Usage   : $fstr = $feat->toString($str)
 Function: Returns the homol as string delimited by $str
 Returns : string
 Args    : string

=cut

sub toString {
  my ($self,$tab) = @_;

  my $str = "";

  $tab = "\t" unless $tab;

  $str= $self->id     . $tab .
        $self->start  . $tab . 
        $self->end    . $tab . 
        $self->hstart . $tab . 
	$self->hend   . $tab .
	$self->strand . $tab .
        $self->source_tag  . $tab . 
        $self->primary_tag . $tab;

  return $str;
}



# rest is left until we know what is going on...
# but you do have all the SeqFeature documentation: reproduced here
# for convience...

=pod

=head1 Methods inherited from SeqFeature

=head2 start

 Title   : start
 Usage   : $start = $feat->start
 Function: Returns the start coordinate of the feature
 Returns : integer
 Args    : none

=cut


=pod

=head2 end

 Title   : end
 Usage   : $end = $feat->end
 Function: Returns the end coordinate of the feature
 Returns : integer
 Args    : none

=cut


=pod

=head2 strand

 Title   : strand
 Usage   : $strand = $feat->strand()
 Function: Returns strand information, being 1,-1 or 0
 Returns : -1,1 or 0
 Args    : none

=cut

sub strand {
  my ($self,$value) = @_;

   if($value)  {
      if ($value eq "+") {$value =  1;}
      if ($value eq "-") {$value = -1;}
      if ($value eq ".") {$value =  0;}

      $self->{'strand'} = $value;
    }

    return $self->{'strand'};
}

=pod

=head2 sub_SeqFeature

 Title   : sub_SeqFeature
 Usage   : @feats = $feat->sub_SeqFeature();
 Function: Returns an array of sub Sequence Features
 Returns : An array
 Args    : none

=cut

=pod

=head2 primary_tag

 Title   : primary_tag
 Usage   : $tag = $feat->primary_tag()
 Function: Returns the primary tag for a feature,
           eg 'exon'
 Returns : a string 
 Args    : none

=cut

=pod

=head2 source_tag

 Title   : source_tag
 Usage   : $tag = $feat->source_tag()
 Function: Returns the source tag for a feature,
           eg, 'genscan' 
 Returns : a string 
 Args    : none

=cut

=pod

=head2 has_tag

 Title   : has_tag
 Usage   : $value = $self->has_tag('some_tag')
 Function: Returns the value of the tag (undef if 
           none)
 Returns : 
 Args    :

=head2 all_tags

 Title   : all_tags
 Usage   : @tags = $feat->all_tags()
 Function: gives all tags for this feature
 Returns : an array of strings
 Args    : none



=head2 gff_string

 Title   : gff_string
 Usage   : $str = $feat->gff_string
 Function: provides the feature information in GFF
           version 2 format.
 Returns : A string
 Args    : None



=head1 RangeI methods

These methods are inherited from RangeI and can be used
directly from a SeqFeatureI interface. Remember that a 
SeqFeature is-a RangeI, and so wherever you see RangeI you
can use a feature ($r in the below documentation).

=head2 overlaps

  Title   : overlaps
  Usage   : if($feat->overlaps($r)) { do stuff }
            if($feat->overlaps(200)) { do stuff }
  Function: tests if $feat overlaps $r
  Args    : a RangeI to test for overlap with, or a point
  Returns : true if the Range overlaps with the feature, false otherwise


=head2 contains

  Title   : contains
  Usage   : if($feat->contains($r) { do stuff }
  Function: tests whether $feat totally contains $r
  Args    : a RangeI to test for being contained
  Returns : true if the argument is totaly contained within this range


=head2 equals

  Title   : equals
  Usage   : if($feat->equals($r))
  Function: test whether $feat has the same start, end, strand as $r
  Args    : a RangeI to test for equality
  Returns : true if they are describing the same range


=head1 Geometrical methods

These methods do things to the geometry of ranges, and return
triplets (start, stop, strand) from which new ranges could be built.

=head2

  Title   : intersection
  Usage   : ($start, $stop, $strand) = $feat->intersection($r)
  Function: gives the range that is contained by both ranges
  Args    : a RangeI to compare this one to
  Returns : nothing if they don't overlap, or the range that they do overlap


=head2 union

  Title   : union
  Usage   : ($start, $stop, $strand) = $feat->union($r);
          : ($start, $stop, $strand) = Bio::RangeI->union(@ranges);
  Function: finds the minimal range that contains all of the ranges
  Args    : a range or list of ranges to find the union of
  Returns : the range containing all of the ranges

=cut

1;
