package Spangle::HomolSet;

=pod

  =head1 NAME
  
  Spangle::HomolSet - a set of possibly overlapping database hits

  =head1 SYNOPSIS
  
  $cl = new Spangle::HomolSet
  
=head1 DESCRIPTION

  Stores and constructs possible new genes out of database hits.

=head1 CONTACT

  Describe contact details here

=head1 APPENDIX

  The rest of the documentation details each of the object methods. Internal methods are usually preceded with a _

=cut

use strict;  
use vars   qw(@ISA);

# Object preamble - inherits from Bio::Root::RootI
use Spangle::Homol;
use Spangle::Exon;
use Spangle::SeqLength_DB;   # used to fetch database entry lengths

use Bio::Root::RootI;

@ISA = qw(Bio::Root::RootI);

sub new {
    my( $pkg, @args ) = @_;

    my $self = bless {}, $pkg;
    $self->{_homols} = {};
    $self->{_cluster} = undef;
    $self->throw("Unexpected arguments [@args]") if @args;
    return $self;
}

sub contains {
  my ($self,$id) = @_;

  $id =~ tr/a-z/A-Z/;

  if ($self->{_homols}{$id}) {
     return 1;
  }
  return 0;
}

=pod

=head2 each_cluster

  Title   : each_cluster
  Usage   : @cl = $homolset->each_cluster
  Function: Returns an array of clusters
  Returns : @Bio::SeqFeature::Generic
  Args    : none

=cut

sub each_cluster {
  my $self = shift;

  if (defined($self->{_cluster})) {
    return $self->{_cluster}->sub_SeqFeature;
  }
}

=pod

=head2 each_feature

  Title   : each_feature
  Usage   : @f = $homolset->each_feature
  Function: Returns all homology features in the set
  Returns : @Spangle::Homol
  Args    : none

=cut

sub each_feature {
  my $self = shift;
  my @arr;

  if (defined($self->{_homols})) {

      foreach my $id (keys %{$self->{_homols}}) {
	  push(@arr,@{$self->{_homols}{$id}});
      }
      
      return @arr;
  }
  
}

=pod

=head2 each_feature_by_ID

  Title   : each_feature_by_ID
  Usage   : @f = $homolset->each_feature_by_ID
  Function: Returns all homology features in the set keyed by id.
  Returns : Reference to a hash of @Homol keyed by Homol->id
  Args    : none

=cut

sub each_feature_by_ID {
  my $self = shift;
  
   return $self->{_homols};
  
}
=pod

=head2 each_ID

  Title   : each_ID
  Usage   : @id = $homolset->each_ID
  Function: Returns all the ids of the Homol features
  Returns : @string
  Args    : none

=cut

sub each_ID {
  my $self = shift;
  
  if (defined($self->{_homols})) {
   return keys %{$self->{_homols}};
}
  
}

=pod

=head2 sort

  Title   : sort
  Usage   : $homolset->sort
  Function: Sorts all homols by start coordinate
  Returns : @Homol
  Args    : none

=cut

sub sort {
  my $self = shift;

  my @arr = $self->each_feature;

  @arr = sort { $a->start <=> $b->start } @arr;

  return @arr;
}

=pod

=head2 add_homol

  Title   : add_homol
  Usage   : $homolset->add_homol
  Function: Adds a homology feature into the set.
  Returns : nothing
  Args    : none

=cut

sub add_homol {
  my ($self,$h) = @_;
  
  push(@{$self->{_homols}{$h->id}},$h);


  if ($self->{_count}{$h->id}) {
    $self->{_count}{$h->id}++;
  } else {
    $self->{_count}{$h->id} = 1;
  }

#  print("Count is " . $self->{_count}{$h->id} . " for " . $h->id . "\n");
}  

=pod

=head2 add_feature_array

  Title   : add_feature_array
  Usage   : $homolset->add_feature_array(@f);
  Function: Adds a homology feature array into the set.
  Returns : nothing
  Args    : none

=cut

sub add_feature_array {
  my ($self,@f) = @_;

  foreach my $h (@f) {
    $self->add_homol($h);
  }
}


=pod

=head2 merge

  Title   : merge
  Usage   : $homolset->merge
  Function: Merges two or more homol features into one if they are close enough together
  Returns : nothing
  Args    : none

=cut

sub merge {
  my ($self,$overlap,$query_gap,$homol_gap) = @_;
  
  $overlap   = 20  unless $overlap;
  $query_gap = 15  unless $query_gap;
  $homol_gap = 15  unless $homol_gap;
  
  
  delete $self->{_count};
  
  foreach my $id (keys %{$self->{_homols}}) {
    
    my $count = 0;
    my @newfeatures;
    my @features = @{$self->{_homols}{$id}};
    
    @features = sort { $a->start <=> $b->start} @features;

    # put the first feature in the new array;
    push(@newfeatures,$features[0]);
    
    for (my $i=0; $i < $#features; $i++) {
      my $id  = $features[$i]  ->id;
      my $id2 = $features[$i+1]->id;
      
      # First case is if start of next hit is < end of previous
      if ($features[$i]->end > $features[$i+1]->start && 
	 ($features[$i]->end - $features[$i+1]->start) < $overlap) {
	
	if ($features[$i]->strand eq "1") {
	  $newfeatures[$count]-> end($features[$i+1]->end);
	  $newfeatures[$count]->hend($features[$i+1]->hend);
	} else {
	  $newfeatures[$count]-> end($features[$i+1]->end);
	  $newfeatures[$count]->hend($features[$i+1]->hstart);
	}
	
	# add the scores - is this the right thing for BLAST?
	$newfeatures[$count]->score($newfeatures[$count]->score + $features[$i+1]->score);
	
	if ($features[$i+1]->hstart == $features[$i+1]->hend) {
	  $features[$i+1]->strand($features[$i]->strand);
	}
	
	# Allow a small gap if < $query_gap, $homol_gap
      } elsif (($features[$i]->end < $features[$i+1]->start) &&
	       abs($features[$i+1]->start - $features[$i]->end) <= $query_gap) {
	
	if ($features[$i]->strand eq "1") {
	  $newfeatures[$count]->end($features[$i+1]->end);
	  $newfeatures[$count]->hend($features[$i+1]->hend);
	} else {
	  $newfeatures[$count]->end($features[$i+1]->end);
	  $newfeatures[$count]->hstart($features[$i+1]->hstart);
	}
	
	$newfeatures[$count]->score($newfeatures[$count]->score + $features[$i+1]->score);
	
	if ($features[$i+1]->hstart == $features[$i+1]->hend) {
	  $features[$i+1]->strand($features[$i]->strand);
	}
	
      } else {
	# we can't extend the merged homologies so start a
	# new homology feature
	
	# first do the coords on the old feature
	if ($newfeatures[$count]->hstart > $newfeatures[$count]->hend) {
	  my $tmp = $newfeatures[$count]->hstart;
	  $newfeatures[$count]->hstart($newfeatures[$count]->hend);
	  $newfeatures[$count]->hend($tmp);
	}
	
	$count++;
	$i++;
	
	push(@newfeatures,$features[$i]);
	$i--;
      }
    }
    
    # Adjust the last new feature coords
    if ($newfeatures[$#newfeatures]->hstart > $newfeatures[$#newfeatures]->hend) {
      my $tmp = $newfeatures[$#newfeatures]->hstart;
      $newfeatures[$#newfeatures]->hstart($newfeatures[$#newfeatures]->hend);
      $newfeatures[$#newfeatures]->hend($tmp);
    }
    
    # Set the gff features to the new merged set of features

    $self->{_homols}{$id} = [];

    foreach my $f (@newfeatures) {
      $self->add_homol($f);
    }

  }
}

=head2 filter

  Title   : filter
  Usage   : $homolset->filter($method,$score)
  Function: Filters the homols by method and score
  Returns : nothing
  Args    : none

=cut

sub filter {
  my ($self,$method,$score) = @_;

  $score = 200 unless $score;

  foreach my $id ($self->each_ID) {

    my $h = $self->{_homols}{$id};
    if ($h->[0]) {
      if ($h->[0]->source_tag eq $method) {
        my $found = 0;
        foreach my $h2 (@$h) {
	  if ($h2->score > $score) {
	    $found = 1;
	    last;
	  }
        }
        if (!$found) {
	  print("Ignoring $id\n");
	  $self->{_ignore}{$id} = 1;
        }
      }
    }
  }
}

=pod

head2 cluster

  Title   : cluster
  Usage   : $homolset->cluster
  Function: Makes an array of features each of which contains overlapping homologies
  Returns : nothing
  Args    : none

=cut

sub cluster {
  my ($self,$gap) = @_;
  
  $gap = 5 unless $gap;

  # Features hashed by database id
  my $hash = $self->{_homols};

  return unless 
  # The main cluster feature
  $self->{_cluster} = new Bio::SeqFeature::Generic;

  my $top = $self->{_cluster};

  # Sort homolset by start coord.
  my @homols = $self->sort;

  return unless $#homols >= 0;

  
 # Create the first cluster object
  my $subseq = new Bio::SeqFeature::Generic;

  $subseq->strand(1);

  # Start off the cluster with the first feature
  $subseq->add_sub_SeqFeature($homols[0],'EXPAND');

  if ($homols[0]->source_tag eq "BLASTX") {
    $subseq->strand($homols[0]->strand);
  }

  $top   ->add_sub_SeqFeature($subseq,'EXPAND');

  # Loop over the rest of the features
  my $count = 0;

  HOMOL: foreach my $h (@homols) {
      print("Clustering " . $h->id . "\n");
      # Dont' cluster homols with the ignore tag
      next HOMOL if $self->{_ignore}{$h->id};

      # Only cluster homols with >1 hit
      if ($#{$hash->{$h->id}} == 0) {
	next HOMOL;
      }
      
      if ($count > 0) {
	my @overlap = $h->match($subseq);    print("Overlap :@overlap:$#overlap\n");
	
	# Add to cluster if overlap
	if ($overlap[0]) {
	  $subseq->add_sub_SeqFeature($h,'EXPAND');
	  
	  if ($h->source_tag eq "BLASTX") {
	    $subseq->strand($h->strand);
	  } 
	  
	} else {
	  
	  # Start a new cluster
	  $subseq = new Bio::SeqFeature::Generic;
	  $subseq->add_sub_SeqFeature($h,'EXPAND');
	  
	  if ($h->source_tag eq "BLASTX") {
	    $subseq->strand($h->strand);
	  } else {
	    $subseq->strand(1);
	  }
	  
	  # And add to the top cluster feature
	  $top->add_sub_SeqFeature($subseq,'EXPAND');
	  
	}
      }
      $count++;
    }
}

=pod

=head2 is_in_cluster

  Title   : is_in_cluster
  Usage   : my $in = $homolset->is_in_cluster($id,$cluster)
  Function: Checks whether a feature with id $id is in a cluster
  Returns : 0,1
  Args    : String,Bio::SeqFeature::Generic

=cut

sub is_in_cluster {
   my($self,$id,$cluster) = @_;

#   print("Looking in cluster " . $cluster->start . "\t" . $cluster->end . "\n");
   foreach my $h ($cluster->sub_SeqFeature) {
      return(1) if $h->id eq $id;
   }
   return 0;
}


=pod

=head2 find_common_ends

  Title   : find_common_ends
  Usage   : $homolset->find_common_ends
  Function: Resets the end of the cluster to the most frequent coordinate
  Returns : nothing
  Args    : none

=cut


sub find_common_ends {
  my ($self,$clone) = @_;

  my $top = $self->{_cluster};
  my @patch = $top->sub_SeqFeature;

  @patch = sort { $a->start <=> $b->start } @patch;
  my $count = 0;

  foreach my $p (@patch) {
#     print("PATCHY " . $p->start . "\t" . $p->end . "\n");

      my %start;
      my %end;

      my %allstart;
      my %allend;
      
      my $maxstart = -1;
      my $maxend   = -1;

      my $newstart = $p->start;
      my $newend   = $p->end;



HOMOL: foreach my $f ($p->sub_SeqFeature) {

       next HOMOL if $f->source_tag eq "BLASTX";
       next HOMOL if $f->source_tag =~ "blastn";

       if ($count < $#patch && $self->is_in_cluster($f->id,$patch[$count+1])) {
	 $end{$f->end}++;
       } elsif ($count < $#patch-1 && $self->is_in_cluster($f->id,$patch[$count+2])) {
	 $end{$f->end}++;
       } elsif ($count < $#patch-2 && $self->is_in_cluster($f->id,$patch[$count+3])) {
	 $end{$f->end}++;
       } elsif ($count < $#patch-3 && $self->is_in_cluster($f->id,$patch[$count+4])) {
	 $end{$f->end}++;
       } elsif ($count < $#patch-4 && $self->is_in_cluster($f->id,$patch[$count+5])) {
	 $end{$f->end}++;
       } elsif ($count < $#patch-5 && $self->is_in_cluster($f->id,$patch[$count+6])) {
	 $end{$f->end}++;
       } elsif ($count < $#patch-6 && $self->is_in_cluster($f->id,$patch[$count+7])) {
	 $end{$f->end}++;
       } elsif ($count < $#patch-7 && $self->is_in_cluster($f->id,$patch[$count+8])) {
	 $end{$f->end}++;
       }
       
       if ($count > 0 && $self->is_in_cluster($f->id,$patch[$count-1])) {
	 $start{$f->start}++;
       } elsif ($count > 1 && $self->is_in_cluster($f->id,$patch[$count-2])) {
	 $start{$f->start}++;
       } elsif ($count > 2 && $self->is_in_cluster($f->id,$patch[$count-3])) {
	 $start{$f->start}++;
       } elsif ($count > 3 && $self->is_in_cluster($f->id,$patch[$count-4])) {
	 $start{$f->start}++;
       } elsif ($count > 4 && $self->is_in_cluster($f->id,$patch[$count-5])) {
	 $start{$f->start}++;
       } elsif ($count > 5 && $self->is_in_cluster($f->id,$patch[$count-6])) {
	 $start{$f->start}++;
       } elsif ($count > 6 && $self->is_in_cluster($f->id,$patch[$count-7])) {
	 $start{$f->start}++;
       } elsif ($count > 7 && $self->is_in_cluster($f->id,$patch[$count-8])) {
	 $start{$f->start}++;
       }
       
       $allstart{$f->start}++;
       $allend  {$f->end}++;
       
       
     }
     
     $count++;
     
     my $startflag = 0;
     my $endflag   = 0;
     
     
     # These are loops over linked hits
     foreach my $k (keys %start) {
       if ($start{$k} > $maxstart) { 
	 $startflag = 1;
	 $maxstart = $start{$k}; 
	 $newstart = $k;
       }
     }
     
     
      foreach my $k (keys %end) {
	if ($end{$k} > $maxend) { 
	  $endflag = 1;
	  $maxend = $end{$k};
	  $newend = $k;
	}
      }
      
      # These are loops over all the hits - 
      # this is only called if we have no linked hits
      if ($startflag == 0) {
	$maxstart = -1;
	foreach my $k (keys %allstart) {
	  if ($allstart{$k} > $maxstart) { 
	    $startflag = 1;
	    $maxstart = $allstart{$k}; 
	    $newstart = $k;
	  }
	}
      }
      
      if ($endflag == 0) {
	$maxend = -1;
	foreach my $k (keys %allend) {
	  if ($allend{$k} > $maxend) { 
	    $endflag = 1;
	    $maxend = $allend{$k};
	    $newend = $k;
	  }
	}
      }
      
      # Now we need to adjust these end points to coincide with any 
      # GENSCAN, fgenesh or fgenes sites that occur within
      # umm 15 bases either side of this end points
      
      my @types = ("GENSCAN","fgenesh","FGENES");

      my %lpredhash;
      my %rpredhash;
      my $predstrand;
      foreach my $g ($clone->eachPredictedGene) {
	foreach my $ty (@types) {
	  if ($g->source_tag =~ $ty) {

	    foreach my $ex ($g->each_Exon) {
	      if (abs($ex->start - $newstart) < 15) {
		$lpredhash{$ex->start}++;
                $predstrand = $ex->strand;
	      }
	      if (abs($ex->end - $newend) < 15) {
		$rpredhash{$ex->end}++;
                $predstrand = $ex->strand;
	      }
	    }
	  }
	}
      } # end foreach $g
      
      # Now - do we find the nearest splice? or the most common splice?
      # or the nearest most common splice?

      # Find the max splice
      my $max = -1;
      my $maxkey;

      foreach my $key (keys %lpredhash) {
	print("Looping over $key " . $lpredhash{$key} . "\n");
	if ($lpredhash{$key} > $max) {
	  $max    = $lpredhash{$key};
	  $maxkey = $key;
	  print("Resetting max = $max $maxkey\n");
	} elsif ($lpredhash{$key} == $max) {
	  if (abs($key - $newstart) < abs($maxkey - $newstart)) {
	    $max = $lpredhash{$key};
	    $maxkey = $key;
	  print("Resetting max = $max $maxkey\n");
	  }
	}
      }
      if ($max != -1) {
	$newstart = $maxkey;
           
      }

      $max = -1;

      foreach my $key (keys %rpredhash) {
	print("Looping over $key " . $rpredhash{$key} . "\n");
	if ($rpredhash{$key} > $max) {
	  $max    = $rpredhash{$key};
	  $maxkey = $key;
	  print("Resetting max = $max $maxkey\n");
	} elsif ($rpredhash{$key} == $max) {
	  if (abs($key - $newend) < abs($maxkey - $newend)) {
	    $max = $rpredhash{$key};
	    $maxkey = $key;
	  print("Resetting max = $max $maxkey\n");
	  }
	}
      }
      if ($max != -1) {
	$newend = $maxkey;
      }

     $p->start($newstart);
     $p->end  ($newend);
     
     
     # Default strand is forward
     if ($predstrand ne "") {
        $p->strand($predstrand);
     }
     if ($p->strand != -1 && $p->strand != 1) {
       $p->strand(1);
     }
   }
}


sub cluster2pattern {
  my ($cluster,$window,$clone,$type) = @_;

  my $counts;
  my @pred = $clone->eachProposedGene;
  my $pred = \@pred;

  my @subseq = $cluster->sub_SeqFeature;

  my $gene;
  my $exon;

  foreach my $g (@$pred) {
    foreach my $ex ($g->each_Exon) {
      my @overlap = Spangle::Homol::match($cluster,$ex);
      if ($overlap[0] == 1 && abs($overlap[1]) < $window) {
	$gene = $g;
	$exon = $ex;

#	print("Cluster start/end " . $cluster->start . "/" . $cluster->end . " ". $exon->start . "/" . $exon->end . "\n");
      }
    }
  }

  return unless defined($exon);

  my @types = ("BLASTX",
	       "vertebrate_mRNA_eg",
	       "EST_eg");
  
  
  my @preds = ("GENSCAN",
	       "fgenesh",
	       "FGENES");
  
  my $outstr = "";

  # Loop over the predicted positions
  if ($type eq "left") {
  for (my $j = $cluster->start - $window ; $j <= $cluster->start + $window ; $j++) {

    # Loop over the homols
    foreach my $f (@subseq) {
      
      # Do we have an overlap for this homol
      if ($f->start <= $j && $f->end >= $j) {
	my $score = $f->score_per_base/100.0;
	if ($f->source_tag eq "BLASTX") {
	  $score /= 5;
	}
	
	$counts->{$j}{$f->source_tag} = $counts->{$j}{$f->source_tag} + $score;
	
      }
      
    }


    foreach my $f (@preds) {
      my $found = 0;
      my $fg;


      foreach my $g ($clone->eachPredictedGene) {
	if ($g->source_tag =~ /$f/) {
	  $fg = $g;
	  $found = 1;
	}
	if($found == 1) {
	  foreach my $ex ($fg->each_Exon) {

	    if ($ex->start <= $j && $ex->end >= $j) {
	      $counts->{$j}{$f} = 1;
	    }
	  }
	}
      }
    }

    my $tot = 0;
      
    foreach my $type (@types) {
      $tot += $counts->{$j}{$type};
    }
    
    foreach my $type (@types) {
      if ($counts->{$j}{$type} ne "") {
	$outstr .= (1.0 *$counts->{$j}{$type}) . " ";
      } else {
	$outstr .= "0 ";
      }
    }
    foreach my $type (@preds) {
      if ($counts->{$j}{$type} ne "") {
	$outstr .= ($counts->{$j}{$type}) . " ";
      } else {
	$outstr .= "0 ";
      }
    }
    $outstr .= "\n";
  }
  for (my $j = $cluster->start - $window ; $j <= $cluster->start + $window ; $j++) {
    # Now we need the prediction data
    my $found = 0;
      
    if ($j == $exon->start) {
      $outstr .= "1 ";
    } else {
      $outstr .= "0 ";
    }
    
  }



  print("#\n$outstr\n");
}
  my $exon2;
  foreach my $g (@$pred) {
    foreach my $ex ($g->each_Exon) {
      my @overlap = Spangle::Homol::match($cluster,$ex);
      if ($overlap[0] == 1 && abs($overlap[2]) < $window) {
	$gene = $g;
	$exon2 = $ex;
      }
    }
  }

  return unless defined($exon2);

  print STDERR "Found exon " . $exon2->primary_tag . "\n";
  
  # Already declared above in same context
  #my @types = ("BLASTX",
  #         "vertebrate_mRNA_eg",
  #         "EST_eg");
  #
  #
  #my @preds = ("GENSCAN",
  #         "fgenesh",
  #         "FGENES");
  
  $outstr = "";
  # Loop over the predicted positions
  if ($type eq "right") {
  for (my $j = $cluster->end - $window ; $j <= $cluster->end + $window ; $j++) {

    # Loop over the homols
    foreach my $f (@subseq) {
      
      # Do we have an overlap for this homol
      if ($f->start <= $j && $f->end >= $j) {
	my $score = $f->score_per_base/100.0;
	if ($f->source_tag eq "BLASTX") {
	  $score /= 5;
	}
	
	$counts->{$j}{$f->source_tag} = $counts->{$j}{$f->source_tag} + $score;
	
      }
    }

    foreach my $f (@preds) {
      my $found = 0;
      my $fg;

      foreach my $g ($clone->eachPredictedGene) {
	if ($g->source_tag =~ /$f/) {
	  $fg = $g;
	  $found = 1;
	}

	if($found == 1) {
	  foreach my $ex ($fg->each_Exon) {
	    if ($ex->start <= $j && $ex->end >= $j) {
	      $counts->{$j}{$f} = 1;
	    }
	  }
	}
      }
    }


    my $tot = 0;
      
    foreach my $type (@types) {
      $tot += $counts->{$j}{$type};
    }
    
    foreach my $type (@types) {
      if ($counts->{$j}{$type} ne "") {
	$outstr .= (1.0 *$counts->{$j}{$type}) . " ";
      } else {
	$outstr .= "0 ";
      }
    }


    foreach my $type (@preds) {
      if ($counts->{$j}{$type} ne "") {
	$outstr .= ($counts->{$j}{$type}) . " ";
      } else {
	$outstr .= "0 ";
      }
    }
    $outstr .= "\n";
  }
  for (my $j = $cluster->end - $window ; $j <= $cluster->end + $window ; $j++) {
    # Now we need the prediction data
    my $found = 0;
      
    if ($j == $exon2->end) {
      $outstr .= "1 ";
    } else {
      $outstr .= "0 ";
    }
    
  }
  print("#\n$outstr\n");
}
}

  
  
  

=pod

=head2 cluster2patch

  Title   : cluster2patch
  Usage   : \$homolset->cluster2patch(\$cluster)
  Function: Makes patch features from clustered HomolSet
  Returns : nothing
  Args    : Bio::SeqFeature::Generic

=cut

sub cluster2patch {
    my ($self,$cluster,$tol) = @_;
    
    $tol = 20 unless $tol;
    
    my @patch;
    my @subseq;

    my $count = 0;
    my $lsplice = 0;
    my $rsplice = 0;

    my $splice = [];

    @subseq = $cluster->sub_SeqFeature;
    @subseq = sort { $a->start <=> $b->start} @subseq;

    # Make the first patch
    $patch[0] = new Bio::SeqFeature::Generic;
    my $fh    = $subseq[0];
    $patch[0]->add_sub_SeqFeature($fh,'EXPAND');

    # Do we have spliced left and right ends?
    my $seqlen   = $self->get_length($fh);
    push(@{$splice->[0]},$fh->splice);

    # Now deal with the rest
    for (my $i = 1; $i<=$#subseq; $i++) {  
	
	my $homol  = $subseq[$i];
	my $seqlen = $self->get_length($homol);
	
	# loop over all the patches made until we find one
	my $found = 0;
	my $cc = 0;
	
	while ($cc <= $#patch) {
	    my $p = $patch[$cc];
	    
	    my @overlap = $homol->match($p);

	    my $lo = 0;
	    my $ro = 0;
	    
	    ($lo,$ro) = $homol->test_overlap($p);
	    
	    if ($lo && $ro) {
		
		# Change the splicing if necessary
		my @tmpspl = $homol->splice;
		
		if ($tmpspl[0] == 1) {$splice->[$cc][0] = 1;}
		if ($tmpspl[1] == 1) {$splice->[$cc][1] = 1;}
		
		$patch[$cc]->add_sub_SeqFeature($homol,'EXPAND');
		
		$found = 1;
		
	    }
	    $cc++;
	}
	
	if ($found == 0) {
	    $count++;

	    $patch[$count] = new Bio::SeqFeature::Generic;
            $patch[$count]->add_sub_SeqFeature($homol,'EXPAND');

	    my @splice = $homol->splice;
            push(@{$splice->[$count]},@splice);

	    $found = 1;
	}
    }


    $self->patches(@patch);

    return @patch;
}

=head2 patches

  Title   : patches
  Usage   : @patches = $homolset->patches
  Function: Returns array of homol patcges
  Returns : @Spangle::Homol
  Args    : none

=cut

sub patches {
  my ($self,@arg) = @_;


  if (!defined($self->{_patches})) {
    $self->{_patches} = new Bio::SeqFeature::Generic;
  }

  if (@arg) {
    foreach my $f (@arg) {
      print("Adding patch " . $f->start . "\t" . $f->end . "\n");
      $self->{_patches}->add_sub_SeqFeature($f,'EXPAND');
    }
  }
  
  return $self->{_patches};

}


=head2 link_patches

  Title   : link_patches
  Usage   : $homolset->link_patches
  Function: Links the clustered and patched features together
  Returns : nothing
  Args    : none

=cut



sub link_patches {
  my ($self,$tol,$do_splice,$fd,$fa,$rd,$ra) = @_;
  
  $tol = 10 unless $tol;
  
  my $genes;
  my $genes2;
  
#  my @patches = $self->patches;
  my $top = $self->{_cluster};
  print("Top is $top\n");
  
  my @patches = $top->sub_SeqFeature;

  print("Number of patches is $#patches\n");

  # Sort the clusters by start position
  @patches = sort { $a->start <=> $b->start } @patches;
  
  # Loop over all clusters
 PATCH1:  for (my $c = 0; $c < $#patches; $c++) {
    my $p1    = $patches[$c];
    my @sub   = $p1->sub_SeqFeature;
    
    print("Finding links in cluster number $c \n");
    
    # Sort the supporting features by score
    @sub = sort  { $b->score <=> $a->score } @sub;
    
    my $found = 0;
    my $c2    = $c+1;
    
    # Now look in all the other clusters to find a link
  PATCH2:    while ($c2 <= $#patches) {
      
      my $p2   = $patches[$c2];
      my @sub2 = $p2->sub_SeqFeature;
      
      # Only look at a cluster if it starts after the end of
      # the current cluster

      if ($p2->start > $p1->end) {
	# Sort by score
	@sub2 = sort { $a->score <=> $b->score } @sub2;
	
	my $foundlink = 0;

	# Now loop over the supporting features in the first linking cluster
	foreach my $s1 (@sub) {
#	  print($s1->toString . "\n");
	  # Does the second cluster contain this database id?
	  my $s2 = $self->get_id($p2,$s1->id);

	  # If we have the same id in two clusters then...
	  if (ref($s2) ne "" ) {
#	    print("Found ids in two clusters\n");
#	    print($s2->toString . "\n");
	    # Check the strand is the same
	    if ($s1->strand eq $s2->strand) {
	      
	      # Only allow a $tol discontinuity in the sequence
	      if ($s1->strand == $s2->strand && (abs($s1->hend - $s2->hstart) < $tol || abs($s1->hstart - $s2->hend) < $tol)) {

		print($s1->toString . "\n");
		print($s2->toString . "\n");

		$found = 1;
		$foundlink = 1;
	      }
	    }

	    if ($foundlink == 1) {
	      # We have found a link between p1 and p2
	      print("Creating link from " . $p1->start . "\t" . $p1->end . "\n");
	      print("to                 " . $p2->start . "\t" . $p2->end . "\n");
	      push(@{$p1->{_forward}} ,$p2);
	      push(@{$p2->{_backward}},$p1);
	      next PATCH1;
	    }
	  }

	} # end of loop over p2 sub seqfeatures

      } # end of p2->start > p1->end
      $c2++;
    }  # END of loop over p2

  } # End of loop over p1

  # We now have all the links between patches
  # Build genes from the links.
  
  for (my $c = 0; $c < $#patches; $c++) {
    my $p = $patches[$c];
    
    if (!defined($p->{_backward}) && defined($p->{_forward})) {
      #     print("Creating new subseq\n");
      my $tr = Spangle::Subseq->new;
      
      $tr->source_tag("SPAN.$c");
      $tr->primary_tag("SPAN");
      
      my $ex = new Spangle::Exon;
      #      print("New exon " . $p->start . "\t" . $p->end . "\n");
      $ex->start ($p->start);
      $ex->end   ($p->end);
      $ex->strand($p->strand);

      $tr->add_Exon($ex);

      while (ref($p->{_forward}) ne "") {
	$p = $p->{_forward}[0];

	my $ex = new Spangle::Exon;
	
	$ex->start ($p->start);
	$ex->end   ($p->end);
	$ex->strand($p->strand);

	foreach my $sup ($p->sub_SeqFeature) {
	  $ex->add_supporting_feature($sup);
	}
	
#	print("New exon " . $p->start . "\t" . $p->end . "\n");
	$tr->add_Exon($ex);

      }

      my $strand = 1;
      my $min = 100000000;
      my $max = 0;

      foreach my $ex ($tr->each_Exon) {
	foreach my $sup ($ex->each_supporting_feature) {
	  if ($sup->source_tag eq "BLASTX") {
	    $strand = $sup->strand;
	  }
	}
	$ex->{_supporting_feature} = [];

	
      }

      foreach my $ex ($tr->each_Exon) {
	$ex->strand($strand);

	if ($ex->start <= $min) {$min = $ex->start;}
	if ($ex->end   >= $max) {$max = $ex->end;}
	
      }

      $tr->start($min);
      $tr->end  ($max);

      # Add the new transcript to the hash
      $self->add_transcript($tr);

    }
  }
}

=head2 add_transcript

  Title   : add_transcript
  Usage   : $homolset->add_transcript($tr);
  Function: Adds a transcript to the homolset
  Returns : nothing
  Args    : Bio::EnsEMBL::Transcript

=cut

sub add_transcript {
  my ($self,$trans) = @_;

  if (!defined($self->{_transcript})) {
    $self->{_transcript} = [];
  }
  push(@{$self->{_transcript}},$trans);
}

=head2 each_transcript

  Title   : each_transcript
  Usage   : @tr = $homolset->each_transcript
  Function: Returns an array of transcripts
  Returns : @Bio::EnsEMBL::Transcript
  Args    : none

=cut

sub each_transcript {
  my $self = shift;

  if (!defined($self->{_transcript})) {
    $self->{_transcript} = [];
  }
  return @{$self->{_transcript}};
}
   

=head2 get_length

  Title   : get_length
  Usage   : $homolset->get_length($h)
  Function: Gets the complete length of the database hit
  Returns : int
  Args    : Homol

=cut

sub get_length {
  my ($self,$homol) = @_;

  $homol || $self->throw("No. homol input. Can't get length");
  
  # Create the hash if necessary
  if (!defined($self->{_length_hash})) {
    $self->{_length_hash} = new Spangle::SeqLength_DB;
  }
  
  # Set the length in the homol feature
  my $len =  $self->{_length_hash}->get_length($homol->id);
  $homol->length($len);
  
  return $homol->length;
}

=head2 get_id 

  Title   : get_id
  Usage   : $feat = $homolset->get_id($feat,$id);
  Function: Looks among subfeatures and returns feature with id id
  Returns : Spangle::Homol
  Args    : string

=cut

sub get_id {
  my ($self,$feat,$id) = @_;

  foreach my $sub ($feat->sub_SeqFeature) {
    if ($sub->id eq $id) {
      return $sub;
    }
  }
}

=head2 reduce

  Title   : reduce
  Usage   : $homolset->reduce
  Function: Reduces redundancy in the transcripts generated by link_cluster
            If a shorter transcript overlaps with a longer one
            it is deleted from the transcript array.
  Returns : nothing
  Args    : none

=cut


sub reduce {
  my ($self) = @_;

  my @tran = $self->each_transcript;

  print("Numver of transcripts = " . ($#tran+1) . "\n");
  return unless ($#tran >= 0);

  # Sort the transcripts by number of exons
  @tran = sort { $b->each_Exon <=> $a->each_Exon } @tran;

  my @nonred ;
  push(@nonred,$tran[0]);

  my $count2 = 0;

  my @tmp = $tran[0]->each_Exon;

  for (my $i = 1; $i <= $#tran; $i++) {

    my $tr   = $tran[$i];
    my @tmp  = $tr->each_Exon;

    my $done = 0;
    
    # Loop over non-red gffs
    for (my $j = 0; $j <= $#nonred; $j++) {
      
      # This is count of exons that overlap
      my $count = 0;
      
      if ($done == 0) {
	my $nr   = $nonred[$j];
	my @nrex = $nr->each_Exon;

	# Loop over non-red exons
	foreach my $nrex (@nrex) {

	  # Loop over our test span
	  foreach my $ex ($tr->each_Exon) {      

	    my @overlap = Spangle::Homol::match($ex,$nrex);

	    if ($overlap[0] == 1) {
	      $count++;
	      # Test for complete coverage of the span
	    }

	  }
	  if ($count >= $#tmp) {
	    $done = 1;
	  }
	}
      }
    }
    
    # If not overlap make a new cluster
    if ($done == 0) {
      push(@nonred,$tr);
      $count2++;
    }
  }
  
    
  return @nonred;
}

  
sub delete_feature {
  my ($self,$feat) = @_;

  foreach my $id (keys %{$self->{_homols}}) {
    my $count = 0;
    my @arr   = @{$self->{_homols}{$id}};

    for ($count = 0; $count <= $#arr; $count++) {
      if ($feat == $arr[$count]) {
	splice(@{$self->{_homols}{$id}},$count,1);
	return 1;
      }
    }
  }
  return 0;
}
1;
