package Spangle::Link;

=pod

  =head1 NAME
  
  Spangle::Link - Link object

  =head1 SYNOPSIS
  
  $cl = new Spangle::Link
  $cl = new Spangle::Clone($id,$db);  # Make a new link with id and acedb handle
  
=head1 DESCRIPTION

A module to store and manipulate linked clone objects

=head1 CONTACT

Describe contact details here

=head1 APPENDIX

The rest of the documentation details each of the object methods. Internal methods are usually preceded with a _

=cut

use strict;  
use vars   qw(@ISA);

use Spangle::Clone;
use Spangle::Ace;

use Bio::Seq;

# Object preamble - inherits from Bio::SeqFeature::Generic

use Bio::SeqFeature::Generic;

@ISA = qw(Bio::SeqFeature::Generic);

sub new {
    my($pkg, @args) = @_;

    my $self = $pkg->SUPER::new(@args);

    # Initialize the variables
    $self->{_clones}    = [];
    $self->{_subseqs}   = [];
    
    $self->primary_tag('Link');
    $self->source_tag ('Spangle');

    $self->_make(@args);
    return $self;
}

sub _make {
    my ($self,$id,$db) = @_;
    
    $id || $self->throw("No link id entered\n");
    
    $self->id   ($id);
    $self->acedb($db);
    
}

sub id {
    my ($self,$arg) = @_;
    
    if (defined($arg)) {
	$self->{_id} = $arg;
    } else {
	return $self->{_id};
    }
}

sub acedb {
    my ($self,$arg) = @_;
    if (defined($arg)) {
	$self->{_acedb} = $arg;
    } else {
	return $self->{_acedb};
    }
}


sub eachClone {
    my $self = shift;
    
    if (defined($self->{_clones})) {
	return sort {$a->start <=> $b->start } @{$self->{_clones}};
    }
    
}

sub eachSubsequence {
    my $self = shift;
    
    if (defined($self->{_subseqs})) {
	return sort {$a->start <=> $b->start } @{$self->{_subseqs}};
    }
    
}


=head2 add_clone

  Title   : add_clone
  Usage   : $clone->add_clone($clone)
  Function: Adds a clone into the link
  Returns : nothing
  Args    : none

=cut

sub add_clone {
    my ($self,$clone) = @_;

    if ($clone->end < $clone->start) {
	my $tmp = $clone->end;
	$clone->end   ($clone->start);
	$clone->start ($tmp);
	$clone->strand(-1);
    }

    if (defined($clone)) {
	push(@{$self->{_clones}},$clone);
    } 

}


=head2 add_subsequence

  Title   : add_subsequence
  Usage   : $clone->add_subsequence($clone)
  Function: Adds a subsequence into the link
  Returns : nothing
  Args    : none

=cut

sub add_subsequence {
    my ($self,$clone) = @_;

    if ($clone->end < $clone->start) {
	my $tmp = $clone->end;
	$clone->end   ($clone->start);
	$clone->start ($tmp);
	$clone->strand(-1);
    }

    if (defined($clone)) {
	push(@{$self->{_subseqs}},$clone);
    } 
}

=head2 get_clone

  Title   : get_clone
  Usage   : $clone->get_clone($cloneid)
  Function: Retrieves a clone object by id
  Returns : Spangle::Clone
  Args    : none

=cut

sub get_clone {
  my ($self,$id) = @_;

  foreach my $cl ($self->eachClone) {
     if ($cl->id eq $id) {
       return $cl;
     }
  }
}


sub split_subseqs {
    my ($self,$clone) = @_;

    my @sub1;
    my @sub2;
    my @sub3;

    foreach my $sub ($self->eachSubsequence) {
	if ($sub->end < $clone->start) {
	    push(@sub1,$sub);
	} elsif ($sub->start > $clone->end) {
	    push(@sub2,$sub);
	} else {
	    push(@sub3,$sub);
	}
    }
    foreach my $sub (@sub3) {
	$self->warn("Subsequence " . $sub->id . "will have the wrong coordinates as it overlaps the changed clone");
    }
    return (\@sub1,\@sub2,\@sub3);
}

=head2 remove_clone

  Title   : remove_clone
  Usage   : my (@newlinks) = $link->remove_clone($clone)
  Function: Removes a clone from the link object
  Returns : Array of links (1 or 2 depending on where the deleted 
			    clone is)
  Args    : Spangle::Clone

=cut

sub remove_clone {
    my ($self,$clone) = @_;
    
    my @new1;
    my @new2;

    my $link1 = new Spangle::Link;
    my $link2 = new Spangle::Link;

    print("Removing clone " . $clone->id . "\n");

    my ($sub1,$sub2,$sub3) = $self->split_subseqs($clone);

    if (defined($clone)) {
	my @clones = $self->eachClone;
	my $found = 0;
	
	while ((my $cl = shift(@clones)) && $found == 0) {

	    if ($cl->id eq $clone->id) {
		$found = 1;
		last;
	    } else {
		push(@new1,$cl);
	    }
	}

	push(@new2,@clones);

	# Should check that new1 actually contains clones here

	$self->{_clones} = [];

	foreach my $cl1 (@new1,@new2) {
	    $self->add_clone($cl1);
	}
	
	print("Link1  - adding " . $#new1 . " clones\n");	
	print("Link2  - adding " . $#new2 . " clones\n");	
	
	if ($#new2 >= 0) {
	    $link2->id   ($new2[0]->id);
	    $link2->acedb($self->acedb);
	    
	    foreach my $cl (@new2) {
		$link2->add_clone($cl);
	    }
	    foreach my $cl (@$sub2) {
		$link2->add_subsequence($cl);
	    }
	    
	    $link2->shift_coords(-1 * $new2[0]->start+1,0);
	    
	}

	if ($#new1 >= 0) {
	    $link1->id   ($new1[0]->id);
	    $link1->acedb($self->acedb);
	    
	    foreach my $cl (@new1) {
		$link1->add_clone($cl);
	    }

	    foreach my $cl (@$sub1,@$sub3) {
		$link1->add_subsequence($cl);
	    }
	    
	    $link1->shift_coords(-1 * $new1[0]->start+1,0);
	    
 	}

    } 
    
    return ($clone,$link1,$link2);
}

=head2 delete_clone

  Title   : remove_clone
  Usage   : $link->delete_clone($clone)
  Function: Removes a clone from the link object
  Returns : Nothing
  Args    : Spangle::Clone

=cut

sub delete_clone {
    my ($self,$clone) = @_;
    
    print("Deleting clone " . $clone->id . "\n");

    if (defined($clone)) {
	my @clones = $self->eachClone;
	my @new;

	while (my $cl = shift(@clones)) {

	    if ($cl->id ne $clone->id) {
		push(@new,$cl);
	    }
	}

	# Should check that new1 actually contains clones here

	$self->{_clones} = [];

	foreach my $cl1 (@new) {
	    $self->add_clone($cl1);
	}
    }
}

=head2 remove_clone_by_ID

  Title   : remove_clone_by_ID
  Usage   : $link->remove_clone($cloneid)
  Function: Removes a clone from the link object
  Returns : (Spangle::Clone,Spangle::Link,Spangle::Link)
  Args    : none

=cut

sub remove_clone_by_ID {
    my ($self,$cloneid) = @_;

    my $clone;

    my $link1;
    my $link2;

    if (defined($cloneid)) {
	foreach my $cl ($self->eachClone) {
	    if ($cl->id eq $cloneid) {
		$clone = $cl;
	    }
	}

	my @tmp = $self->eachClone;

	($clone,$link1,$link2) = $self->remove_clone($clone);
    } 

    return ($clone,$link1,$link2);
}

=head2 length

  Title   : length
  Usage   : $len = $link->length
  Function: Returns the length of the link
  Returns : int
  Args    : none

=cut

sub length {
  my ($self) = @_;
  my $len = 0;

  foreach my $cl ($self->eachClone) {
    if ($cl->start > $len) {$len = $cl->start;}
    if ($cl->end   > $len) {$len = $cl->end;}
  }

  return $len;
}

=head2 shift_coords

  Title   : shift_coords
  Usage   : $link->shift_coords($offset,$start)
  Function: Shifts a link by $offset bases starting from $start
  Returns : nothing
  Args    : none

=cut

sub shift_coords {
  my ($self,$offset,$start) = @_;

  $start = 0 unless $start;

  foreach my $cl ($self->eachClone,$self->eachSubsequence) {

      my $newstart = $cl->start;
      my $newend   = $cl->end;

      if ($cl->start > $start) { $newstart = $cl->start + $offset;}
      if ($cl->end   > $start) { $newend   = $cl->end   + $offset;}
     
      $cl->start($newstart);
      $cl->end  ($newend);

  }
}


=head2 reverse

  Title   : reverse
  Usage   : $link->reverse
  Function: Reverses a link
  Returns : nothing
  Args    : none

=cut

sub reverse {
  my ($self) = @_;

  my $len = $self->length;

  foreach my $cl ($self->eachClone) {
    my $newstart = $len - $cl->start + 1;
    my $newend   = $len - $cl->end   + 1;

    $cl->start($newstart);
    $cl->end  ($newend);
  }
}

    
=head2 merge

  Title   : merge
  Usage   : $link->merge($link2)
  Function: Merges link2 into link1 if they share 1 common clone
  Returns : nothing
  Args    : none

=cut

sub merge {
  my ($link1,$link2) = @_;


  my ($cloneid,$orient) = $link1->_find_common_clone($link2);
  print("\nCommon/orient $cloneid $orient\n");

  my @tmp1 = $link1->eachClone;
  my @tmp2 = $link2->eachClone;

  my $clone1 = $link1->get_clone($cloneid);
  my $clone2 = $link2->get_clone($cloneid);

  my $shift = $clone1->start;
     $shift = $clone1->end if ($clone1->end < $clone1->start);

  print("Shifting by $shift\n");

  $link2->shift_coords($shift-1,0);

  # There are 4 orientation cases

  if ($orient == 1) {
      # Case 1.
      # 1) ----------->
      # 2)          --------->

      $link2->delete_clone($clone2);
      
  } elsif ($orient == 2) {
      # Case 2.
      # 1) ----------->
      # 2)          <---------
      
      $link2->reverse;
      $link2->delete_clone($clone2);

  } elsif ($orient == 3) {
      # Case 3.
      # 1) <-----------
      # 2)          ---------->
      
      $link1->reverse;
      $link2->delete_clone($clone2);
      
  } elsif ($orient == 4) {
      # Case 4.
      # 1) <-----------
      # 2)          <---------
      
      $link2->delete_clone($clone2);
  } 

  foreach my $cl ($link2->eachClone) {
    $link1->add_clone($cl);
  }
  foreach my $cl ($link2->eachSubsequence) {
    $link1->add_subsequence($cl);
  }

  print("done merging\n");
}

sub _find_common_clone {
  my ($link1,$link2) = @_;

  my @ids;
  
  foreach my $cl ($link1->eachClone) {
      if ($link2->get_clone($cl->id)) {
      push(@ids,$cl->id);
    }
  }

  if ($#ids > 0) {
    $link1->throw("ERROR: more than 1 common clone to the links (@ids)");
  }
  
  my $common = $ids[0];
  print("Common clone is $common\n");
  my $clone1 = $link1->get_clone($common);
  my $clone2 = $link2->get_clone($common);

  my $orient = 0;

  if  ($clone1->strand ==  1 && $clone2->strand ==  1) {
     $orient = 1;
  } elsif ($clone1->strand ==  1 && $clone2->strand == -1)   {
     $orient = 2;
  } elsif ($clone1->strand == -1 && $clone2->strand ==  1)   {
     $orient = 3;
  } elsif ($clone1->strand == -1 && $clone2->strand == -1)   {
     $orient = 4;
  }

  return ($clone1->id,$orient);
}

sub print {
    my ($self) = @_;

    my @clones = $self->eachClone;
    my $num = $#clones+1;

    print("\n===== Link id : " . $self->id . "\n");
    print("Number of clones is $num\n");
    printf("\n%15s %7s %7s %7s\n","Id","Start","End","Strand");
    
    foreach my $cl (@clones) {
	printf("%15s %7d %7d %7d\n",$cl->id,$cl->start,$cl->end,$cl->strand);
    }
    print("\n");
}
    
=head2 print_ace

  Title   : print_ace
  Usage   : $link->print_ace
  Function: Prints ace string for link
  Returns : string
  Args    : none

=cut

sub print_ace {
  my ($self) = @_;

  my $id;
  my $ace;

  my @clones = $self->eachClone;
  my @subs   = $self->eachSubsequence;

  if ($self->id eq "") {
    $id = "LINK_" . $clones[0]->id;
  } else {
    $id = $self->id;
  }

  $ace .= "-D Sequence $id\n";
  $ace .= "\nSequence $id\nLink\n";

  foreach my $cl (@clones,@subs) {
      if ($cl->strand == -1) {
	  $ace .= "Subsequence " . $cl->id . " " . $cl->end . " " . $cl->start . "\n";
      } else {
	  $ace .= "Subsequence " . $cl->id . " " . $cl->start . " " . $cl->end . "\n";
      }
  }

  foreach my $cl (@clones) {
    $ace .= "\nSequence " . $cl->id . "\nSource $id\n";
  }  

  return $ace;
}

sub _contains_one_of {
    my ($self,$clone,@clones) = @_;
    
    return 1 if $#clones < 0;

    foreach my $cl (@clones) {
	return 1 if ($cl eq $clone->id);
    }
    return 0;
}

=head2 trim_dna

  Title   : trim_dna
  Usage   : $link->trim_dna
  Function: Trims dna, sets trim_left/right tags  and returns ace string
  Returns : string
  Args    : none

=cut

sub trim_dna {
    my ($self,@trimclones) = @_;
    my $acestr;
    my @clones = $self->eachClone;
    
  CLONE:   foreach my $clone ($self->eachClone) {
      $self->acedb(Spangle::Ace::connect('humsrv1',310000)) unless $self->acedb->ping;
      
      my $trim    = 1;                

      next CLONE unless ($clone->id !~ /\.SPAN/ && $clone->id !~ /\.TRIMMED/ && $clone->id !~ /\.C22/);
      next CLONE unless $self->_contains_one_of($clone,@trimclones);

      eval {

	  $self->trim_clone($clone);

	  my $lab = $clone->lab;

	  if ($lab eq "SC" || $lab eq "Sanger Centre") {
	      print("Sanger clone - not trimming DNA\n");
	  }

	  $acestr .= $clone->trimace;

	  undef($clone);
	  
      };
      
      if ($@) {
	  warn($@);
	  print("Error processing clone. Skipping\n");
      }
  }
    
  return ($acestr);
}

sub trim_clone {
    my ($self,$clone) = @_;
    my $trim = 1;

    print(STDERR "\nProcessing clone " . $clone->id . "\n"); 
    print(STDERR "Strand is " . $clone->strand . "\t" . $clone->start . "\t" . $clone->end ."\n");

    $clone->acedb($self->acedb);

    my ($new_trim_left,$new_trim_right) = $self->_find_overlaps($clone);   
    
    $clone->trim_left ($new_trim_left);
    $clone->trim_right($new_trim_right);

    $clone->trim;

    my $diff    = $self->_compare_dna($clone);

}

=head2 dump_dna

  Title   : dump_dna
  Usage   : $link->dump_dna
  Function: Dumps fasta string (masked or raw sequence)
  Returns : string
  Args    : none

=cut

sub dump_dna {
  my ($self,$mask) = @_;
  
  my @clones  = $self->eachClone;
  my $nclones = $#clones+1;
  my $count   = 1;
  my $alldna  = "";
  my $linklen = 0;

  my @clones  = $self->eachClone;
     @clones  = sort {$a->start <=> $b->start} @clones;

 CLONE: foreach my $clone (@clones) {
     next CLONE unless ($clone->id !~ /\./);
    
     eval {
	 print(STDERR "Processing clone " . $clone->id . " $count/$nclones\n");
	 $count++;
      
	 $self->trim_clone($clone);
      
	 if ($clone->start > $linklen) {$linklen = $clone->start;}
	 if ($clone->end   > $linklen) {$linklen = $clone->end;}
      
	 next CLONE if ($clone->{_trim_left} > $clone->{_trim_right});
	
	 my $trimdna = $clone->trimdna;
	 
	 print(STDERR "Length of dna is " . (length($trimdna)). "\n");

	 $alldna .= $trimdna;

	 undef($clone);

     };
    
     if ($@) {
	 warn($@);
	 print(STDERR "Error processing clone. Skipping\n");
     }
 }
  
  print(STDERR "Link length is $linklen\n");
  print(STDERR "Dna  length is " . (length($alldna)) . "\n");
  
  return $alldna;

}

sub _write_trimmed_fasta_sequence {
    my ($self,$clone,$dna) = @_;
    my $acestr;
    my $strand = $clone->strand;

    print(STDERR "Strand = $strand\n");
    my ($substart,$subend ) = $self->find_subseq($clone);

    print(STDERR "DNA start - end : $substart $subend\n");
    my $newdna = substr($dna,$substart-1,($subend - $substart + 1));

    if ($strand == -1) {
	$newdna =~ tr/atcgATCG/tagcTAGC/;
	$newdna = reverse($newdna);
    }

    return $newdna;
    
}



sub _compare_dna {
  my ($self,$clone) = @_;

  my $dna    = $clone->fetch_DNA;
  my $length = abs($clone->start - $clone->end) + 1;
  my $dnalen = length($dna);
  my $diff   = ($length - $dnalen);

  print(STDERR "DNA length is " . $dnalen . "\n");
  print(STDERR "Link clone length is $length\n");
  print(STDERR "Difference in DNA lengths for " . $clone->id . " is " . $diff . "\n");

  $self->throw("ERROR: Length of clone in link is different from clone dna length\n") if $diff;

  return ($length = $dnalen);
}

sub _find_overlaps {
  my ($self,$clone) = @_;

  my $new_trim_left  = 0;
  my $new_trim_right = 0;


  foreach my $cl ($self->eachClone) {

    if ($cl != $clone && $cl->id !~ /\./) {

      my @overlap = Spangle::Homol::match($clone,$cl);

      if ($clone->id eq "cN5H6") {
	print($cl->id . " @overlap\n");
      }
      if ($overlap[0] == 1) {
	if ($overlap[1] <= 0 && $overlap[2] > 0) {
	  if ($overlap[2] > $new_trim_left) {
	    $new_trim_left = $overlap[2];
	  }
	} elsif ($overlap[1] > 0 && $overlap[2] <= 0) {
	  if ($overlap[1] > $new_trim_right) {
	    $new_trim_right = $overlap[1];
	  }
	} elsif ($overlap[1] > 0 && $overlap[2] > 0) {
	  if ($overlap[1] > $new_trim_left) {
	    $new_trim_left = $overlap[1];
	  }
	  if ($overlap[2] > $new_trim_right) {
	    $new_trim_right = $overlap[2];
	  }
	}
      }
    }
  }
  
  # Now convert the link coordinates to clone coordinates
  my $strand = 1;
  if ($clone->start > $clone->end) {$strand = -1;}

  if ($strand == 1) {
    my $len = ($clone->end - $clone->start + 1);

    if ($new_trim_left != 0) {
      $new_trim_left  = $len - $new_trim_left + 1;
    } else {
      $new_trim_left = 1;
    }

    if ($new_trim_right != 0) {
      $new_trim_right = $new_trim_right;
    } else {
      $new_trim_right = $len;
    }
    
  } else {
    my $len = ($clone->start - $clone->end + 1);
    my $tmp = $new_trim_left;

    if ($new_trim_right != 0) {
      $new_trim_right =  $new_trim_right;
    } else {
      $new_trim_right = $len;
    }
    if ($new_trim_left != 0) {
      $new_trim_left = $len - $tmp + 1;
    } else {
      $new_trim_left = 1;
    }
  }

  print("New left/right = $new_trim_left $new_trim_right\n");

  return ($new_trim_left,$new_trim_right);
}

=head2 parse

  Title   : parse
  Usage   : $link->parse(@data)
  Function: Parses ace lines containing subsequences
  Returns : nothing
  Args    : none

=cut

sub parse {
  my ($self,@data) = @_;

  LINE : foreach my $line (@data) {
    if ($line =~ /^Sequence +: +(.*)/) {
      my $id = $1;
      $id =~ s/\"//g;
      
      $self->id($1);

    } elsif ($line =~ /^Subsequence[\t ]+(.*)[\t ]+(\d+) +(\d+)/) {
      
      my $id    = $1;
      my $start = $2;
      my $end   = $3;
      my $strand= 1;

      $id =~ s/\"//g;



      my $clone = new Spangle::Clone;

      $clone->id   ($id);
      $clone->start($start);
      $clone->end  ($end);
      
      if ($clone->start > $clone->end) { $strand = -1;}

      $clone->strand($strand);

      if ($id =~ /\./) {
	  $self->add_subsequence($clone);
      } else {
	  $self->add_clone($clone);
      }

    }
  }
}


=head2 fetch

  Title   : fetch
  Usage   : $link->fetch
  Function: Fetches a link from an ace database
  Returns : nothing
  Args    : none

=cut


sub fetch {
  my ($self,$id) = @_;


  $self->id($id) if $id;

  $self->throw("No id specified. Can't fetch link")              unless $self->id;
  $self->throw("No database handle specified. Can't fetch link") unless $self->acedb;

  my $data = $self->acedb->raw_query("find sequence " . $self->id);
     $data = $self->acedb->raw_query("show -a");

     $data =~ s/\0//g;

  my @data = split(/\n/,$data);

  $self->parse(@data);
}

sub dump_dna2 {
  my ($self) = @_;

  my $linklen = 0;
  my $linkdna;

  CLONE:   foreach my $clone ($self->eachClone) {
    next CLONE unless ($clone->id !~ /\./);
      
    eval {
      print("Processing clone " . $clone->id . "\n");

      $clone->acedb($self->acedb);

#      my ($trim_left,$trim_right) = $clone->fetch_trims;

#      my $dna                     = $clone->fetch_DNA_file; 
#      my $diff                    = $self->_compare_dna($dna,$clone);
#      my $strand                  = 1;
#      my $len                     = length($dna);

#      if ($clone->start > $clone->end) { $strand = -1; }

#      if ($clone->start > $linklen) {$linklen = $clone->start;}
#      if ($clone->end   > $linklen) {$linklen = $clone->end;}

#      my $substart;
#      my $subend;

#      if ($strand == 1) {
#	$substart = $trim_left;
#	$subend   = $len;
#      } else {
#	$substart = 1;
#	$subend   = $trim_right;
#      }
      
#      print("DNA start - end : $substart $subend\n");

#      my $newdna = substr($dna,$substart-1,($subend - $substart + 1));

#      (my $badchars = $newdna) =~ s/[atgcATGC]//g;

#      if (length($badchars) > 0) {
#	print("ERROR: Bad characters \"$badchars\" in clone " . $clone->id . "\n");
#      }
      
#      $linkdna .= $newdna;

    };


    if ($@) {
      die "Problem with clone " . $clone->id ;
    }             
  }

#  print("Length of DNA  is " . (length($linkdna)) . "\n");
#  print("Length of link is $linklen\n");

#  $linkdna =~ s/(.{72})/$1\n/g;

#  return $linkdna;
}
  

sub make_link {
    my ($self,$cloneids) = @_;

    my $db       = $self->acedb;
    my @cloneids = @$cloneids;

    # Generate the link
    my $id        = shift @cloneids;
    my $link      = new Spangle::Link($id,$db);

    # Put the first clone in the link
    my $clone    = new Spangle::Clone;
       $clone->id($id);
       $clone->acedb($db);

    my $dna      = new Bio::Seq(-seq => $clone->fetch_DNA(0));

    $dna->id($clone->id);
    
    $clone->start (1);
    $clone->end   ($dna->length);
    $clone->strand(1);
    
    $link->add_clone($clone);
    
    my $dna1   = $dna;

    my $clone1 = new Spangle::Clone($id,$db);
       $clone1->start(1);
       $clone1->end($dna1->length);
       $clone1->strand(1);
    
    my $linkend = 0;

    foreach my $id (@cloneids) {
	
	my $clone2 = Spangle::Clone->new;
  
	$clone2->id   ($id);
	$clone2->acedb($db);

	my $dna2      = new Bio::Seq(-seq => $clone2->fetch_DNA(0));
	
	$dna2->id($id);
	
	print("\n### Running blast to join clones " . $clone1->id . " "  . $clone2->id . "\n");
	my $blast = new Spangle::BlastEnd($dna1,$dna2);
	
	my ($orientation,$offset) =    $blast->find_overlap;
	
	print("### Done running blast\n\n");

	if (!defined($orientation) || !defined($offset)) {
	    die "ERROR: Couldn't join clones " . $clone1->id . " and " . $clone2->id . "\n";
	}
	
	print("Overlap  is $offset\n");
	
	$clone2->start($linkend + $offset);
	$clone2->end  ($linkend + $offset + $dna2->length - 1);
	
	if ($orientation == 2 || $orientation == 4) {
	    print("Setting strand for " . $clone2->id . " to -1\n");
	    $clone2->strand(-1);
	} else {
	    print("Setting strand for " . $clone2->id . " to 1\n");
	    $clone2->strand(1);
	}

	if ($linkend == 0 && ($orientation == 3 || $orientation == 4)) {
	    my @tmp = $link->eachClone;
	    $tmp[0]->strand(-1);
	    print("Resetting " . $tmp[0]->id  . " strand to -1\n");
	}
	
	$link->add_clone($clone2);
	
	$clone1->id    ($clone2->id);
	$clone1->start ($clone2->start);
	$clone1->end   ($clone2->end);
	$clone1->strand($clone2->strand);
	
	$dna1 = $dna2;
	
	$linkend   = $clone1->start -1;
    }

    return $link;
}

sub replace_clone {
    my ($self,$cloneid) = @_;

    print("\nRemoving clone $cloneid from " . $self->id . "\n");
    my ($clone,$link1,$link2) = $self->remove_clone_by_ID($cloneid); # Split the link

    my @clones1 = $link1->eachClone;
    my @clones2 = $link2->eachClone;

    print("Making a new link\n");

    if ($#clones1 == -1) {
	$link2->print;
    
	my $clone2 = $clones2[0];

	my @newclones;
	
	push(@newclones,$clone->id,$clone2->id);
	
	my $newlink = $self->make_link(\@newclones);

	$newlink->merge($link2);
	$newlink->id($self->id);
	
	return $newlink;

    } elsif ($#clones2 == -1) {
	$link1->print;
    
	my $clone1 = $clones1[$#clones1];

	my @newclones;
    
	push(@newclones,$clone1->id,$clone->id);
    
	my $newlink = $self->make_link(\@newclones);

	$link1->merge($newlink);
	$link1->id($self->id);

	return $link1;

    } else {
	
	my $clone1 = $clones1[$#clones1];
	my $clone2 = $clones2[0];
	
	my @newclones;
	
	push(@newclones,$clone1->id,$clone->id,$clone2->id);
	
	my $newlink = $self->make_link(\@newclones);
	
	$link1->merge($newlink);
	$link1->merge($link2);
	
	$link1->id($self->id);
	
	return $link1;
    }
}
    








