package Spangle::Subseq;

=pod

=head1 NAME
  
Spangle::Subseq - ACEDB subsequence object

=head1 SYNOPSIS
  
  $cl = new Spangle::Subseq
  
=head1 DESCRIPTION

  A module to store and manipulate subsequences
  attached to an ACEDB clone.

=head1 CONTACT

  Describe contact details here

=head1 APPENDIX

  The rest of the documentation details each of the object methods. Internal methods are usually preceded with a _

=cut

use strict;  
use vars   qw(@ISA);


# Object preamble - inherits from Bio::SeqFeature::Generic

use Spangle::Transcript;

@ISA = qw(Spangle::Transcript);

sub new {
    my($pkg, @args) = @_;
    
    my $self = $pkg->SUPER::new(@args);
  
    # Initialize all data structures here
    $self->{_method}          = "";
    $self->{_locus}           = "";
    $self->{_start_not_found} = "";
    $self->{_end_not_found}   = "";
    $self->{_remark}          = [];
    $self->{_note}            = "";
    $self->{_CDS_start}       = "";
    $self->{_CDS_end}       = "";
    $self->{_pseudogene}      = "";
    $self->{_frame}                = "";
    
    $self->{_printFlag} = 1;
  
    return $self;
}

=pod

=head2 check_translation

  Title   : check_translation
  Usage   : $clone->check_translation
  Function: Checks the splice sites and exons for translatable ORFs.
  Returns : nothing
  Args    : none

=cut

=head2 add_supporting_feature

  Title   : add_supporting_feature
  Usage   : $clone->add_supporting_feature
  Function: Adds a piece of supporting Evidence to a subsequence
  Returns : nothing
  Args    : Spangle::Exon,Bio::SeqFeature::Generic

=cut

sub add_supporting_feature {
  my ($self,$exon,$f) = @_;

  $exon->add_sub_SeqFeature($f);

}

=head2 eachSupportingFeature

  Title   : eachSupportingFeature
  Usage   : @f = $clone->eachSupportingFeature
  Function: Returns all supporting features
  Returns : @Bio::SeqFeature::Generic
  Args    : none

=cut

sub eachSupportingFeature {
  my $self = shift;

  }

=head2 exonOverlap

  Title   : exonOverlap
  Usage   : $clone->exonOverlap($sub)
  Function: Calculates how many exons overlap with another subsequence 
  Returns : int (number of exons)
  Args    : Subsequence

=cut

=head2 primary_tag

  Title   : type
  Usage   : $clone->primary_tag($str)
  Function: Sets the type of the subsequence(proposed,predicted etc)
  Returns : stirng
  Args    : none or string

=cut

sub primary_tag {
  my($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_primary_tag} = $arg;
  } 

  return $self->{_primary_tag};
}

=head2 source_tag

  Title   : source_tag
  Usage   : $clone->source_tag($source)
  Function: Sets the source of the subseuqence (GENSCAN,Spangle etc)/
  Returns : string
  Args    : none or string

=cut

sub source_tag {
  my($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_source_tag} = $arg;
  } 

  return $self->{_source_tag};
}

=head2 start

  Title   : start
  Usage   : $clone->start
  Function: Sets the start coord of the subseuqence 
  Returns : int
  Args    : none or int

=cut

sub start {
  my($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_start} = $arg;
#    print("Changing start for " . $self->source_tag . " " . $arg . "\n");
  } 

  return $self->{_start};
}

=head2 end

  Title   : end
  Usage   : $clone->end($end)
  Function: Sets the end coord of the subseuqence 
  Returns : int
  Args    : none or int

=cut

sub end {
  my($self,$arg) = @_;

  if (defined($arg)) {
    $self->{_end} = $arg;
  } 

  return $self->{_end};
}

=head2 strand

  Title   : strand
  Usage   : $clone->strand($strand)
  Function: Sets the strand of the subseuqence (+,1 or -,-1)
  Returns : string or int
  Args    : none or int

=cut

sub strand {
  my($self,$arg) = @_;

  if (defined($arg)) {
    if ($arg eq "+" || $arg == 1) {
      $self->{_strand} = 1;
    } elsif($arg eq "-" || $arg == -1) {
      $self->{_strand} = -1;
    }
  }
  return $self->{_strand};
}

=head2 phase

  Title   : phase
  Usage   : $clone->phase($phase)
  Function: Sets the phase of the subsequence (0,1,2)
  Returns : 0,1,2,
  Args    : none or 0,1,2

=cut

sub phase {
  my($self,$arg) = @_;

  if (defined($arg)) {
    if ($arg <= 2 && $arg >= 0) {
      $self->{_phase} = $arg;
    }
  }
  return $self->{_phase};
}

=head2 method

  Title   : method
  Usage   : $clone->method
  Function: Gets the method of the subsequence
  Returns : string
  Args    : none or string

=cut

sub method {
  my($self,$arg) = @_;

  if (defined($arg)) {
      $self->{_method} = $arg;
  }

  return $self->{_method};
}


=head2 locus

  Title   : locus
  Usage   : $clone->locus
  Function: Gets the locus name 
  Returns : string
  Args    : none or string

=cut

sub locus {
  my($self,$arg) = @_;

  if (defined($arg)) {
      $self->{_locus} = $arg;
  }

  return $self->{_locus};
}


=head2 addRemark

  Title   : addRemark
  Usage   : $clone->addRemark($s)
  Function: Adds a remark to the array
  Returns : string
  Args    : none or string

=cut

sub addRemark {
  my($self,$arg) = @_;

  if (defined($arg)) {
      push(@{$self->{_remark}},$arg);
  }

}

=head2 eachRemark

  Title   : eachRemark
  Usage   : @rem = $clone->eachRemark
  Function: Returns the remark array
  Returns : array of string
  Args    : none

=cut

sub eachRemark {
  my($self) = @_;

  return @{$self->{_remark}};

}

=head2 note

  Title   : note
  Usage   : $clone->note
  Function: Gets the note string
  Returns : string
  Args    : none or string

=cut

sub note {
  my($self,$arg) = @_;

  if (defined($arg)) {
      $self->{_note} = $arg;
  }

  return $self->{_note};
}


=head2 frame

  Title   : frame
  Usage   : $clone->frame
  Function: Returns offset value for translation of start_not_found subsequence
  Returns : 1, 2, 3
  Args    : none or 1, 2, 3

=cut

sub frame {
  my($self,$arg) = @_;

  if (defined($arg)) {
        if ($arg <= 3 && $arg >= 1) {
            $self->{_frame} = $arg;
        }
  }

  return $self->{_frame};
}



=head2 start_not_found

  Title   : start_not_found
  Usage   : $clone->start_not_found
  Function: Returns true if the start of the gene is not in this subseuqence
  Returns : 0,1
  Args    : none or 0,1

=cut

sub start_not_found {
  my($self,$arg) = @_;

  if (defined($arg)) {
      $self->{_start_not_found} = $arg;
  }

  return $self->{_start_not_found};
}


=head2 end_not_found

  Title   : end_not_found
  Usage   : $clone->end_not_found
  Function: Returns true if the end of the gene is not in this subseuqence
  Returns : 0,1
  Args    : none or 0,1

=cut

sub end_not_found {
  my($self,$arg) = @_;

  if (defined($arg)) {
      $self->{_end_not_found} = $arg;
  }

  return $self->{_end_not_found};
}

=head2 pseudogene

  Title   : pseudogene
  Usage   : $clone->pseudogene
  Function: Returns true if the gene is a pseudogene
  Returns : 0,1
  Args    : none or 0,1

=cut

sub pseudogene {
  my($self,$arg) = @_;

  if (defined($arg)) {
      $self->{_pseudogene} = $arg;
  }

  return $self->{_pseudogene};
}


=head2 CDS_start

  Title   : CDS_start
  Usage   : $clone->CDS_start
  Function: Returns number of the base that starts the coding region (local coordinates)
  Returns : int
  Args    : none or int

=cut

sub CDS_start {
  my($self,$arg) = @_;

  if (defined($arg)) {
      $self->{_CDS_start} = $arg;
  }

  return $self->{_CDS_start};
}

=head2 CDS_end

  Title   : CDS_end
  Usage   : $clone->CDS_end
  Function: Returns number of the base that ends the coding region (local coordinates)
  Returns : int
  Args    : none or int

=cut

sub CDS_end {
  my($self,$arg) = @_;

  if (defined($arg)) {
      $self->{_CDS_end} = $arg;
  }

  return $self->{_CDS_end};
}

=head2 find_supporting_evidence

  Title   : find_supporting_evidence
  Usage   : $subseq->find_supporting_evidence($homolset)
  Function: Given a Homolset of database hits finds which
            ones overlap with the subsequence coords
  Returns : nothing
  Args    : HomolSet

=cut

sub find_supporting_evidence {
  my ($self,$homolset) = @_;
  
  #print STDERR "Finding for $self $homolset\n"; 
  defined($homolset) || $self->throw("No homolset input. Can't find supporting evidence");

  my @homols = $homolset->each_feature;

  foreach my $ex ($self->each_Exon) {
    foreach my $h (@homols) {
      # Set the length if necessary - this involves more calls than necessary
      # We should store the lengths in the homolset. Maybe get_lengths in homolset object?

      if (!defined($h->length)) {
	$homolset->get_length($h);
        #print STDERR "Got length\n";
      }

      my ($lo,$ro,@overlap) = $h->test_overlap($ex);
      #print STDERR "Tested overlap\n";
      if ($overlap[0]) {
	$ex->add_supporting_feature($h);
      }
      #print STDERR "done\n";
    }
  }
}

=head2 printFlag

  Title   : printFlag
    Usage   : $clone->printFlag($print);
  Function: Sets a flag saying whether to print the subsequence or not
  Returns : 0,1
  Args    : 0,1

=cut

sub printFlag {
  my ($self,$arg) = @_;
  

  if (defined($arg)) {

      if ($arg == 0 || $arg == 1) {
	  $self->{_printFlag} = $arg;
      } else {
	  $self->throw("Can't set printFlag to $arg. Must be 0 or 1\n");
      }
  }

  return $self->{_printFlag};
}
# Here is the Transcript documentation


=head2  add_Exon

     Title   : add_Exon
     Usage   : $trans->add_Exon($exon)
     Returns : Nothing
     Args    :

=head2  each_Exon

     Title   : each_Exon
     Usage   : foreach $exon ( $trans->each_Exon)
     Function: Returns an array of exons in the transcript
     Example : my @exons = $tr->each_Exon
     Returns : An array of exon objects
     Args    : none

=head2  flush_Exon

     Title   : flush_Exon
     Usage   : Removes all Exons from the array.
     Function:
     Example :
     Returns : 
     Args    :

=head2  translate

     Title   : translate
     Usage   : $pep = $feat->translate()
     Function: Returns the peptide translation of the gene - in the correct phase
     Returns : Bio::Seq
     Args    : none

=head2  dna_seq

     Title   : dna_Seq
     Usage   : $dna = $feat->dna_seq
     Function: Returns the dna sequence of the gene
     Returns : Bio::Seq
     Args    : none

=head2  sort

     Title   : sort
     Usage   : $feat->sort()
     Function: Sorts the exon features by start coordinate
               Sorts forward for forward strand and reverse for reverse strand
     Returns : none
     Args    : none

=cut
