
### Hum::Fox::AceData

package Hum::Fox::AceData;

use strict;
use Carp;
use Hum::Fox;
use Hum::Ace::AceText;

sub new {
    my( $pkg ) = @_;
    
    return bless {}, $pkg;
}

sub fetch_current_by_name {
    my( $pkg, $name ) = @_;
    
    confess "no name" unless $name;
    return $pkg->_generic_fetch(qq{ is_current = 'Y' AND name = '$name' });
}

sub fetch_by_name_and_version {
    my( $pkg, $name, $version ) = @_;
    
    confess "no name"    unless $name;
    confess "no version" unless $version;
    return $pkg->_generic_fetch(qq{ name = '$name' AND version = $version });
}

sub fetch_by_db_id {
    my( $pkg, $id ) = @_;
    
    confess "no db_id" unless $id;
    my $moniker = $pkg->moniker;
    return $pkg->_generic_fetch(qq{ ${moniker}_id = $id });
}

sub _generic_fetch {
    my( $pkg, $where_clause ) = @_;

    my $moniker = $pkg->moniker;
    my $sth = Hum::Fox::fox_prepare_statement(qq{
        SELECT ${moniker}_id
          , name
          , version
          , who
          , UNIX_TIMESTAMP(edit_time)
          , ace_data
        FROM ace_$moniker
        WHERE $where_clause
        });
    $sth->execute;
    
    my ($id, $name, $version, $who, $when, $data) = $sth->fetchrow
        or return;  # undef on failure
    
    $data ||= '';
    
    my $self = $pkg->new;
    $self->db_id    ($id);
    $self->name     ($name);
    $self->version  ($version);
    $self->who      ($who);
    $self->when     ($when);
    $self->data     ($data);
    
    return $self;
}

sub fetch_from_id_sth {
    my( $pkg, $sth ) = @_;
    
    $sth->execute;
    
    my( @obj );
    while (my ($id) = $sth->fetchrow) {
        my $acedata = $pkg->fetch_by_db_id($id);
        push(@obj, $acedata);
    }
    return @obj;
}

sub fetch_by_link_one_to_many {
    my( $pkg, @obj ) = @_;
    
    return $pkg->_fetch_generic_by_link_one_to_many(
        q{ AND many.is_current = 'Y' }, @obj,
        );
}

sub fetch_archival_by_link_one_to_many {
    my( $pkg, @obj ) = @_;
    
    return $pkg->_fetch_generic_by_link_one_to_many('', @obj);
}

sub _fetch_generic_by_link_one_to_many {
    my( $pkg, $where_clause, @obj ) = @_;
    
    confess "no objects given" unless @obj;
    
    my $one_in_clause = join(',', map "'". $_->db_id ."'", @obj);
    my $one  = $obj[0]->moniker;
    my $many =    $pkg->moniker;
    
    my $join_id = $many . '_id';
    my $sql = qq{
        SELECT many.$join_id
        FROM ace_${one}_link_$many lnk
          , ace_$many many
        WHERE lnk.$join_id = many.$join_id
          $where_clause
          AND lnk.${one}_id IN ($one_in_clause)
        };
    #warn $sql;
    my $sth = Hum::Fox::fox_prepare_statement($sql);
    
    return $pkg->fetch_from_id_sth($sth);
}

sub fetch_by_link_many_to_one {
    my( $pkg, @obj ) = @_;
    
    confess "no objects given" unless @obj;
    
    my $many_in_clause = join(',', map "'". $_->db_id ."'", @obj);
    my $one  =    $pkg->moniker;
    my $many = $obj[0]->moniker;
    
    my $join_id = $one . '_id';
    my $sql = qq{
        SELECT DISTINCT(one.$join_id)
        FROM ace_${one}_link_$many lnk
          , ace_$one one
        WHERE lnk.$join_id = one.$join_id
          AND one.is_current = 'Y'
          AND lnk.${many}_id IN ($many_in_clause)
        };
    #warn $sql;
    my $sth = Hum::Fox::fox_prepare_statement($sql);
    
    return $pkg->fetch_from_id_sth($sth);
}

sub fetch_previous {
    my( $self ) = @_;
    
    my $version = $self->version;
    return if $version <= 1;
    $version--;
    return ref($self)->fetch_by_name_and_version($self->name, $version);
}

sub moniker {
    my( $self ) = @_;
    
    my $pkg = ref($self);
    confess("Module '$pkg' does not implement the 'moniker' method");
}

sub ace_class {
    my( $self ) = @_;
    
    my $pkg = ref($self);
    confess("Module '$pkg' does not implement the 'ace_class' method");
}

sub db_id {
    my( $self, $db_id ) = @_;
    
    if ($db_id) {
        $self->{'_db_id'} = $db_id;
    }
    return $self->{'_db_id'};
}

sub name {
    my( $self, $name ) = @_;
    
    if ($name) {
        $self->{'_name'} = $name;
    }
    return $self->{'_name'};
}

sub version {
    my( $self, $version ) = @_;
    
    if (defined $version) {
        $self->{'_version'} = $version;
    }
    return $self->{'_version'} || 0;
}

sub who {
    my( $self, $who ) = @_;
    
    if ($who) {
        $self->{'_who'} = $who;
    }
    return $self->{'_who'} || Hum::Fox::current_uname();
}

sub when {
    my( $self, $when ) = @_;
    
    if ($when) {
        $self->{'_when'} = $when;
    }
    return $self->{'_when'} || time;
}

sub data {
    my( $self, $data ) = @_;
    
    if (defined $data) {
        $self->{'_data'} = $data;
    }
    return $self->{'_data'} || '';
}

# Useful if you want to hand munge the data
sub mutable_data {
    my( $self, $mut ) = @_;
    
    if ($mut) {
        $self->data($mut->ace_string);
    } else {
        return Hum::Ace::AceText->new($self->data);
    }
}

sub ace_string {
    my( $self ) = @_;
    
    my $class = $self->ace_class;
    my $name  = $self->name or confess "name not set";
    my $data  = $self->data;
    return qq{\n$class : "$name"\n$data\n};
}

{
    my($self_file, $othr_file);
    
    END {
        unlink($self_file) if $self_file;
        unlink($othr_file) if $othr_file;
    }
    
    sub ace_diff {
        my( $self, $othr ) = @_;
        
        $self_file ||= "/tmp/ace_diff.self.$$";
        $othr_file ||= "/tmp/ace_diff.othr.$$";
        
        local( *SELF, *OTHR );
        open SELF, "> $self_file" or confess "Can't write to '$self_file' : $!";
        open OTHR, "> $othr_file" or confess "Can't write to '$othr_file' : $!";
        print SELF $self->ace_string;
        print OTHR $othr->ace_string;
        close SELF;
        close OTHR;
        
        my $diff = `acediff $self_file $othr_file 2>/dev/null`;
        unlink($self_file, $othr_file);
        $diff =~ s{//.*}{}mg;
        if ($diff =~ /\w/) {
            warn "Got diff: >>$diff<<\n";
            return $diff;
        } else {
            return;
        }
    }
}

sub differ {
    my( $self, $othr ) = @_;
    
    my $self_data = $self->_sorted_ace_data;
    my $othr_data = $othr->_sorted_ace_data;
    
    # Return true if different (not equal)
    return $self_data ne $othr_data;
}

sub _sorted_ace_data {
    my( $self ) = @_;
    
    return join("\n", sort split /\n/, $self->ace_string) . "\n";
}

sub fuse {
    my( $self, @others ) = @_;
    
    my( %ace_lines );
    foreach my $obj ($self, @others) {
        my $data = $obj->data;
        while ($data =~ /(.+)/g) {
            $ace_lines{$1} = 1;
        }
    }
    $self->data(join '', map "$_\n", sort keys %ace_lines);
}

sub store_new_version {
    my( $self ) = @_;
    
    my $pkg = ref($self);
    my $name = $self->name;
    if (my $othr = $pkg->fetch_current_by_name($name)) {
        
        #warn "Got ", ref($self), " '$name'\n";
        
        # Put the current version number into self
        $self->version($othr->version);
        
        unless ($self->differ($othr)) {
            # Same as stored version, so fill in
            # db_id in self and return.
            $self->db_id($othr->db_id);
            return;
        }
    }
    
    # Store if not in db, or is new version
    $self->store;
}

sub store {
    my( $self ) = @_;
    
    my $name        = $self->name or confess "name not set";
    if (my $id = $self->db_id) {
        confess ref($self), " '$name' is already stored with db_id '$id'";
    }
    my $moniker     = $self->moniker;
    my $version     = $self->get_next_version;
    my $when        = $self->when;
    my $who         = $self->who;
    my $ace_data    = $self->data;
    
    my ($type) = ref($self) =~ /([^:]+)$/;
    
    warn "  storing $type '$name' v$version\n";
    
    my $table = "ace_$moniker";
    
    my $set_not_current = Hum::Fox::fox_prepare_statement(qq{
        UPDATE $table SET is_current = 'N' WHERE name = '$name'
        });
    
    my $insert = Hum::Fox::fox_prepare_statement(qq{
        INSERT $table (${moniker}_id, name, version, is_current, edit_time, who, ace_data)
        VALUES (NULL, ?, ?, 'Y', FROM_UNIXTIME(?), ?, ?)
        });
    
    $set_not_current->execute;
    $insert->execute($name, $version, $when, $who, $ace_data);
    
    my $db_id = $insert->{'mysql_insertid'}
        or confess "No insertid from statement handle";
    $self->db_id($db_id);
    
    # Save version back to object
    $self->version($version);
}

sub get_next_version {
    my( $self ) = @_;
    
    my $moniker = $self->moniker;
    my $name    = $self->name;
    my $sth = Hum::Fox::fox_prepare_statement(qq{
        SELECT MAX(version) from ace_$moniker WHERE name = '$name'
        });
    $sth->execute;
    
    my ($max) = $sth->fetchrow;
    $max ||= 0;
    $max++;
    return $max;
}

sub store_link {
    my( $self, $link ) = @_;
    
    my $self_m  = $self->moniker;
    my $link_m  = $link->moniker;
    
    my $self_col = "${self_m}_id";
    my $link_col = "${link_m}_id";
    
    my $table   = "ace_${link_m}_link_$self_m";
    
    my $self_id = $self->db_id
        or confess "No db_id in ", ref($self), "  ", $self->name;
    my $link_id = $link->db_id
        or confess "No db_id in ", ref($link), "  ", $link->name;
    
    my $sth = Hum::Fox::fox_prepare_statement(
        qq{ REPLACE $table ($link_col, $self_col) VALUES ($link_id, $self_id) }
        );
    $sth->execute;
}

sub set_not_current {
    my( $self ) = @_;
    
    my $id      = $self->db_id;
    my $moniker = $self->moniker;
    my $table = 'ace_' . $moniker;
    my $id_col = $moniker . '_id';
    my $sth = Hum::Fox::fox_prepare_statement(
        qq{ UPDATE $table SET is_current = 'N' WHERE $id_col = $id }
        );
    $sth->execute;
}

1;

__END__

=head1 NAME - Hum::Fox::AceData

=head1 AUTHOR

James Gilbert B<email> jgrg@sanger.ac.uk

