
### Hum::Fox::Lock

package Hum::Fox::Lock;

use strict;
use Carp;
use Hum::Fox;
use Hum::Submission qw{ acetime };
use Sys::Hostname 'hostname';

sub make_lock {
    my( $pkg, $ace_dir ) = @_;
    
    confess "Missing ace_dir argument" unless $ace_dir;
    
    my $who       = Hum::Fox::current_uname();
    my $when      = time();
    my $host_name = hostname();
    
    my $self = bless {}, $pkg;
    $self->who($who);
    $self->when($when);
    $self->host_name($host_name);
    $self->ace_dir($ace_dir);
    $self->write_lock;
    
    return $self;
}

sub fetch_Lock_by_dir_and_host {
    my( $pkg, $dir, $host ) = @_;
    
    confess "Need dir and host; got '$dir' and '$host'"
        unless $dir and $host;
    return $pkg->_fetch_generic(qq{ ace_dir = '$dir' AND host_name = '$host' });
}

sub fetch_Lock_by_id {
    my( $pkg, $id ) = @_;

    confess "No id given" unless $id;
    return $pkg->_fetch_generic(qq{ ace_lock_id = $id });
}

sub _fetch_generic {
    my( $pkg, $where_clause ) = @_;

    # Can be class or instance method
    if (my $p = ref($pkg)) {
        $pkg = $p;
    }
    
    my $sth = Hum::Fox::fox_prepare_statement(qq{
        SELECT ace_lock_id
          , who
          , UNIX_TIMESTAMP(edit_time)
          , host_name
          , ace_dir
        FROM ace_lock
        WHERE $where_clause
        });
    $sth->execute;
    
    my ($id, $who, $when, $host_name, $ace_dir) = $sth->fetchrow;
    confess "No ace_lock matching '$where_clause'" unless $id;

    my $self = bless {}, $pkg;
    $self->ace_lock_id($id);
    $self->who($who);
    $self->when($when);
    $self->host_name($host_name);
    $self->ace_dir($ace_dir);
    
    return $self
}

sub ace_lock_id {
    my( $self, $ace_lock_id ) = @_;
    
    if ($ace_lock_id) {
        $self->{'_ace_lock_id'} = $ace_lock_id;
    }
    return $self->{'_ace_lock_id'};
}

sub who {
    my( $self, $who ) = @_;
    
    if ($who) {
        $self->{'_who'} = $who;
    }
    return $self->{'_who'};
}

sub when {
    my( $self, $when ) = @_;
    
    if ($when) {
        $self->{'_when'} = $when;
    }
    return $self->{'_when'};
}

sub host_name {
    my( $self, $host_name ) = @_;
    
    if ($host_name) {
        $self->{'_host_name'} = $host_name;
    }
    return $self->{'_host_name'};
}

sub ace_dir {
    my( $self, $ace_dir ) = @_;
    
    if ($ace_dir) {
        $self->{'_ace_dir'} = $ace_dir;
    }
    return $self->{'_ace_dir'};
}

sub write_lock {
    my( $self ) = @_;
    
    my $who       = $self->who       or confess       "who not set";
    my $when      = $self->when      or confess      "when not set";
    my $host_name = $self->host_name or confess "host_name not set";
    my $ace_dir   = $self->ace_dir   or confess   "ace_dir not set";
    
    my $sth = Hum::Fox::fox_prepare_statement(qq{
        INSERT ace_lock(ace_lock_id
              , who
              , edit_time
              , host_name
              , ace_dir)
        VALUES (NULL
              , '$who'
              , FROM_UNIXTIME($when)
              , '$host_name'
              , '$ace_dir')
        });
    $sth->execute;
    my $ace_lock_id = $sth->{'mysql_insertid'}
        or confess "failed to get insertid from sth";
    $self->ace_lock_id($ace_lock_id);
}

sub lock_objects {
    my( $self, @obj ) = @_;
    
    foreach my $o (@obj) {
        $self->lock_object($o);
    }
}

sub lock_object {
    my( $self, $obj ) = @_;
    
    my $ace_lock_id = $self->ace_lock_id
        or confess "ace_lock_id missing.  Has lock been written?";

    my $name   = $obj->moniker;
    my $obj_id = $obj->db_id
        or confess "got no db_id from object '$obj'";

    my $table = 'ace_lock_'. $name;
    my $id_field = $name . '_id';
    my $sth = Hum::Fox::fox_prepare_statement(qq{
        INSERT $table ($id_field, ace_lock_id)
        VALUES ($obj_id, $ace_lock_id)
        });
    eval{ $sth->execute };
    if ($@) {
        if ($@ =~ /duplicate/i) {
            die $self->nice_lock_error($table, $id_field, $obj_id);
        } else {
            confess $@;
        }
    } else {
        return 1;
    }
}

sub nice_lock_error {
    my( $self, $table, $id_field, $obj_id ) = @_;
    
    my $sth = Hum::Fox::fox_prepare_statement(qq{
        SELECT ace_lock_id
        FROM $table
        WHERE $id_field = $obj_id
        });
    $sth->execute;
    my ($lock_id) = $sth->fetchrow;
    
    my $lock = $self->fetch_Lock_by_id($lock_id);
    return "Error $table exists for '$obj_id' and is locked by:\n\n  "
        . $lock->nice_string . "\n\n";
}

sub nice_string {
    my( $self ) = @_;
    
    return sprintf(
        "%s on %s since %s (dir = '%s')",
        $self->who,
        $self->host_name,
        acetime($self->when),
        $self->ace_dir,
        );
}

sub fingerprint {
    my( $self ) = @_;
    
    my ($prog) = $0 =~ m{([^/]+)$};
    my $who = $self->who;
    my $when = acetime($self->when);
    return "$prog-$who-$when";
}

sub list_locked_ids {
    my( $self, $moniker ) = @_;
    
    my $id = $self->ace_lock_id
        or confess "Missing ace_lock_id";
    my $table = 'ace_lock_'. $moniker;
    my $id_col = $moniker .'_id';
    my $sth = Hum::Fox::fox_prepare_statement(
        qq{ SELECT $id_col FROM $table WHERE ace_lock_id = $id }
        );
    $sth->execute;
    
    my( @lock_id );
    while (my ($lck) = $sth->fetchrow) {
        push(@lock_id, $lck);
    }
    return @lock_id;
}

sub release_lock {
    my( $self ) = @_;
    
    my $id = $self->ace_lock_id
        or confess "ace_lock_id not set";
    foreach my $table ('ace_lock', $self->lock_table_list) {
        my $sth = Hum::Fox::fox_prepare_statement(
            qq{ DELETE FROM $table WHERE ace_lock_id = $id }
            );
        $sth->execute;
    }
}

{
    my @lock_tables = qw{
        ace_lock_genomic
        ace_lock_subseq
        ace_lock_ana_seq
        ace_lock_locus
        };

    sub lock_table_list {
        return @lock_tables;
    }
}

1;

__END__

=head1 NAME - Hum::Fox::Lock

=head1 AUTHOR

James Gilbert B<email> jgrg@sanger.ac.uk

