
### Hum::Fox::Session

package Hum::Fox::Session;

use strict;
use Carp;
use Fcntl qw{ O_WRONLY O_EXCL O_CREAT };
use Symbol 'gensym';

use Hum::Fox::Lock;
use Hum::Fox::AceData::GenomeSequence;
use Hum::Fox::AceData::SubSequence;
use Hum::Fox::AceData::Locus;
use Hum::AnaStatus::Sequence;

sub new {
    my( $pkg ) = @_;
    
    return bless {}, $pkg;
}

sub open_fh_to_ace_file {
    my( $self, $file ) = @_;
    
    my $fh = gensym();
    open $fh, "> $file" or confess "Can't write to '$file' : $!";
    $self->ace_data_filehandle($fh);
    return $fh;
}

sub ace_dbh {
    my( $self, $dbh ) = @_;
    
    if ($dbh) {
        $self->{'_ace_dbh'} = $dbh;
    }
    return $self->{'_ace_dbh'};
}

sub disconnect_ace_dbh {
    my( $self ) = @_;

    $self->{'_ace_dbh'} = undef;
}

sub close_ace_data_filehandle {
    my( $self ) = @_;
    
    my $fh = $self->ace_data_filehandle;
    close($fh) or confess "Error from closing ace_data_filehandle : $!";
    $self->{'_ace_data_filehandle'} = undef;
}

sub ace_data_filehandle {
    my( $self, $fh ) = @_;
    
    if ($fh) {
        $self->{'_ace_data_filehandle'} = $fh;
    }
    return $self->{'_ace_data_filehandle'};
}

sub session_lock {
    my( $self, $dir ) = @_;
    
    if ($dir) {
        my( $lock );
        if (UNIVERSAL::isa($dir, 'Hum::Fox::Lock')) {
            $lock = $dir;
        } else {
            $lock = Hum::Fox::Lock->make_lock($dir);
        }
        $self->{'_session_lock'} = $lock;
    }
    return $self->{'_session_lock'};
}

sub fetch_ace_data_for_ana_seqs {
    my( $self, @ana_seq ) = @_;
    
    # Lock the ana_seq objects
    my $lock = $self->session_lock;
    if ($lock) {
        $lock->lock_objects(@ana_seq);
        
        # Make sure the fox list file is created
        $self->fox_list_write_handle

        ### Realign old subsequences here?
    }
    
    
    # Test if we have genomic sequence in annotation
    # store - may not if it is the first time this
    # sequence is annotated.
    ### Should lock anyway?
    if (my @genomic = Hum::Fox::AceData::GenomeSequence
        ->fetch_by_link_one_to_many(@ana_seq))
    {
        # Record db_id's of GenomeSequence objects
        $self->print_ace_data(@genomic);
        
        # May not have any SubSequences, so test as well
        if (my @subseq = Hum::Fox::AceData::SubSequence
            ->fetch_by_link_one_to_many(@genomic))
        {
            # Record db_id's of GenomeSequence objects
            $self->print_ace_data(@subseq);
            
            # Now get loci
            if (my @locus = Hum::Fox::AceData::Locus
                ->fetch_by_link_many_to_one(@subseq))
            {
                $self->print_ace_data(@locus);
                
                ## Now fetch any subsequences attached to this locus
                ## that we haven't yet seen.
                #if (my @locus_subseq =
                #    $self->fetch_by_link_one_to_many($locus))
                #{
                #    my $seen_subseq = map {$_->name, 1} @subseq;
                #    my( @unseen_locus_subseq );
                #    foreach my $loc_sub (@locus_subseq) {
                #        push(@unseen_locus_subseq, $loc_sub)
                #            unless $seen_subseq{$loc_sub->name};
                #    }
                #    if (@unseen_locus_subseq) {
                #        $self->print_ace_data(@unseen_locus_subseq);
                #    }
                #}
            }
        }
    }
    
    $self->close_fox_list_write_handle if $lock;
}

sub print_ace_data {
    my( $self, @obj ) = @_;
    
    my $fh = $self->ace_data_filehandle
        or confess "Missing ace_data_filehandle";
    my $lock = $self->session_lock;
    if ($lock) {
        $lock->lock_objects(@obj);
        $self->save_fox_list(@obj);
    }
    foreach my $acedata (@obj) {
        print $fh $acedata->ace_string;
    }
}

sub save_fox_list {
    my( $self, @obj ) = @_;
    
    my $fox_fh = $self->fox_list_write_handle;
    foreach my $data (@obj) {
        print $fox_fh join("\t",
            ref($data),
            $data->db_id,
            $data->ace_class,
            $data->name,
            ), "\n";
    }
}

sub fox_list_write_handle {
    my( $self ) = @_;
    
    my( $fox_fh );
    unless ($fox_fh = $self->{'_fox_list_write_handle'}) {
        my $fox_file = $self->fox_list_path;
        $fox_fh = gensym();
        sysopen($fox_fh, $fox_file, O_WRONLY | O_EXCL | O_CREAT)
            or confess "Can't write to new file '$fox_file' : $!";
        $self->{'_fox_list_write_handle'} = $fox_fh;
    }
    return $fox_fh;
}

sub close_fox_list_write_handle {
    my( $self ) = @_;
    
    if (my $fh = $self->{'_fox_list_write_handle'}) {
        close($fh) or confess "Error closing fox_list_write_handle : $!";
        $self->{'_fox_list_write_handle'} = undef;
    }
}

sub fox_list_path {
    my( $self ) = @_;
    
    my $dir = $self->session_lock->ace_dir;
    my $fox_list = "$dir/.fox_list";
}

sub record_changed_objects_in_acediff {
    my( $self, $diff ) = @_;
    
    my $changed  = $self->list_changed_objects_in_acediff($diff);
    my $fox_file = $self->fox_list_path;
    local *FOX_LIST;
    open FOX_LIST, $fox_file or die "Can't read '$fox_file' : $!";
    while (<FOX_LIST>) {
        chomp;
        my ($fox_class, $db_id, $ace_class, $ace_name) = split /\t/, $_, 4;
        if ($changed->{$ace_class}{$ace_name}) {
            $self->add_change_record($fox_class, $ace_name, $db_id);
        }
    }
    close FOX_LIST;
}

sub add_change_record {
    my( $self, $fox_class, $ace_name, $db_id ) = @_;
    
    $self->{'_changes'}{$fox_class}{$ace_name} = $db_id;
}

sub delete_change_record {
    my( $self, $obj ) = @_;
    
    my $fox_class = ref($obj);
    my $ace_name = $obj->name;
    
    delete $self->{'_changes'}{$fox_class}{$ace_name};
}

sub list_changed_objects_in_acediff {
    my( $self, $diff ) = @_;
    
    local $/ = "";
    local(*ACEDIFF);
    open ACEDIFF, $diff or confess "Can't read '$diff' : $!";
    
    my $change = {};
    while (<ACEDIFF>) {
        next if m{^//}s;
        my ($class, $name) = /^(\S+)\s+"([^"]+)/s;
        unless ($class) {
            warn "\nCan't see class and name in:\n$_";
            next;
        }
        $change->{$class}{$name} = 1;
    }
    unless (keys %$change) {
        warn "No changes detected\n";
    }
    return $change;
}

sub set_remaining_changed_not_current {
    my( $self ) = @_;
    
    foreach my $fox_class (keys %{$self->{'_changes'}}) {
        my ($obj_type) = $fox_class =~ /([^:]+)$/;
        my $class_hash = $self->{'_changes'}{$fox_class};
        while (my ($name, $db_id) = each %$class_hash) {
            warn " deleting $obj_type '$name'\n";
            my $obj = $fox_class->fetch_by_db_id($db_id);
            $obj->set_not_current;
        }
    }
}

sub save_annotation_data {
    my( $self ) = @_;
    
    my $lock = $self->session_lock
        or confess "No session_lock - Can't store data";

    my @ana_seq_id = $lock->list_locked_ids('ana_seq');
    my( @ana_seq );
    foreach my $asid (@ana_seq_id) {
        my $seq = Hum::AnaStatus::Sequence
            ->new_from_ana_seq_id($asid);
        push(@ana_seq, $seq);
    }
    $self->save_ace_data_for_ana_seqs(undef, @ana_seq);
        
    $self->set_remaining_changed_not_current;
}

sub save_ace_data_for_ana_seqs {
    my( $self, $store_time, @ana_seq ) = @_;
    
    # Record the time, so that all the objects
    # stored by this method get the same time.
    $store_time ||= time();
    
    my( %subseq_names, %subseq_genomic );
    foreach my $seq (@ana_seq) {
        my $full_name = $seq->full_sequence_name;
        warn "Storing '$full_name'\n";

        unless ($self->ace_find('Genome_Sequence', $full_name)) {
            warn "Can't find Genome_Sequence '$full_name'\n";
            next;
        }

        my $genomic = Hum::Fox::AceData::GenomeSequence->new;
        $genomic->name($full_name);
        $genomic->when($store_time);

        my $genomic_data = '';

        # Get the subsequences
        my $subseq_tags = $self->clean_raw_answer("show -a Subsequence");

        # Subsequence         "dJ313I6.1" 29502 31123
        my $subseq_pat = qr/^Subsequence\s+"($full_name\.\d[^"]*)/;
        foreach my $line (split /\n/, $subseq_tags) {
            if (my ($sub_name) = $line =~ /$subseq_pat/) {
                $genomic_data .= $line . "\n";
                $subseq_names{$sub_name}++;
                $subseq_genomic{$sub_name} = $genomic;
            }
        }


        # Get data from the Genome_sequence
        foreach my $tag (qw{

            Clone_left_end 
            Clone_right_end

            Annotation

            Keyword

            })
        {
            my $ans = $self->clean_raw_answer("show -a $tag");

            # Remove first line
            $ans =~ s/^Sequence.+$//m;

            # Add all the other lines
            while ($ans =~ /^(\w.*\n)/mg) {
                $genomic_data .= $1;
            }
        }

        # Get EMBL_dump_info DE_line and CC_line data.
        my $embl_data = $self->clean_raw_answer('show -a EMBL_dump_info');
        foreach my $line (split /\n/, $embl_data) {
            if ($line =~ /(DE|CC)_line/) {
                $genomic_data .= $line . "\n";
            }
        }

        # Get poly-A features
        my $features = $self->clean_raw_answer('show -a Feature');
        foreach my $line (split /\n/, $features) {
            $genomic_data .= "$line\n" if $line =~ /^Feature\s+"[pP]olyA/i;
        }
        $genomic->data($genomic_data);
        $genomic->store_new_version;
        $genomic->store_link($seq); # Link back to ana_seq
        $self->delete_change_record($genomic);
    }
    ### End of foreach AnaSequence
    
    # Get all the subseqences
    my( %loci );
    foreach my $name (keys %subseq_names) {
        if (my $data = $self->get_ace_data('Sequence', $name)) {
            my( $locus );
            if (my ($loc_name) = $data =~ /Locus\s+"([^"]+)/) {
                unless ($locus = $loci{$loc_name}) {
                    $locus = $self->get_and_store_locus($loc_name, $store_time);
                    $loci{$loc_name} = $locus;
                }
            }
            my $sub = Hum::Fox::AceData::SubSequence->new;
            $sub->name($name);
            $sub->data($data);
            $sub->when($store_time);
            $sub->store_new_version;
            
            my $genomic = $subseq_genomic{$name}
                or confess "No Genomic sequence '$name'";
            
            $sub->store_link($genomic);
            $sub->store_link($locus) if $locus;
            
            $self->delete_change_record($sub);
        } else {
            warn "No SubSequence object named '$name'\n";
        }
    }
    
    # Store links from the loci back to any subsequences
    # that weren't in this edit Session.
    foreach my $locus (values %loci) {
        my $data = $locus->data;
        while ($data =~ /^Positive_sequence\s+"([^"]+)/mg) {
            my $name = $1;
            unless ($subseq_names{$name}) {
                my $sub = Hum::Fox::AceData::SubSequence
                    ->fetch_current_by_name($name);
                if ($sub) {
                    $sub->store_link($locus);
                } else {
                    warn "No such SubSequence '$name'\n";
                }
            }
        }
    }
}

sub get_and_store_locus {
    my( $self, $name, $store_time ) = @_;

    # Get all the locus objects
    if (my $data = $self->get_ace_data('Locus', $name)) {
        my $locus = Hum::Fox::AceData::Locus->new;
        $locus->name($name);
        $locus->data($data);
        $locus->when($store_time);
        
        unless ($self->delete_change_record($locus)) {
            # If there isn't a change record for this locus, but
            # there is a current locus in the database, then
            # we need to merge in the data from the existing
            # locus.
            my $current = Hum::Fox::AceData::Locus
                ->fetch_current_by_name($name);
            if ($current) {
                $locus->fuse($current);
            }
        }
        
        $locus->store_new_version;
        return $locus;
    } else {
        warn "No Locus object named '$name'\n";
        return;
    }
}

sub ace_find {
    my( $self, $class, $name ) = @_;
    
    my $ans = $self->raw_answer(qq{find $class "$name"});
    
    my ($n) = $ans =~ /Found (\d+) objects/;
    return $n;
}

sub get_ace_data {
    my( $self, $class, $name ) = @_;
    
    if ($self->ace_find($class, $name)) {
        my $data = $self->clean_raw_answer('show -a');
        
        # Remove object name header
        $data =~ s/^\s*$class.+?\n//s
            or confess "Can't remove header from data:\n$data";
        
        return $data;
    }
    return;
}

sub clean_raw_answer {
    my( $self, $query ) = @_;
    
    my $ans = $self->raw_answer($query);
        
    # Remove comments
    $ans =~ s{^//.+\n}{}mg;
    
    # Replace '\n\<newline>' with '\n'
    $ans =~ s/\\n\\\n/\\n/g;
    
    return $ans;
}

sub raw_answer {
    my( $self, $query ) = @_;
    
    my $ace = $self->ace_dbh
        or confess "No ace_dbh (handle to ACeDB database)";
    my $ans = $ace->raw_query($query);
    
    # Remove NULLs -- we never want them
    $ans =~ s/\0//g;
    
    return $ans;
}

1;

__END__

=head1 NAME - Hum::Fox::Session

=head1 AUTHOR

James Gilbert B<email> jgrg@sanger.ac.uk

