
### Bio::EnsEMBL::Ace::Filter::Gene

package Bio::EnsEMBL::Ace::Filter::Gene;

use strict;
use Carp;
use Bio::SeqIO;
use Data::Dumper;
use Bio::EnsEMBL::Ace::Filter;
use vars '@ISA';
@ISA = ('Bio::EnsEMBL::Ace::Filter');


sub new {
    my ($class) = @_;

    my $self = bless {}, $class;

    return $self;
}

sub ace_data {
    my ( $self, $contig ) = @_;
        
    my @transcripts = $self->get_transcripts( $contig );
        
    return $self->ace_data_from_transcripts( $contig, \@transcripts );
}

sub get_transcripts {
    my ( $self, $contig ) = @_;
    
    if ( $contig ) {
        my( @transcript_list );
        my @genes = $contig->get_Genes_by_Type( $self->analysis_object->logic_name, 0 );
        foreach my $gene (@genes) {
            my $gene_name = $gene->stable_id;
            foreach my $t ($gene->each_Transcript) {
                $t->gene_name($gene_name);
                push(@transcript_list, $t);
            }
        }
        return @transcript_list;
    } else {
        confess "No contig given";
    }
}

sub ace_data_from_transcripts {
    my ( $self, $contig, $transcripts ) = @_;

    my $contig_id = $contig->id;
    my $genomic_ace = qq{\nSequence "$contig_id"\n};
    my $subseq_ace  = '';
    
    foreach my $transcript ( @$transcripts ) {
        my( $gen, $sub );
        eval{
            ($gen, $sub) = $self->_make_ace_data($contig, $transcript);
        };
        if ($@) {
            warn $@;
        } else {
            $genomic_ace .= $gen;
            $subseq_ace  .= $sub;
        }
    }

    my $url_ace = '';
    
    if (my $url_str = $self->url_string) {
        my %genes = map {$_, 1} grep $_, map $_->gene_name, @$transcripts;
        foreach my $gene_name (keys %genes) {
            my $url = sprintf($url_str, $gene_name);
            #my $url = "http\\:\\/\\/www.ensembl.org\\/Homo_sapiens\\/geneview?gene=%s";
            $url_ace .= qq{\nURL : "$gene_name"\n}
                .qq{URL "$url"\n};
        }
    }

    if ($genomic_ace) {
        return qq{\nSequence : "$contig_id"\n$genomic_ace\n$subseq_ace\n$url_ace\n};
    }
    else {
        return '';
    }
}

sub next_gene_int {
    my( $self ) = @_;
    
    return ++$self->{'_gene_int'};
}

sub _make_ace_data {
    my( $self, $contig, $transcript, $gene_id ) = @_;

    ### Need to append number to stable_id
    my $transcript_id = $transcript->stable_id;
    
    my $contig_id = $contig->id;
    
    my ($start_not_found, $end_not_found, @exons_in_transcript) =
        $self->transcript_exons_in_contig( $contig, $transcript );
    unless (@exons_in_transcript) {
        die "transcript '$transcript_id' has no exons in contig!\n";
    }
    #print STDERR "\nTRAN: ", $transcript->stable_id, "\n";
    #foreach my $ex (@exons_in_transcript) {
    #    printf STDERR "  %9d  %9d  %s\n", $ex->start, $ex->end, ($ex->strand == 1 ? '+' : '-');
    #}

    # Append number to incomplete transcripts in case
    # the other part of the transcript is in another
    # clone in the same session.
    if ($start_not_found or $end_not_found) {
        my $i = $self->next_gene_int;
        $transcript_id .= ".$i";
        $transcript->stable_id($transcript_id);
    }

    my $subseq_data = qq{\nSequence : "$transcript_id"\nSource "$contig_id"\n};
    

    ### We don't deal properly with non-coding transcripts or
    ### transcripts that have a cds which is outside this contig.
    my( $ace_phase );
    if (my $translation = $transcript->translation) {
        my ($CDS_start, $CDS_end) = $self->CDS_start_end($translation, @exons_in_transcript);
        $subseq_data .= qq{CDS $CDS_start $CDS_end\n};
        $ace_phase = $self->ace_phase(@exons_in_transcript);
    }

    if ($start_not_found or $ace_phase) {
        if (defined $ace_phase) {
            $subseq_data .= qq{Start_not_found $ace_phase\n};
        } else {
            $subseq_data .= qq{Start_not_found\n};
        }
    }

    if ($end_not_found) {
        $subseq_data .= qq{End_not_found\n};
    }

    ### This line will need to change for non-ensembl genes!
    my $meth_name = $self->analysis_object->logic_name;
    #$subseq_data .= qq{Web_location "Sanger-EnsEMBL"\nCDS_predicted_by "$meth_name"\nMethod "$meth_name"\n};
    $subseq_data .= qq{CDS_predicted_by "$meth_name"\nMethod "$meth_name"\n};
    if (my $description = $transcript->description) {
        $subseq_data .= qq{Remark "$description"\n};
    }
    if (my $gene_name = $transcript->gene_name) {
        $subseq_data .= qq{Web_location "$gene_name"\n};
    }

    my $first_exon_start = $self->first_exon_start( @exons_in_transcript );
    my $last_exon_end    = $self->last_exon_end   ( @exons_in_transcript );

    if ($exons_in_transcript[0]->strand == 1) {
        $subseq_data .= $self->fwd_strand_exon_list( $first_exon_start, @exons_in_transcript );
    } else {
        $subseq_data .= $self->rev_strand_exon_list(    $last_exon_end, @exons_in_transcript );
    }

    # The following coordinates have a compensating +1 to deal with an off
    # by one in the Ensembl API for certain gene types!!!
    if ($self->is_off_by_one) {
        $first_exon_start++;
        $last_exon_end++;
    }

    ### Now deal with offsets and inversions on golden path
    my $map_contig = $self->get_contig_overlapping($contig, $first_exon_start, $last_exon_end);
    if ($map_contig->orientation == 1) {
        my $raw_start = $map_contig->rawcontig_start;
        $first_exon_start += $raw_start - 1;
        $last_exon_end    += $raw_start - 1;
    } else {
        my $raw_end = $map_contig->rawcontig_end;
        $first_exon_start = $raw_end - $first_exon_start + 1;
        $last_exon_end    = $raw_end - $last_exon_end    + 1;
    }

    if ($exons_in_transcript[0]->strand == -1) {
        ($first_exon_start, $last_exon_end) = ($last_exon_end, $first_exon_start);
    }
    return(qq{Subsequence "$transcript_id" $first_exon_start $last_exon_end\n}, $subseq_data);
}

sub get_contig_overlapping {
    my( $self, $contig, $start, $end ) = @_;
    
    foreach my $map_ctg ($contig->_vmap->each_MapContig) {
        if ($start <= $map_ctg->end and $end >= $map_ctg->start) {
            return $map_ctg;
        }
    }
    confess "Can't find MapContig overlapping '$start' to '$end'\n";
}

{
    my %phase = (
        "0" => "1",
        "1" => "3",
        "2" => "2",
    );

    sub ace_phase {
        my ( $self, @exons_in_transcript ) = @_;

        my $start_phase = $exons_in_transcript[0]->phase;
        return $phase{$start_phase};
    }
}

sub first_exon_start {
    my ( $self, @exons_in_transcript ) = @_;

    my $first_exon_start;
    if ( $exons_in_transcript[0]->strand == 1 ) {
        $first_exon_start = $exons_in_transcript[0]->start;
    }
    else {
        $first_exon_start = $exons_in_transcript[ $#exons_in_transcript ]->start;
    }
    return $first_exon_start;
}

sub last_exon_end {
    my ( $self, @exons_in_transcript ) = @_;

    my $last_exon_end;

    if ( $exons_in_transcript[0]->strand == 1 ) {
        $last_exon_end = $exons_in_transcript[ $#exons_in_transcript ]->end;
    }

    else {
        $last_exon_end = $exons_in_transcript[0]->end;
    }
    return $last_exon_end;
}

sub CDS_start_end {
    my ( $self, $translation, @exons_in_transcript, ) = @_;

    my $start_exon = $translation->start_exon;
    my   $end_exon = $translation->end_exon;

    my( $CDS_start, $CDS_end );
    my $spliced_length = 0;
    foreach my $exon (@exons_in_transcript) {
        if ($exon == $start_exon) {
            $CDS_start = $spliced_length + $translation->start;
        }
        if ($exon == $end_exon) {
            $CDS_end   = $spliced_length + $translation->end;
            last;   # But it's probably the last one anyway!
        }
        $spliced_length += $exon->length;
    }
    $CDS_start ||= 1;
    $CDS_end   ||= $spliced_length;
    return($CDS_start, $CDS_end);
}

sub fwd_strand_exon_list {
    my( $self, $pos, @exons_in_transcript ) = @_;
    
    my $exon_list = '';
    foreach my $exon (@exons_in_transcript) {
        my $start = $exon->start - $pos + 1;
        my $end   = $exon->end   - $pos + 1;
        $exon_list .= qq{Source_Exons $start $end\n};
    }
    return $exon_list;

}

sub rev_strand_exon_list {
    my( $self, $pos, @exons_in_transcript ) = @_;
    
    my $exon_list = '';
    foreach my $exon (@exons_in_transcript) {
        my $start = $pos - $exon->end   + 1;
        my $end   = $pos - $exon->start + 1;
        $exon_list .= qq{Source_Exons $start $end\n};
    }
    return $exon_list;
}

              
sub transcript_exons_in_contig {
    my( $self, $contig, $transcript ) = @_;

    unless ($contig and $transcript) {
        confess "Need both contig and transcript";
    }

    my $contig_id = $contig->id;

    # Get the ids of all the rawcontigs that make up the VirtualContig
    my %raw_contig_id = map {$_->contig->id, 1} $contig->_vmap->each_MapContig;
    #warn map "  contig ID = $_\n", keys %raw_contig_id;

    # Split sticky exons
    #my @exons = map {$_->isa('Bio::EnsEMBL::StickyExon') ? $_->each_component_Exon : $_}
    #    $transcript->get_all_Exons;
    
    # Reject sticky exons
    my @exons = map {$_->isa('Bio::EnsEMBL::StickyExon') ? () : $_}
        $transcript->get_all_Exons;

    my( @exons_in_transcript );
    my $snf = 0;
    my $enf = 0;
    foreach my $exon (@exons) {
        my $ex_con_id = $exon->contig->id;

        # The exon will either have the id of one of the RawContigs
        # in the VirtualContig (eg: Ensembl genes), or will have the
        # same name as the VirtualContig (eg: genscans).
        # Maybe this will be consistent in the new schema?
        if ($ex_con_id eq $contig_id or $raw_contig_id{$ex_con_id}) {
            push(@exons_in_transcript, $exon);
        }
        else {
            #warn "exon '$ex_con_id' is not in VirtualContig\n";
            if (@exons_in_transcript) {
                $enf = 1;
            } else {
                $snf = 1;
            }
        }
    }       

    return($snf, $enf, @exons_in_transcript);    
}

sub is_off_by_one {
    my ( $self, $obo ) = @_;

    if (defined $obo) {
        $self->{'_obo'} = $obo;
    }
    
    return $self->{'_obo'};
}

sub url_string {
    my( $self, $url_string ) = @_;
    
    if ($url_string) {
        $self->{'_url_string'} = $url_string;
    }
    return $self->{'_url_string'};
}


1;

__END__

=head1 NAME - Bio::EnsEMBL::Ace::Filter::Gene

=head1 ArUTHOR

Stephen Keenan B<email> keenan@sanger.ac.uk

