### Bio::EnsEMBL::Ace::Filter::Similarity

package Bio::EnsEMBL::Ace::Filter::Similarity;

use strict;
use Carp;
use Data::Dumper;
use strict;
use Bio::SeqIO;
use Bio::EnsEMBL::Ace::Filter;
use vars '@ISA';
@ISA = ('Bio::EnsEMBL::Ace::Filter');

sub new {
    my ($class) = @_;

    my $self = bless {}, $class;
    return $self;
}

sub ace_data {
    my ( $self, $contig ) = @_;

    my $seqname = $contig->id;

    my $ace_data = '';
    my $count    = 0;

    my $ana_obj = $self->analysis_object or confess "analysis_object not set";
    my $feature_type = $ana_obj->logic_name;
    my $ana_id       = $ana_obj->dbID;

    my $raw_contig_id = $self->contig_internal_id($contig);
    my $db_obj = $contig->dbobj;
    my $sth = $db_obj->prepare(qq{
        SELECT seq_start
          , seq_end
          , hid
          , hstart
          , hend
          , strand
          , perc_id
        FROM feature
        WHERE contig = $raw_contig_id
          AND analysis = $ana_id
        ORDER BY hid
        });
    $sth->execute;
    my ( $start,  $end,  $hid,  $hstart,  $hend,  $strand,  $percent_id);
    $sth->bind_columns(
        \$start, \$end, \$hid, \$hstart, \$hend, \$strand, \$percent_id);
    
    my $homol_method = $self->method_tag;
    my $homol_tag    = $self->homol_tag;
    my $class        = $self->hit_class;
    my $percent_cutoff = $self->percent_identity_cutoff;

    my $current_hid = '';
    my $current_full;
    my $hit_ace  = '';
    my $gen_ace = qq{\nSequence : "$seqname"\n};
    while ($sth->fetch) {
        # Using percent ID instead of score
        if ($class eq 'Protein') {
            # Bug in BPlite?  blastx has max 33% identity
            $percent_id = 1 + (3 * $percent_id);
        }

        if ($percent_cutoff) {
            next unless $percent_id > $percent_cutoff;
        }
    
        if ($hid ne $current_hid) {
            $current_hid = $hid;
            my( $prefix, $desc ) = $self->get_hseqprefix_description($hid, $db_obj);
            unless ($prefix) {
                $prefix = $self->hseq_prefix;
                warn "Missing description for '$hid'\n";
            }
            ### Backslash escape description text
            $current_full = $prefix . $hid;
            $hit_ace .= qq{\n$class : "$current_full"\nTitle "$desc"\n};
        }
        if ($strand == -1) {
            ($hstart, $hend) = ($hend, $hstart);
            
            # Alternate way of switching strand:
            #($start, $end) = ($end, $start);
        }
        
        # Using score
        #$hit_ace .= qq{DNA_homol "$seqname" "$homol_method" $score $hstart $hend $start $end\n};
        #$gen_ace .= qq{$homol_tag "$current_full" "$homol_method" $score $start $end $hstart $hend\n};
        $hit_ace .= qq{DNA_homol "$seqname" "$homol_method" $percent_id $hstart $hend $start $end\n};
        $gen_ace .= qq{$homol_tag "$current_full" "$homol_method" $percent_id $start $end $hstart $hend\n};
    }

    return $gen_ace . $hit_ace;
        #. qq{\nMethod : "$homol_method"\nScore_bounds 60 100\nWidth 2\n};
}


sub percent_identity_cutoff {
    my( $self, $percent_identity_cutoff ) = @_;
    
    if (defined $percent_identity_cutoff) {
        $self->{'_percent_identity_cutoff'} = $percent_identity_cutoff;
    }
    return $self->{'_percent_identity_cutoff'};
}


#sub swall_prefix {
#    my ( $self, $hid ) = @_;
#
#    if ( $hid =~ /_/ ) {
#        return 'Sw:';
#    }
#    else {
#        return 'Tr:';
#    }
#}

sub homol_tag {
    my ( $self, $homol_tag ) = @_;

    if ($homol_tag) {
        $self->{'_homol_tag'} = $homol_tag;
    }
    return $self->{'_homol_tag'};

}

sub method_tag {
    my ( $self, $method_tag ) = @_;

    if ($method_tag) {
        $self->{'_method_tag'} = $method_tag;
    }
    return $self->{'_method_tag'};

}

sub hseq_prefix {
    my ( $self, $hseq_prefix ) = @_;

    if ($hseq_prefix) {
        $self->{'_hseq_prefix'} = $hseq_prefix;
    }
    return $self->{'_hseq_prefix'};

}

sub get_hseqprefix_description {
    my ( $self, $feature_hid, $dbh ) = @_;

    my $hit_desc_hid = '';
    my $query = qq{
SELECT hit_db_prefix, description
FROM hit_descriptions.hit_description
WHERE hid = '$feature_hid'
};
   
    my $sth = $dbh->prepare($query);
    $sth->execute();
    my ($hit_prefix, $description) = $sth->fetchrow;

    unless ($description) {
        $description = 'No description found';
    }

    return $hit_prefix, $description;
}

sub hit_class {
    my ($self) = @_;

    my ($class);
    if ( $self->analysis_object->program =~ /blastx/i ) {
        $class = 'Protein';
    }
    else {
        $class = 'Sequence';
    }
    return $class;
}

1;

__END__

=head1 NAME - Bio::EnsEMBL::Ace::Filter::Similarity

=head1 AUTHOR

Stephen Keenan B<email> keenan@sanger.ac.uk

